/** Module event
 *
 * For handling events in workspace via LOCAL source
 *
 * Requires:
 *  - local3x.qrt named 'LOCAL'
 *  - qu.ql (found in common_ql/utils)
 *  - logging.ql (found in common_ql/logging)
 *
 * Idea:
 * To control execution of out functions attached to tables and graphs. Calling
 * is done from an out function to the next by referring to the next function's
 * name. A function can call as many other functions as desired.
 *
 * All functions to be involved must register themselves with their name, so
 * they can be called from another function. Function calling is always done by
 * the name of the out function to run.
 *
 * NOTE: Involved out functions should never throw exceptions!
 *
 * Example:
 *
 * out void update_swap()
 * {
 *   event.this_function("update_swap"); // Always start with this line
 *   ...
 *   event.call_function("show_swaps"); // Show swaps again (they have changed...)
 *   event.call_function("show_prices"); // Show (recalculated) prices again after update
 *   ...
 *   event.next();  // Always do this before end of function and any return statement in the function.
 * }
 *
 * out matrix(string) show_prices()
 * {
 *   event.this_function("show_prices"); // Always, so it can be called. Here it is called from update_swap()
 *   ...
 *   event.next(); // Even if this function has no event.call_function(...) it must call event.next(), since there could be other calls waiting. Here the call to show_swaps is probably waiting to happen.
 * }
 *
 *
 * Details:
 * The function calls are kept as a queue, so the next function to be called by event.next()
 * will be the last one requested by event.call_function(...). In the example
 * above show_prices is called first, then any function requested by
 * show_prices, and then show_swaps.
 *
 * No call to a function which is not yet registered will be made. A function
 * is registered by calling event.this_function(func_name). This is to prevent
 * calls to functions that aren't event attached to any table (calling those
 * would bring the chain of calls to event.next() to a halt since the
 * unattached function would not be run).
 *
 */


qu.count_set __RIC_COUNT = new qu.count_set();
string __LOCAL_EVENT_NUMBER_FID = "event_number";
void __conditional_publish(string ric) {
    if (__RIC_COUNT.get_count(ric) == 0) {
        local.publish('LOCAL', ric, [__LOCAL_EVENT_NUMBER_FID], ['0']);
        sleep(2); // To ensure ric is ready for reading
        __RIC_COUNT.increase(ric);
    }
}
void __local_event(string ric) {
    __RIC_COUNT.increase(ric);
    number num = __RIC_COUNT.get_count(ric);
    local.publish('LOCAL', ric, [__LOCAL_EVENT_NUMBER_FID], [str(num)]);
}
void __listen_to_event(string ric, string option(nullable) func_name = null<string>) {
    __conditional_publish(ric);
    rt.get(ric, __LOCAL_EVENT_NUMBER_FID, 'LOCAL');
}

module event {
    logging.Context logging_context = logging.context("EventManager");


    // Dummy class to allow making settings in compile stage
    class preset_obj {};
    preset_obj __preset_obj__;

    public enum event_type option(category: "Workspace") {
        ET_REQUEST option(str: "request"),
            ET_SKIPPED_CALL option(str: "skipped call"),
            ET_CALL option(str: "call"),
            ET_RUN_PLANNED option(str: "run"),
            ET_RUN_UNPLANNED option(str: "run (unplanned)"),
            ET_SKIPPED_NEXT option(str: "skipped next"),
            ET_RUN option(str: "unused") // Exists for backward COMPILE compability - no events will be of this type
    };

    public class Event option(category: "Workspace") {
    public:
        Event(string func_name, timestamp time, event_type type);
        string func_name();
        timestamp time();
        event_type type();

        void add_info(string info);
        string extra_info();

        void set_state_desc(string desc);
        string state_desc();

    private:
        string func_name_;
        timestamp time_;
        event_type type_;
        vector(string) extra_info_;
        string state_desc_;
    };
    Event.Event(string func_name, timestamp time, event_type type) :
        func_name_(func_name),
        time_(time),
        type_(type)
        {}
    string Event.func_name() = func_name_;
    timestamp Event.time() = time_;
    event_type Event.type() = type_;
    void Event.add_info(string info) { push_back(extra_info_, info); }
    string Event.extra_info() = qu.str_join(extra_info_, ", ");
    void Event.set_state_desc(string desc) { state_desc_ = desc; }
    string Event.state_desc() = state_desc_;




    number RL_NOTHING = 0;
    number RL_INFO = 1;
    number RL_DEBUG = 2;

    public class EventManager option(category: "Workspace") {
    public:
        EventManager();
        void enter_function(string func_name);
        void push(string func_name);
        void next();

        void remember();
        void remember_debug();
        void forget();
        logical user_controlled_history();
        void enable();
        void disable();

        void enable_timing();
        void disable_timing();

        vector(string) event_func_names();
        vector(timestamp) event_times();
        vector(event_type) event_types();
        vector(string) event_info();
        vector(string) event_states();
        vector(string) event_requesters(); // Backward compatibility

    private:
        logical enabled_;
        qu.queue_str_unique call_queue_;
        string expected_function_; // called but not yet run
        set_str registered_func_names_;
        number remember_level_;
        string last_registered_; // set by enter_function(), unset by next(), remembered along with pushes
        Event last_start_event_; // set by enter_function(), unset by next()
        vector(Event) events_;
        logical user_controlled_history_; // True if a history controlling function has been called
        logical use_timing_; // True if this_function() and next() should start and stop timers

        void _remember();
        void _remember_debug();
        void _forget();
        string _state_description();
    };
    EventManager.EventManager() :
        enabled_(true),
        call_queue_(new qu.queue_str_unique()),
        registered_func_names_(set_str()),
        remember_level_(RL_NOTHING),
        user_controlled_history_(false),
        use_timing_(false)
    {
        // logging.set_logger(new logging.FileLogger("C:\\temp\\eventmgr.log"));
        logging.get_file_logger().clear_file();
        _forget();
    }
    void EventManager.enter_function(string func_name)
    {
        if (!enabled_) return;
        if (use_timing_) qu.timing.start_top(func_name);
        __listen_to_event(func_name);
        logical planned = qu.equaln(func_name, expected_function_);
        registered_func_names_.add(func_name);
        logging.info(logging_context, [func_name, " is run", planned ? "" : " (unplanned)"]);
        last_registered_ = func_name;
        if (remember_level_ >= RL_INFO) {
            last_start_event_ = new Event(func_name, now(), planned ? ET_RUN_PLANNED : ET_RUN_UNPLANNED);
            last_start_event_.set_state_desc(_state_description());
            push_back(events_, last_start_event_);
        }
        if (planned) {
            expected_function_ = null<string>;
        }
        // Clear out queued calls to this function
    }
    void EventManager.push(string func_name)
    {
        if (!enabled_) return;
        call_queue_.push(func_name);
        logging.debug(logging_context, ["push ", func_name]);
        if (remember_level_ >= RL_INFO) {
            Event event = new Event(func_name, now(), ET_REQUEST);
            event.add_info(strcat(["By ", last_registered_]));
            event.set_state_desc(_state_description());
            push_back(events_, event);
        }
    }
    void EventManager.next()
    {
        if (!enabled_) return;
        if (use_timing_) qu.timing.end(last_registered_);
        string func_to_call = null<string>;
        if (null(expected_function_)) {
            while (null(func_to_call) && !call_queue_.empty()) {
                func_to_call = call_queue_.pop();
                if (!registered_func_names_.includes(func_to_call)) {
                    logging.info(logging_context, ["skipping call to ", func_to_call, " (not registered)"]);
                    if (remember_level_ >= RL_DEBUG) {
                        Event event = new Event(func_to_call, now(), ET_SKIPPED_CALL);
                        event.set_state_desc(_state_description());
                        push_back(events_, event);
                    }
                    func_to_call = null<string>;
                }
            }
        } else {
            if (remember_level_ >= RL_DEBUG) {
                Event e = new Event(last_registered_, now(), ET_SKIPPED_NEXT);
                e.add_info(strcat(["Waiting for ", expected_function_]));
                e.set_state_desc(_state_description());
                push_back(events_, e);
            }
        }

        if (null(func_to_call)) {
            logging.info(logging_context, "nothing to call");
        } else {
            logging.info(logging_context, ["calling ", func_to_call]);
            __local_event(func_to_call);
            expected_function_ = func_to_call;
            if (remember_level_ >= RL_INFO) {
                Event event = new Event(func_to_call, now(), ET_CALL);
                event.set_state_desc(_state_description());
                push_back(events_, event);
            }
        }
        if (remember_level_ >= RL_INFO) {
            last_start_event_.add_info(strcat([str(now() - last_start_event_.time()), " ms"]));
            last_start_event_ = null<Event>;
        }
        last_registered_ = null<string>;
    }
    void EventManager.remember()
    {
        user_controlled_history_ = true;
        _remember();
    }
    void EventManager.remember_debug()
    {
        user_controlled_history_ = true;
        _remember_debug();
    }
    void EventManager.forget()
    {
        user_controlled_history_ = true;
        _forget();
    }
    logical EventManager.user_controlled_history() = user_controlled_history_;
    void EventManager.enable() { enabled_ = true; }
    void EventManager.disable() { enabled_ = false; }

    void EventManager.enable_timing() { use_timing_ = true; }
    void EventManager.disable_timing() { use_timing_ = false; }
    vector(string) EventManager.event_func_names() = clone_vector(events_.func_name());
    vector(timestamp) EventManager.event_times() = clone_vector(events_.time());
    vector(event_type) EventManager.event_types() = clone_vector(events_.type());
    vector(string) EventManager.event_info() = clone_vector(events_.extra_info());
    vector(string) EventManager.event_states() = clone_vector(events_.state_desc());

    // Backward compatibility
    vector(string) EventManager.event_requesters() = this.event_info();

    void EventManager._remember() { remember_level_ = RL_INFO; }
    void EventManager._remember_debug() { remember_level_ = RL_DEBUG; }
    void EventManager._forget()
    {
        remember_level_ = RL_NOTHING;
        last_registered_ = null<string>;
        resize(events_, 0);
    }

    string EventManager._state_description()
    {
        if (!enabled_) {
            return "disabled";
        }
        string desc = "";
        if (remember_level_ < RL_DEBUG) {
            return desc;
        }

        if (!null(expected_function_)) {
            desc = strcat(["Expects ", expected_function_, ". "]);
        }

        desc = strcat([desc, "Queue: [", qu.str_join(call_queue_.contents(), ", "), "]"]);
        desc = strcat([desc, " Current: ", last_registered_]);
        return desc;
    }


    EventManager __event_manager__ = new EventManager();

    logical history_shown = false;

    matrix(text_rgb) _history_matrix(logical reversed)
    {
        if (!__event_manager__.user_controlled_history()) {
            // If settings function hasn't been called, user hasn't taken
            // control of history handling. Since the user calls this
            // function, it is obvious that history is desired, so
            // start remembering!
            // Unfortunately, at the first call to this function won't
            // no history will exist - but that can't be fixed since we
            // don't want to record history without the user wanting it.
            __event_manager__.remember();
        }

        vector(string) fnames = __event_manager__.event_func_names();
        vector(timestamp) times = __event_manager__.event_times();
        vector(event_type) types = __event_manager__.event_types();
        vector(string) infos = __event_manager__.event_info();
        vector(string) states = __event_manager__.event_states();

        integer n_rows = v_size(fnames);
        integer n_cols = 5;

        if (n_rows == 0) {
            matrix(text_rgb) ret_mx[1, n_cols];
            return ret_mx;
        }

        matrix(text_rgb) ret_mx[n_rows, n_cols];

        for (integer row = 0; row < n_rows; ++row) {
            integer index = reversed ? n_rows - row - 1 : row;
            string fname = fnames[index];
            string time  = str(times[index]);
            string type  = string(types[index]);
            string info  = null(infos[index]) ? "" : infos[index];
            string state = states[index];

            number color = null<number>;
            switch (types[index]) {
                case ET_REQUEST:
                    color = rgb(240, 230, 140);
                    break;
                case ET_SKIPPED_CALL:
                    color = rgb(200, 200, 200);
                    break;
                case ET_CALL:
                    color = rgb(150, 250, 190);
                    break;
                case ET_RUN_PLANNED:
                    color = rgb(170, 210, 250);
                    break;
                case ET_RUN_UNPLANNED:
                    color = rgb(190, 225, 250);
                    break;
                case ET_SKIPPED_NEXT:
                    color = rgb(230, 230, 230);
                    break;
            }

            ret_mx[row, 0] = text_rgb(time, null<number>, color, false);
            ret_mx[row, 1] = text_rgb(type, null<number>, color, false);
            ret_mx[row, 2] = text_rgb(fname, null<number>, color, false);
            ret_mx[row, 3] = text_rgb(info, null<number>, color, false);
            ret_mx[row, 4] = text_rgb(state, null<number>, color, false);
        }

        history_shown = true;
        return ret_mx;
    }


    void _settings(logical events_disabled, logical history_enabled, logical keep_shown_history, logical debug_level)
    {
        if (events_disabled) {
            __event_manager__.disable();
        } else {
            __event_manager__.enable();
        }

        if (history_enabled) {
            if (!keep_shown_history && history_shown) {
                __event_manager__.forget();
            }
            if (debug_level) {
                __event_manager__.remember_debug();
            } else {
                __event_manager__.remember();
            }
        } else {
            __event_manager__.forget();
        }
        history_shown = false;
    }

    /**** PUBLIC INTERFACE BELOW ****/


    /** Mandatory functions for correct behavior **/
    // Always call at start of out function
    public void this_function(string func) option(category: "Workspace") { __event_manager__.enter_function(func); }
    // Call whenever to put another function in call queue
    public void call_function(string func) option(category: "Workspace") { __event_manager__.push(func); }
    // Always call before leaving out function
    public void next() option(category: "Workspace") { __event_manager__.next(); }

    /** Totally disable events (and re-enable) **/
    public preset_obj disable_events() { __event_manager__.disable(); return __preset_obj__; }
    public preset_obj enable_events() { __event_manager__.enable(); return __preset_obj__; }

    /** History keeping **/
    public preset_obj remember() { __event_manager__.remember(); return __preset_obj__; }
    public preset_obj remember_debug() { __event_manager__.remember_debug(); return __preset_obj__; }
    public preset_obj forget() { __event_manager__.forget(); return __preset_obj__; }

    /** Timing **/
    public preset_obj enable_timing() { __event_manager__.enable_timing(); return __preset_obj__; }
    public preset_obj disable_timing() { __event_manager__.disable_timing(); return __preset_obj__; }

    /** To use as out functions in workspace tables **/
    // Get a neat matrix with recorded event history. History recording must be enabled, either through settings function below or explicit call to remember or remember_debug above.
    public matrix(text_rgb) history_matrix(logical reversed = false) = _history_matrix(reversed);
    // Settings, if desired
    public void settings(logical events_disabled, logical history_enabled, logical keep_shown_history, logical debug_level) { _settings(events_disabled, history_enabled, keep_shown_history, debug_level); }


    /*** DEPRECATED - will be removed ***/
    EventManager get_manager() = __event_manager__;

}
