option(null: hard);

module http_api {

void url_percent_decode(out string s)
{
    // https://www.urlencoder.org/
	if (null(s))
		return;

	vector(string) pct_splitted = str_tokenize(s, "%");
	integer n = v_size(pct_splitted);
	for (integer i = 1; i < n; ++i) {
		string hex_and_more = pct_splitted[i];
		string hex = sub_string(hex_and_more, 0, 2);
		integer ch = str_hex_to_integer(hex);
		string decoded_ch = str_build([ch]);
		pct_splitted[i] = strcat(decoded_ch, sub_string_end(hex_and_more, 2));
	}

	if (n > 1)
		s = strcat(pct_splitted);
}

void url_percent_decode(vector(string) v)
{
	integer n = v_size(v);
	for (integer i = 0; i < n; ++i) {
		try {
			url_percent_decode(v[i]);
		} catch {
			throw (E_UNSPECIFIC, strcat(["Invalid URL percent encoding: ", v[i]]));
		}
	}
}

class URL {
public:
	URL(string url);

	logical like(string compare_url);
	integer size();
	string at(integer pos);

	string path();

	vector(string) params();
	integer param_count(string param);
	string param_value(string param, string option(nullable) dflt = null<string>); // Exception if count > 1
	string param_value_req(string param); // Exception if count != 1
	vector(string) param_multi_value(string param); // Never exception, works for missing and single params too

	vector(string) unused_params();

private:
	string orig_;

	vector(string) parts_;

	map_istr_v_str params_;
	set_str unused_params_;

	void __dbg_print(__dbg_label l);
};

URL.URL(string url)
{
	if (!str_startswith(url, "/")) {
		throw (E_UNSPECIFIC, "Does not start with slash");
	}

	url = sub_string_end(url, 1);

	orig_ = url;

	vector(string) base_params = str_tokenize(url, "?");
	parts_ = str_tokenize(base_params[0], "/");
	url_percent_decode(parts_);

	params_ = new map_istr_v_str();
	unused_params_ = set_str();

	if (v_size(base_params) == 1) {
		// no params
	} else if (v_size(base_params) == 2) {
		vector(string) query_params = str_tokenize(base_params[1], "&");
		integer nq = v_size(query_params);
		for (integer i = 0; i < nq; ++i) {
			string qp = query_params[i];
			string key;
			string val;
			integer equal_pos = str_search(qp, "=");
			if (equal_pos < 0) {
				key = qp;
			} else {
				key = sub_string_start(qp, equal_pos);
				val = sub_string_end(qp, equal_pos+1);
			}

			url_percent_decode(key);
			url_percent_decode(val);

			unused_params_.add(key);
			vector(string) param_vals = params_.emplace(key);
			push_back(param_vals, val);
		}
	} else {
		throw (E_UNSPECIFIC, "Multiple question marks in URL");
	}
}

logical URL.like(string compare_url)
{
	if (!str_startswith(compare_url, "/")) {
		throw (E_UNSPECIFIC, "Compare URL does not start with slash");
	}

	compare_url = sub_string_end(compare_url, 1);

	vector(string) compare_parts = str_tokenize(compare_url, "/");

	if (v_size(compare_parts) != v_size(parts_)) {
		return false;
	}

	integer n = v_size(parts_);
	for (integer i = 0; i < n; ++i) {
		if (compare_parts[i] == "*") {
			continue;
		}

		string part = parts_[i];
		vector(string) vcmp = str_tokenize(compare_parts[i], "|");
		integer m = v_size(vcmp);
		integer j = 0;
		for ( ; j < m; ++j) {
			if (part == vcmp[j])
				break;
		}

		if (j == m)
			return false; // didn't find part in vcmp
	}

	return true;
}

integer URL.size() = v_size(parts_);

string URL.at(integer pos) = parts_[pos];

string URL.path() = strcat(["/", str_join(parts_, "/")]);

vector(string) URL.params() = params_.get_keys();

integer URL.param_count(string param)
{
	unused_params_.remove(param);
	return v_size(params_.find(param));
}

string URL.param_value(string param, string option(nullable) dflt)
{
	unused_params_.remove(param);
	vector(string) vals = params_.find(param);
	if (v_size(vals) == 0) {
		return dflt;
	} else if (v_size(vals) == 1) {
		return vals[0];
	} else {
		throw (E_UNSPECIFIC, strcat(["Param '", param, "' has multiple values"]));
	}
}

string URL.param_value_req(string param)
{
	if (null(params_.find(param)))
		throw (E_UNSPECIFIC, strcat(["Param '", param, "' expected in URL"]));
	return this.param_value(param);
}

vector(string) URL.param_multi_value(string param)
{
	unused_params_.remove(param);
	vector(string) vals = params_.find(param);
	if (null(vals)) {
		return vector(i:0; "");
	} else {
		return vals;
	}
}

vector(string) URL.unused_params() = unused_params_.get_content();

void URL.__dbg_print(__dbg_label l)
{
	vector(string) param_strs[0];
	vector(string) params = this.params();
	integer n = v_size(params);
	for (integer i = 0; i < n; ++i) {
		vector(string) vals = params_.find(params[i]);
		if (null(vals))
			resize(vals, 0);
		if (v_size(vals) == 1) {
			push_back(param_strs, strcat([params[i], "=", vals[0]]));
		} else {
			push_back(param_strs, strcat([params[i], "=[", str_join(vals, ","), "]"]));
		}
	}

	l.set_text(strcat(["URL { /", str_join(parts_, "/"), " { ", str_join(param_strs, ", "), " } }"]));
}

http.response unused_params_response(URL url)
{
	if (v_size(url.unused_params()) > 0) {
		return http_api.mk_error_response(400, strcat(["Unknown query parameters: ", str_join(url.unused_params(), ", ")]));
	} else {
		return null<http.response>;
	}
}


/*
 * context is just a base class. A server will have a subclass of context.
 *
 */

class context {
};


/*
 * client_info holds information about the connected client, such as user
 * identity, connection information and permissions.
 */

class client_info {
public:
	client_info(http.server_conn conn);
	client_info(http.server_conn_ssl conn);

	string addr();
	integer port();
	string name();

private:
	string addr_;
	integer port_;
	string name_;

	void _init(http.server_conn conn);
};

client_info.client_info(http.server_conn conn)
{
	_init(conn);
}

client_info.client_info(http.server_conn_ssl conn)
{
	_init(conn);
}

void client_info._init(http.server_conn conn)
{
	conn.get_peer(addr_, port_, name_);
}

string client_info.addr() = addr_;
integer client_info.port() = port_;
string client_info.name() = name_;

/*
 * url_path is a URL where some parts are fixed and some parts are to be filled
 * to access correct data. Example path /users/{user}/accounts/{accountName} is
 * will match all four part URLs where part one is "users" and part three is
 * "accounts". Parts two and four are called 'path params' and will often be
 * found in maps with the keys being "user" and "accountName" respectively.
 *
 * The url_path class also makes it easy to test API calls by just passing in
 * the path params.
 */

class url_path {
public:
	url_path(string path);

	string path();
	string pattern(); // "/items/{item}/properties/{property}  =>  "/items/*/properties/*"
	vector(string) url_parts();  // "/items/{item}/properties/{property}"  =>  ["items", "*", "properties", "*"];
	vector(string) path_params();
	string desc_str();
	string mk_url(map_str_str path_param_values);
	string get_path_param(URL url, string element);

private:
	string path_;
	string pattern_;
	vector(string) url_parts_;
	vector(string) path_params_;
	vector(integer) path_param_positions_;

	void __dbg_print(__dbg_label l);
};

void url_path.__dbg_print(__dbg_label l)
{
	l.set_text(strcat(["url_path { ", path_, " }"]));
}

url_path.url_path(string path)
: path_(path)
{
	if (!str_startswith(path, "/"))
		throw (E_UNSPECIFIC, "URL path must start with /");
	if (str_endswith(path, "/"))
		throw (E_UNSPECIFIC, "URL path may not end with /");

	set_str params_seen = set_str();
	resize(path_params_, 0);
	resize(path_param_positions_, 0);

	// modified by loop below
	url_parts_ = str_tokenize(path_, "/")[1:];

	integer n = v_size(url_parts_);
	for (integer i = 0; i < n; ++i) {
		string part = url_parts_[i];
		if (part == "")
			throw (E_UNSPECIFIC, "URL path may not contain \"//\"");
		if (part == "*")
			throw (E_UNSPECIFIC, "URL path may not contain \"/*/\"");
		if (str_startswith(part, "{") && str_endswith(part, "}")) {
			string path_param = sub_string_start(sub_string_end(part, 1), -1);
			if (path_param == "")
				throw (E_UNSPECIFIC, "Path param may not be empty string");
			if (params_seen.includes(path_param))
				throw (E_UNSPECIFIC, strcat("Duplicate path param: ", path_param));
			params_seen.add(path_param);
			push_back(path_params_, path_param);
			push_back(path_param_positions_, i);
			url_parts_[i] = "*";
		}
	}

	pattern_ = strcat("/", str_join(url_parts_, "/"));
}

string url_path.path() = path_;

string url_path.pattern() = pattern_;

vector(string) url_path.url_parts() = url_parts_;

vector(string) url_path.path_params() = path_params_;

string url_path.desc_str() = path_;

string url_path.mk_url(map_str_str path_param_values)
{
	if (v_size(path_params_) == 0)
		return path_;

	vector(string) url_parts = str_tokenize(path_, "/");
	integer n = v_size(url_parts);
	integer path_param_ix = 0;
	for (integer i = 0; i < n; ++i) {
		if (path_param_positions_[path_param_ix] == i) {
			string path_param_key = path_params_[path_param_ix];
			string path_param_val = path_param_values.find(path_param_key);
			if (null(path_param_val)) {
				throw (E_UNSPECIFIC, strcat(["url_path '", this.desc_str(), "' requires query element '", path_param_key, "'"]));
			}
			url_parts[i] = path_param_val;
		}
	}

	return str_join(url_parts, "/");
}

string url_path.get_path_param(URL url, string element)
{
	integer n = v_size(path_params_);
	for (integer i = 0; i < n; ++i) {
		if (path_params_[i] == element) {
			return url.at(path_param_positions_[i]);
		}
	}
	throw (E_UNSPECIFIC, strcat(["No path param '", element, "' in path ", path_]));
}


class request_values {
public:
	request_values();

	string get_path_param(string name);
	string get_query_param(string name);
	vector(string) get_query_multi_param(string name);
	json.node get_body();

	void set_path_param(string name, string value);
	void set_query_param(string name, string option(nullable) value);
	void set_query_multi_param(string name, vector(string) value);
	void set_body(json.node option(nullable) body);

	string params_json();

private:
	map_str_str path_params_;
	map_str_str query_params_;
	map_str_v_str query_multi_params_;
	json.node body_;
};

request_values.request_values()
: path_params_(map_str_str()), query_params_(map_str_str()), query_multi_params_(new map_str_v_str())
{
}

string request_values.get_path_param(string name){
	string value = path_params_.find(name);
	if (null(value))
		throw (E_UNSPECIFIC, strcat(["Internal: No such path param '", name, "'"]));
	return value;
}

string request_values.get_query_param(string name) = query_params_.find(name);

vector(string) request_values.get_query_multi_param(string name) = query_multi_params_.find(name);

json.node request_values.get_body() = body_;

void request_values.set_path_param(string name, string value) { path_params_.add(name, value); }

void request_values.set_query_param(string name, string option(nullable) value)
{
	if (null(value))
		query_params_.remove(name);
	else
		query_params_.add(name, value);
}

void request_values.set_query_multi_param(string name, vector(string) value) { query_multi_params_.add(name, value); }

void request_values.set_body(json.node option(nullable) body) { body_ = body; }

string request_values.params_json()
{
	json.builder jb = new json.builder();
	jb.begin_object();
	{
		jb.begin_member("path_params");
		jb.begin_object();
		vector(string) name, value;
		path_params_.get_content(name, value);
		sort(value, name);
		sort(name);
		for (integer i = 0; i < v_size(name); ++i) {
			jb.begin_member(name[i]);
			jb.append(value[i]);
		}
		jb.end_object();
	}
	{
		jb.begin_member("query_params");
		jb.begin_object();
		vector(string) name, value;
		query_params_.get_content(name, value);
		sort(value, name);
		sort(name);
		for (integer i = 0; i < v_size(name); ++i) {
			jb.begin_member(name[i]);
			jb.append(value[i]);
		}
		jb.end_object();
	}
	{
		jb.begin_member("query_multi_params");
		jb.begin_object();
		vector(string) name = query_multi_params_.get_keys();
		for (integer i = 0; i < v_size(name); ++i) {
			vector(string) values = query_multi_params_.find(name[i]);
			jb.begin_member(name[i]);
			jb.append(values);
		}
		jb.end_object();
	}
	jb.end_object();
	return jb.get();
}

// We wrap http.response in our own class, to future-proof stuff. In the future we
// may want to allow returning more that just a http.response from handler functions.
class response {
public:
	response(http.response resp);

	http.response get();

private:
	http.response resp_;
};

response.response(http.response resp)
	: resp_(resp)
{
}

http.response response.get() = resp_;



typedef
response function(
	context ctxt,
	client_info client,
	request_values req_vals)
	handle_func_t;




class request_handler {
public:
	request_handler(
		http.method method,
		string url,
		handle_func_t handle_func,
		string summary);

	void set_description(string desc);

	void check_complete();

	void set_path_param_desc(string param, open_api.schema sch, string desc);

	void add_query_param(string name, string option(nullable) dflt);
	void add_query_multi_param(string name);

	void set_body(logical required, open_api.schema option(nullable) schema);

	void add_response(open_api.response_desc);

	// Core stuff
	http.method method();
	url_path path();
	handle_func_t func();
	string summary();
	string desc();

	// Path params
	string path_param_desc(string param);
	open_api.schema path_param_schema(string param);

	// Query params
	vector(string) query_params();
	logical query_param_is_multi(string param_name);
	open_api.schema query_param_schema(string param_name);
	string query_param_default(string param_name);

	// Request body
	logical takes_body();
	logical requires_body();
	open_api.schema request_schema();

	// Responses
	vector(open_api.response_desc) responses();

private:
	// Core stuff
	http.method method_;
	url_path path_;
	handle_func_t func_;
	string summary_;
	string desc_;
	map_str_str path_param_descs_;
	map_str_obj<open_api.schema> path_param_schema_;

	// Query params
	set_str query_params_;
	map_str_obj<open_api.schema> query_param_schema_;
	map_str_str query_param_default_;
	set_str query_multi_params_;

	// Request body
	logical takes_body_;
	logical requires_body_;
	open_api.schema request_schema_;

	// Responses
	vector(open_api.response_desc) responses_;

	void __dbg_print(__dbg_label l);
};

void request_handler.__dbg_print(__dbg_label l)
{
	l.set_text(strcat(["request_handler { ", enum_id(method_), " ", path_.path(), " }"]));
}


request_handler.request_handler(http.method method, string url, handle_func_t handle_func, string summary)
: method_(method), path_(new url_path(url)), func_(handle_func), summary_(summary)
	, path_param_descs_(map_str_str()), path_param_schema_(new map_str_obj<open_api.schema>())
	, query_params_(set_str()), query_param_schema_(new map_str_obj<open_api.schema>())
	, query_param_default_(map_str_str()), query_multi_params_(set_str())
	, takes_body_(false), requires_body_(false), request_schema_(null<open_api.schema>)
{
}

void request_handler.set_description(string desc)
{
	desc_ = desc;
}

void request_handler.check_complete()
{
	string err_prefix = strcat([enum_id(method_), " ", path_.path(), ": "]);
	if (v_size(responses_) == 0)
		throw (E_UNSPECIFIC, strcat(err_prefix, "No responses"));

	vector(string) path_params = path_.path_params();
	integer n = v_size(path_params);
	for (integer i = 0; i < n; ++i) {
		if (null(path_param_descs_.find(path_params[i]))) {
			throw (E_UNSPECIFIC, strcat([err_prefix, "Need desc for path param '", path_params[i], "'"]));
		}
	}
}

void request_handler.set_path_param_desc(string param, open_api.schema sch, string desc)
{
	vector(string) path_params = path_.path_params();
	integer n = v_size(path_params);
	for (integer i = 0; i < n; ++i) {
		if (path_params[i] == param) {
			path_param_descs_.add(param, desc);
			path_param_schema_.add(param, sch);
			return;
		}
	}
	throw (E_UNSPECIFIC, strcat(["No such path param: ", param]));
}

void request_handler.add_query_param(string name, string option(nullable) dflt)
{
	if (query_params_.includes(name))
		throw (E_UNSPECIFIC, strcat("Duplicate query param: ", name));
	query_params_.add(name);
	open_api.schema_builder schb = new open_api.schema_builder().set_type_string(dflt, dflt);
	query_param_schema_.add(name, schb.get_schema());
	if (!null(dflt))
		query_param_default_.add(name, dflt);
}

void request_handler.add_query_multi_param(string name)
{
	if (query_params_.includes(name))
		throw (E_UNSPECIFIC, strcat("Duplicate query param: ", name));
	query_params_.add(name);
	open_api.schema_builder schb = new open_api.schema_builder().set_type_string(null<string>, null<string>);
	query_multi_params_.add(name);
	query_param_schema_.add(name, schb.get_schema());
}

void request_handler.set_body(logical required, open_api.schema option(nullable) schema)
{
	switch (method_) {
	case http.CONNECT:
	case http.DELETE:
	case http.GET:
	case http.HEAD:
		throw (E_UNSPECIFIC, "Method does not allow body");
	}

	takes_body_ = true;
	requires_body_ = required;
	request_schema_ = schema;
}

void request_handler.add_response(open_api.response_desc resp)
{
	integer n = v_size(responses_);
	for (integer i = 0; i < n; ++i) {
		if (responses_[i].status_code == resp.status_code)
			throw (E_UNSPECIFIC, strcat("Duplicate response for status code: ", str(resp.status_code)));
	}
	push_back(responses_, resp);
}

http.method request_handler.method() = method_;
url_path request_handler.path() = path_;
handle_func_t request_handler.func() = func_;
string request_handler.summary() = summary_;
string request_handler.desc() = desc_;
string request_handler.path_param_desc(string path_param) = path_param_descs_.find(path_param);
open_api.schema request_handler.path_param_schema(string path_param) = path_param_schema_.find(path_param);

vector(string) request_handler.query_params() = query_params_.get_content();

logical request_handler.query_param_is_multi(string param_name)
{
	if (!query_params_.includes(param_name))
		throw (E_UNSPECIFIC, strcat("Unknown query param: ", param_name));
	return query_multi_params_.includes(param_name);
}

open_api.schema request_handler.query_param_schema(string param_name)
{
	if (!query_params_.includes(param_name))
		throw (E_UNSPECIFIC, strcat("Unknown query param: ", param_name));
	return query_param_schema_.find(param_name);
}

string request_handler.query_param_default(string param_name) = query_param_default_.find(param_name);

logical request_handler.takes_body() = takes_body_;
logical request_handler.requires_body() = requires_body_;
open_api.schema request_handler.request_schema() = request_schema_;

vector(open_api.response_desc) request_handler.responses() = responses_;


void add_path_params(
	request_values req_vals,
	request_handler handler,
	URL url)
{
	url_path path = handler.path();
	vector(string) path_params = path.path_params();
	integer n = v_size(path_params);
	for (integer i = 0; i < n; ++i) {
		string path_param = path_params[i];
		string value = path.get_path_param(url, path_param);
		req_vals.set_path_param(path_param, value);
	}
}

void add_query_params(
	request_values req_vals,
	request_handler handler,
	URL url)
{
	vector(string) accepted_params = handler.query_params();
	integer n = v_size(accepted_params);
	for (integer i = 0; i < n; ++i) {
		string param = accepted_params[i];
		if (handler.query_param_is_multi(param)) {
			req_vals.set_query_multi_param(param, url.param_multi_value(param));
		} else {
			string dflt = handler.query_param_default(param);
			req_vals.set_query_param(param, url.param_value(param, dflt));
		}
	}

	vector(string) unused_params = url.unused_params();
	if (v_size(unused_params) > 5)
		throw (E_UNSPECIFIC, "Several unexpected query params in URL");
	if (v_size(unused_params) > 0)
		throw (E_UNSPECIFIC, strcat(["Unexpected query params: ", str_join(unused_params, ", ")]));
}

void add_body(
	request_values req_vals,
	request_handler handler,
	blob option(nullable) body)
{
	if (!handler.takes_body()) {
		if (!null(body))
			throw (E_UNSPECIFIC, "Unexpected body");
	} else {
		if (null(body) && handler.requires_body()) {
			throw (E_UNSPECIFIC, "Body required");
		}

		try {
			json.node json_body = null(body) ? null<json.node> : json.parse(body.get_string());
			req_vals.set_body(json_body);
		} catch {
			throw (E_UNSPECIFIC, strcat("Malformed body: ", err.message()));
		}
	}
}

} // module http_api

module open_api {

void append_params(json.builder jb, http_api.request_handler h)
{
	vector(string) path_params = h.path().path_params();
	vector(string) query_params = h.query_params();
	if (v_size(path_params) > 0 || v_size(query_params) > 0) {
		jb.begin_member("parameters");
		jb.begin_array();
		integer npp = v_size(path_params);
		for (integer pp = 0; pp < npp; ++pp) {
			string path_param = path_params[pp];
			jb.begin_object();
			jb.begin_member("name");
			jb.append(path_param);
			string desc = h.path_param_desc(path_param);
			schema sch = h.path_param_schema(path_param);
			if (!null(desc)) {
				jb.begin_member("description");
				jb.append(desc);
			}
			if (!null(sch)) {
				jb.begin_member("schema");
				jb.append_node(sch.mk_json_node());
			}
			jb.begin_member("required");
			jb.append(true);
			jb.begin_member("in");
			jb.append("path");
			jb.end_object();
		}
		integer nqp = v_size(query_params);
		for (integer qp = 0; qp < nqp; ++qp) {
			string query_param = query_params[qp];
			schema sch = h.query_param_schema(query_param);
			jb.begin_object();
			jb.begin_member("name");
			jb.append(query_param);
			jb.begin_member("in");
			jb.append("query");
			if (!null(sch)) {
				jb.begin_member("schema");
				jb.append_node(sch.mk_json_node());
			}
			jb.end_object();
		}
		jb.end_array();
	}
}

void append_responses(json.builder jb, http_api.request_handler h)
{
	vector(response_desc) responses = h.responses();
	jb.begin_member("responses");
	jb.begin_object();
	responses.mk_json(jb);
	jb.end_object(); // responses
}

void append_method(json.builder jb, http_api.request_handler h)
{
	string method = str_to_lower(enum_id(h.method()));
	jb.begin_member(method);
	jb.begin_object();
	if (!null(h.desc())) {
		jb.begin_member("description");
		jb.append(h.desc());
	}

	append_params(jb, h);

	append_responses(jb, h);

	jb.end_object(); // method
}

} // module open_api

module http_api {

typedef map_str_obj<map_str_obj<request_handler>> handler_map_t;

class api {
public:
	api(string title, string version);

	void add_schema(string schema_name, open_api.schema schema);

	void add_handler(request_handler h);

	response process(http.request req, context ctxt, client_info client);

	string create_open_api();

private:
	string title_;
	string version_;
	handler_map_t pattern2method2handler_;
	map_str_str pattern2path_;

	map_str_obj<open_api.schema> schemas_;

	void _check_ambiguity(url_path path);
	map_str_obj<request_handler> _find_path_handlers(URL url);

	response _resp_if_open_api_req(http.method method, URL url);
	void _open_api_append_schemas(json.builder jb);
	void _open_api_append_paths(json.builder jb);
	string _create_open_api();

	void __dbg_print(__dbg_label l);
};

void api.__dbg_print(__dbg_label l)
{
	counter_str method_counts = new counter_str();
	vector(string) patterns = pattern2method2handler_.get_keys();
	integer n = v_size(patterns);
	for (integer i = 0; i < n; ++i) {
		vector(string) methods = pattern2method2handler_.find(patterns[i]).get_keys();
		method_counts.increase(methods);
	}

	vector(string) methods;
	vector(number) counts;
	method_counts.get_content(methods, counts);
	sort(counts, methods);
	sort(methods);
	vector(string) method_count_strs = v_strcat().cat(methods).cat(": ").cat(str(counts)).get();

	l.set_text(strcat(["api { ", str(pattern2path_.size()), " paths, ", str_join(method_count_strs, "; "), " }"]));
}

api.api(string title, string version)
: title_(title), version_(version)
	, pattern2method2handler_(new handler_map_t())
	, pattern2path_(map_str_str())
	, schemas_(new map_str_obj<open_api.schema>())
{
}

void api.add_schema(string schema_name, open_api.schema schema)
{
	schemas_.add(schema_name, schema);
}

void api.add_handler(request_handler h)
{
	h.check_complete();

	_check_ambiguity(h.path());

	string url_pattern = h.path().pattern();
	string method_str = enum_id(h.method());
	map_str_obj<request_handler> method2handler = pattern2method2handler_.find(url_pattern);

	if (null(method2handler)) {
		method2handler = new map_str_obj<request_handler>();
		pattern2method2handler_.add(url_pattern, method2handler);
		pattern2path_.add(url_pattern, h.path().path());
	} else {
		// Check that the path is not changed, i.e. we do not accept handlers
		// for url_path "/customer/{NAME}", and for url_path
		// "/customer/{ID}". The handlers must have the same path parameters.
		string existing_path = pattern2path_.find(url_pattern);
		if (h.path().path() != existing_path)
			throw (E_UNSPECIFIC, strcat(["URL path \"", h.path().path(), "\" collides with \"", existing_path, "\""]));

		// Check that handler is unique.
		// This will catch URL's that are similar like this:
		// "/items/{id}" and "/items/{source}",
		// but it will not catch URL's that are ambigous like this:
		// "/items/{id}" and "/{type}/list", since the placeholders
		// are at different positions in the URL.
		request_handler pre_existing = method2handler.find(method_str);
		if (!null(pre_existing))
			throw (E_UNSPECIFIC, strcat(["Already have handler for ", method_str, " ", url_pattern]));
	}

	method2handler.add(method_str, h);
}

map_str_obj<request_handler> api._find_path_handlers(URL url)
{
	vector(string) url_patterns = pattern2method2handler_.get_keys(false);
	integer n = v_size(url_patterns);
	for (integer i = 0; i < n; ++i) {
		string url_pattern = url_patterns[i];
		if (url.like(url_pattern))
			return pattern2method2handler_.find(url_pattern);
	}

	return null<map_str_obj<request_handler>>;
}

response api._resp_if_open_api_req(http.method method, URL url)
{
	if (method != http.GET)
		return null<response>;

	if (url.path() != open_api.get_url) {
		return null<response>;
	}

	string open_api_json = this.create_open_api();
	json.node jn = json.parse(open_api_json);
	return new response(mk_response(200, jn));
}

void api._check_ambiguity(url_path path)
{
	vector(string) path_elements = path.url_parts();
	integer n_elems = v_size(path_elements);

	vector(string) url_patterns = pattern2method2handler_.get_keys(false);
	integer n = v_size(url_patterns);
	for (integer i = 0; i < n; ++i) {
		vector(string) compare_path_elements = str_tokenize(url_patterns[i], "/")[1:];
		if (v_size(compare_path_elements) != n_elems)
			continue;

		if (url_patterns[i] == path.pattern())
			continue;

		// To be able to always differentiate two paths, they must either
		//   - not have the same number of elements (checked above)
		//   - have at least one element position where both paths have fixed string, and these are different
		// E.g. "/items/{item}" and "/{model}/info" are ambiguous for URL "/items/info",
		// but "/items/{item}/price and "/{model}/info/id" can always be distinguished on the third part.
		// However "items/{item}/id" and "/{model}/info/id" are ambiguous.
		logical found_not_ambiguous = false;
		for (integer k = 0; k < n_elems; ++k) {
			string elem = path_elements[k];
			string cmp_elem = compare_path_elements[k];
			if (elem != "*" && cmp_elem != "*" && elem != cmp_elem) {
				found_not_ambiguous = true;
				break;
			}
		}

		if (!found_not_ambiguous)
			throw (E_UNSPECIFIC, strcat(["URL paths \"", path.path(), "\" and \"", pattern2path_.find(url_patterns[i]), "\" are ambiguous"]));
	}
}

response api.process(http.request req, context ctxt, client_info client)
{
	URL req_url = new URL(req.url());

	{ // Request for OPEN API spec?
		response open_api_resp = _resp_if_open_api_req(req.method(), req_url);
		if (!null(open_api_resp))
			return open_api_resp;
	}

	map_str_obj<request_handler> method2handlers = _find_path_handlers(req_url);
	if (null(method2handlers))
		return new response(mk_response(404));

	request_handler handler = method2handlers.find(enum_id(req.method()));
	if (null(handler))
		return new response(mk_method_not_allowed_response(method2handlers.get_keys()));

	request_values req_vals = new request_values();
	try {
		add_path_params(req_vals, handler, req_url);
		add_query_params(req_vals, handler, req_url);
		add_body(req_vals, handler, req.body());
	} catch {
		return new response(mk_error_response(400, err.message()));
	}

	try {
		response resp = handler.func()(ctxt, client, req_vals);
		return resp;
	} catch {
		return new response(mk_error_response(500, err.message()));
	}
}

void api._open_api_append_schemas(json.builder jb)
{
	vector(string) schema_names;
	vector(open_api.schema) schemas;
	schemas_.get_content(schema_names, schemas);
	integer n = v_size(schemas);
	if (n == 0)
		return;

	sort(schemas, schema_names);
	sort(schema_names);

	jb.begin_member("components");
	jb.begin_object();
	jb.begin_member("schemas");
	jb.begin_object();
	for (integer i = 0; i < n; ++i) {
		jb.begin_member(schema_names[i]);
		jb.append_node(schemas[i].mk_json_node());
	}
	jb.end_object(); // schemas
	jb.end_object(); // components
}


void api._open_api_append_paths(json.builder jb)
{
	jb.begin_member("paths");
	jb.begin_object();

	jb.begin_member(open_api.get_url);
	jb.begin_object();
	jb.begin_member("get");
	jb.begin_object();
	jb.begin_member("description");
	jb.append("Get the json for the OPEN API specification");
	jb.begin_member("responses");
	jb.begin_object();
	jb.begin_member("200");
	jb.begin_object();
	jb.begin_member("description");
	jb.append(default_status_message(200));
	jb.end_object();
	jb.end_object();
	jb.end_object();
	jb.end_object();

	vector(string) url_patterns;
	vector(map_str_obj<request_handler>) method2handlers;
	pattern2method2handler_.get_content(url_patterns, method2handlers);
	sort(method2handlers, url_patterns);
	sort(url_patterns);
	integer n = v_size(url_patterns);
	for (integer i = 0; i < n; ++i) {
		string url_pattern = url_patterns[i];
		string path = pattern2path_.find(url_pattern);
		jb.begin_member(path);
		jb.begin_object();
		vector(string) methods = method2handlers[i].get_keys();
		vector(request_handler) handlers = method2handlers[i].find(methods);
		integer m = v_size(methods);
		for (integer j = 0; j < m; ++j) {
			request_handler h = handlers[j];
			open_api.append_method(jb, h);
		}
		jb.end_object(); // path
	}

	jb.end_object(); // paths
}

string api._create_open_api()
{
	json.builder jb = new json.builder();
	jb.begin_object();
	jb.begin_member("openapi");
	jb.append("3.0.1");

	jb.begin_member("info");
	jb.begin_object();
	jb.begin_member("title");
	jb.append(title_);
	jb.begin_member("version");
	jb.append(version_);
	jb.end_object(); // info

	_open_api_append_schemas(jb);

	_open_api_append_paths(jb);

	/* Appearently consumes and produces can't be in the root
	// api only supports json everywhere
	jb.begin_member("consumes");
	jb.begin_array();
	jb.append("application/json");
	jb.end_array();

	jb.begin_member("produces");
	jb.begin_array();
	jb.append("application/json");
	jb.end_array();
	*/

	jb.end_object(); // root

	return jb.get();
}

string api.create_open_api()
{
	try {
		return _create_open_api();
	} catch {
		// It is too dangerous to throw exception here. Ideally, impossible API
		// should have been caught earlier. Other devs cannot be expected to
		// put call to create_open_api in a try-catch clause, and the server
		// should not fail due to OPEN api json not being possible.
		return "{ \"error\": \"Could not create Open API json\" }";
	}
}

}
