/*	
	Utility funcs for instrument wrapper functions
	Developer: Algorithmica Research, Magnus Nyström			
*/


class enum asw_type
option (category: 'Instrument/AssetSwap')
{
	ASW_PAR_VALUE
	option(str: "Par")
	option(desc:
		   "Par value, "
		   "i.e. notional = nominal amount for the bond."),
	ASW_MARKET_CLEAN
	option(str: "CleanPrice")
	option(desc:
		   "Market value clean, "
		   "i.e. notional = clean price of the bond."),
	ASW_MARKET_DIRTY
	option(str: "DirtyPrice")
	option(desc:
		   "Market value dirty, "
		   "i.e. notional = dirty price of the bond."),
	ASW_NO_ADJ
	option(str: "NoAdjustment")
	option(desc:
		   "No adjustments, "
		   "i.e. notional = nominal amount for the bond "
		   "and all difference in value between the bond "
		   "and the swap is reflected in the spread.")
};

//-------------------------------------------------------
//  asw_type_list
//-------------------------------------------------------
vector(string)  asw_type_list()
option (category: 'Instrument/AssetSwap')
{
	vector(asw_type) e;
	expand_enum(e) ;
	return string(e);
}

//-------------------------------------------------------
//  asw_type_map
//-------------------------------------------------------
asw_type asw_type_map(string s)
option (category: 'Instrument/AssetSwap')
{
	asw_type e;
	str_to_enum(s,e);
	return e;
}

module CORE_INT
{
	
	//-------------------------------------------------------
	// asw_uf_be
	//-------------------------------------------------------
	void asw_uf_be(asw_type type, number cp100, number ai100, number nom_bond, 
				   out number prem_disc, out number accr_adj,
				   out number swap_not, out number upfront,
				   out number backend)
	{
		prem_disc 	= (cp100/100.0 - 1.0)*nom_bond;
		accr_adj 	= ai100/100.0*nom_bond;	

		/*if(!null(user_not_pcnt_)){		
			no_ 	= user_not_pcnt_/100 * no_bond_ * fx_;
			uf_ 	= (cp100_+ai100_)*fx_*bond_nom_scale - no_;
			be_ 	= no_ - no_bond_;
			pv_flt_ = pv_bond_cpn_*fx_ - uf_ - be_ * df_ms_;
		}
		else {*/
		switch (type) {
		case asw_type.ASW_PAR_VALUE:		
			swap_not 	= nom_bond;
			upfront 	= accr_adj + prem_disc;
			backend 	= 0;
			break;
		case asw_type.ASW_MARKET_CLEAN:		
			swap_not 	= cp100/100.0*nom_bond;
			upfront 	= accr_adj;
			backend 	= prem_disc;
			break;
		case asw_type.ASW_MARKET_DIRTY:
			swap_not 	= (cp100+ai100)/100.0*nom_bond;
			upfront 	= 0;
			backend 	= accr_adj + prem_disc;
			break;
		case asw_type.ASW_NO_ADJ:
			swap_not 	= nom_bond;
			upfront 	= 0;
			backend 	= 0;
			break;
		default:
			QL_FAIL("invalid asset swap type");
		}		
	}
}

//-------------------------------------------------------
// class asset_swap_helper
// fees (i.e. upfront and backend) are assumed to be added to the fixed leg
//-------------------------------------------------------
class asset_swap_helper
	
{
	public:
		asset_swap_helper (	asw_type,date,date,number ,number,number,disc_func,number	option(nullable),
							number	option(nullable) ,vector(date) ,vector(number) 	,vector(date) ,
							vector(number) ,vector(number) );

		number			accrued_adj();
		number			upfront();
		number			backend();
		number			pv_backend();
		number			pv_fees();
		number			notional_float();
		number			notional_fixed();
		number			premium_adj();
		number			pv_fixed_ex_fees();
		number			pv_float_ex_fees();//same as pv_fixed_ex_fees() if single currency
		number			pv_float();
		number			pv_float(number);
		number 			spread_bpval(number);	
		number			spread();
		number			spread_upfront();
		number			spread_backend();
		number			spread_swap();
		number			spread_accrued();
		number			spread_premium();
		
		vector(date) 	cash_flow_dates();
		vector(number) 	float_cash_flow_amounts();
		void 			cash_flows(out vector(date),out vector(number),out vector(number),out vector(date),out vector(number),out vector(number));		
		void			priced_instrument(number, out number) ;
	
	protected:

		asset_swap_helper (	asw_type,date,date,number ,number,number,number,disc_func,disc_func,number	option(nullable),
							number	option(nullable) ,vector(date) ,vector(number) 	,vector(date) ,
							vector(number) ,vector(number) );
		
		void init();
		logical 		sums(date,date,rate_type,vector(date),vector(number),vector(number),disc_func,number,out number,out number);
		number 			pv_bond_cpns(disc_func disc);
		void 			spread_calc(number pv_flt);
		void 			cf_calc(out vector(date) , out vector(number));
		logical 		is_fund_sprd() ;
		date			trade_;	
		//date 			start_;
		date 			settle_;
		date 			mat_;
		asw_type		type_;
		number 			cp100_;
		number			ai100_;
		disc_func		disc_flt_;
		disc_func		disc_fix_;
		number			fund_sprd_ ;
		number			inv_sprd_ ;
		number 			fx_; //fixleg_ccy = basecurr, ex. fix=EUR, flt=SEK, fx = EURSEK=10
		number 			fx_mat_; //fixleg_ccy = basecurr, ex. fix=EUR, flt=SEK, fx = EURSEK=10 
		vector(date) 	bond_pmt_d_in_;
		vector(number) 	bond_c100_in_;
		vector(date)	vd_in_;
		vector(number)	per_in_;
		vector(number)	vr_in_;
		vector(number)	cf_;
		vector(date) 	cf_d_;
		number			sum_rdfdt_;
		number			sum_dfdt_;
		number			sum_rdfdt_f_;
		number			sum_dfdt_f_;
		number			df_ms_;
		number			pv_bond_cpn_;
		number 			pv_flt_;
		number 			pv_flt_ex_fees_;
		number 			pv_fees_;
		number			aa_;
		number			be_;
		number			no_;
		number			no_bond_;
		number			pd_;
		number			sp_tot_;
		number			sp_uf_;
		number			sp_be_;
		number			sp_ss_;
		number			sp_aa_;
		number			sp_pd_;
		number			uf_;
		rate_type		rt_;
		logical 		is_ccy_;
		logical			init_ok_;		
};

//---------------------------------------
// asset_swap_helper  constructor
//---------------------------------------
asset_swap_helper.asset_swap_helper (	asw_type 		type,
										date 			trade,
										//date 			start,
										date 			settle,
										number 			cp100,
										number			ai100,
										number			bond_nominal,
										disc_func		disc,
										number	option(nullable) fund_sprd,
										number	option(nullable) inv_sprd,
										vector(date) 	bond_pmt_dates,
										vector(number) 	bond_cpns100,
										vector(date) 	flt_pmt_dates,
										vector(number) 	flt_accr_periods,
										vector(number) 	flt_rates)
										: trade_(trade),settle_(settle), type_(type),
										  cp100_(cp100),ai100_(ai100),
										  disc_flt_(disc), disc_fix_(disc),
										fund_sprd_(null(fund_sprd)?0:fund_sprd),
										inv_sprd_(null(inv_sprd)?0:inv_sprd),
										fx_(1),fx_mat_(1),
										bond_pmt_d_in_(clone_vector(bond_pmt_dates)),
										bond_c100_in_(clone_vector(bond_cpns100)),										
										vd_in_(clone_vector(flt_pmt_dates)),
										per_in_(clone_vector(flt_accr_periods)),
										vr_in_(clone_vector(flt_rates)),
										no_bond_(bond_nominal), rt_(RT_SIMPLE),is_ccy_(false),
										init_ok_(false)
{

	init();
}
//---------------------------------------
// asset_swap_helper  constructor cross curr
//---------------------------------------
asset_swap_helper.asset_swap_helper (	asw_type 		type,
										date 			trade,
										//date 			start,
										date 			settle,
										number 			cp100,
										number			ai100,
										number			bond_nominal,
										number 			fx,
										disc_func		disc_fix,
										disc_func		disc_flt,										
										number	option(nullable) fund_sprd,
										number	option(nullable) inv_sprd,
										vector(date) 	bond_pmt_dates,
										vector(number) 	bond_cpns100,
										vector(date) 	flt_pmt_dates,
										vector(number) 	flt_accr_periods,
										vector(number) 	flt_rates)
										: trade_(trade),settle_(settle), type_(type),
										  cp100_(cp100),ai100_(ai100),
										  disc_flt_(disc_flt), disc_fix_(disc_fix),
										fund_sprd_(null(fund_sprd)?0:fund_sprd),
										inv_sprd_(null(inv_sprd)?0:inv_sprd),
										fx_(fx),fx_mat_(null<number>),
										bond_pmt_d_in_(clone_vector(bond_pmt_dates)),
										bond_c100_in_(clone_vector(bond_cpns100)),										
										vd_in_(clone_vector(flt_pmt_dates)),
										per_in_(clone_vector(flt_accr_periods)),
										vr_in_(clone_vector(flt_rates)),
										no_bond_(bond_nominal), rt_(RT_SIMPLE),is_ccy_(true),
										init_ok_(false)
{

	init();
}
//---------------------------------------
// init
//---------------------------------------
void asset_swap_helper.init()
{
	QL_REQUIRE(settle_ >= trade_, "invalid settlement date");
	
	integer b_size = v_size(bond_pmt_d_in_);
	QL_REQUIRE(b_size > 0 && b_size == v_size(bond_c100_in_), "vector size error (bond)");
	QL_REQUIRE(v_size(v_find(bond_c100_in_, [null<number>])) == 0, "invalid cashflow in vector (contains null value)(bond)");
		
	integer f_size = v_size(vd_in_);
	QL_REQUIRE(f_size > 0 && f_size == v_size(per_in_), "vector size error (float)");
	QL_REQUIRE(f_size == v_size(vr_in_), "vector size error (float)");

	QL_REQUIRE(v_size(v_find(vr_in_, [null<number>])) == 0, "invalid rate in vector (contains null value)(float)");
	
	logical ok = sums(trade_,settle_,rt_,vd_in_,per_in_,vr_in_,disc_flt_,0, sum_rdfdt_, sum_dfdt_);
	QL_REQUIRE(ok, "error calculating asset swap sums");

	//if(!null(user_not_pcnt_)){
		//QL_REQUIRE(user_not_pcnt_ > 0 && user_not_pcnt_<= 1, "invalid user notional weight");
	//}
	QL_REQUIRE(no_bond_ > 0, "invalid nominal (must be > 0)");
	
	pv_bond_cpn_ = pv_bond_cpns(disc_fix_);
	QL_REQUIRE(!null(pv_bond_cpn_), "error calculating pv of bond cpns");

	mat_ 			= bond_pmt_d_in_[b_size-1];
	number  df_m	= disc_flt_.disc((mat_ - trade_) / 365);
	number  df_s 	= disc_flt_.disc((settle_ - trade_) / 365);					
	df_ms_ 			= df_m / df_s;
	QL_REQUIRE(!null(df_ms_), "error calculating disc_func values");

	if(is_ccy_) {
		number  _df_m	= disc_fix_.disc((mat_ - trade_) / 365);
		number  _df_s 	= disc_fix_.disc((settle_ - trade_) / 365);					
		number _df_ms 	= _df_m / _df_s;
		QL_REQUIRE(!null(_df_ms), "error calculating disc_func values");
		fx_mat_ = fx_ * _df_ms/df_ms_; 
	}

	CORE_INT.asw_uf_be(type_, cp100_, ai100_, no_bond_, pd_, aa_,no_, uf_,be_);

	pd_ *= fx_;
	aa_	*= fx_;
	no_ *= fx_;
	uf_	*= fx_;
	be_ *= fx_mat_;

	pv_flt_ 		= pv_bond_cpn_*fx_ - uf_ - be_ * df_ms_;
	pv_fees_ 		= uf_ + be_ * df_ms_;
	pv_flt_ex_fees_ = pv_bond_cpn_*fx_;

	
	/*pd_ 	= (cp100_/100.0 - 1.0)*fx_*no_bond_;
	aa_ 	= ai100_/100.0*fx_*no_bond_;	

	
	switch (type_) {
	case asw_type.ASW_PAR_VALUE:		
		no_ 	= fx_ * no_bond_;
		uf_ 	= aa_ + pd_;
		be_ 	= 0;
		break;
	case asw_type.ASW_MARKET_CLEAN:		
		no_ 	= cp100_/100.0*fx_*no_bond_;
		uf_ 	= aa_;
		be_ 	= pd_/fx_* fx_mat_;
		break;
	case asw_type.ASW_MARKET_DIRTY:
		no_ 	= (cp100_+ai100_)/100.0*fx_*no_bond_;
		uf_ 	= 0;
		be_ 	= (aa_ + pd_) / fx_ * fx_mat_;
		break;
	case asw_type.ASW_NO_ADJ:
		no_ 	= fx_ * no_bond_;
		uf_ 	= 0;
		be_ 	= 0;
		break;
	default:
		QL_FAIL("invalid asset swap type");
	}	*/	
	
	
	sum_rdfdt_f_ = sum_rdfdt_;
	sum_dfdt_f_ = sum_dfdt_;

	if((uf_ + be_ * df_ms_ > 0) && fund_sprd_ != 0){
		ok = sums(trade_,settle_,rt_,vd_in_,per_in_,vr_in_,disc_flt_,fund_sprd_, sum_rdfdt_f_, sum_dfdt_f_);
		QL_REQUIRE(ok, "error calculating asset swap sums (funding spread)");
	}
	else if((uf_ + be_ * df_ms_ < 0) && inv_sprd_ != 0){
		ok = sums(trade_,settle_,rt_,vd_in_,per_in_,vr_in_,disc_flt_,inv_sprd_, sum_rdfdt_f_, sum_dfdt_f_);
		QL_REQUIRE(ok, "error calculating asset swap sums (invest spread)");
	}
	
	spread_calc(pv_flt_);
	cf_calc(cf_d_, cf_);
		
	init_ok_ = true;
	return ;
}
//---------------------------------------
// cf_calc
//---------------------------------------
void asset_swap_helper.cf_calc(out vector(date) cf_d, out vector(number) cf)
{
	resize(cf,0);
	resize(cf_d,0);
	push_back(cf, 0);
	push_back(cf_d, settle_);
	
	for(integer i = 0; i < v_size(vd_in_); i++) {
		if(vd_in_[i] <= settle_)
			continue;
		push_back(cf, (vr_in_[i] + sp_tot_) * per_in_[i] * no_ );
		push_back(cf_d, vd_in_[i]);
	}
}
//---------------------------------------
// sums
//---------------------------------------
logical asset_swap_helper.sums(	date 			d_t,
								date 			d_s,
								rate_type		rt,
								vector(date) 	vdi,
								vector(number) 	peri,
								vector(number) 	vri,	
								disc_func		disc,
								number			disc_spread,
								out number 		sum_rdfdt,
								out number 		sum_dfdt)
{
	/*
	 if(vd_[0] < settle_)
		QL_REQUIRE(vd_[1] >= settle_, "invalid date in vector (float)");
	 */
	sum_rdfdt	= 0;
	sum_dfdt	= 0;
	number df_s = disc.disc((d_s - d_t) / 365);
	number dfdt, df_n;

	if(disc_spread == 0){
		for(integer i = 0; i < v_size(peri); i++) {
			if(vdi[i] <= d_s)
				continue;
			df_n = disc.disc((vdi[i] - d_t) / 365.0);
			if (null(df_n)) 
				return false;
						   
			dfdt		= df_n / df_s * peri[i];
			sum_rdfdt 	+= dfdt * (vri[i]);
			sum_dfdt 	+= dfdt;
		}
	}
	else {
		number r;
		number df_d = df_s;
		for(integer i = 0; i < v_size(peri); i++) {
			if(vdi[i] <= d_s)
				continue;
			df_n = disc.disc((vdi[i] - d_t) / 365.0);
			if (null(df_n)) 
				return false;
			r = disc_to_rate(df_n / df_d, rt, peri[i]) + disc_spread;
			df_d = df_n;
			df_n = rate_to_disc(r, rt, peri[i]) ;

			dfdt		= df_n / df_s * peri[i];
			sum_rdfdt 	+= dfdt * (vri[i]);
			sum_dfdt 	+= dfdt;
			
		}
	}
	return true;
}
//---------------------------------------
// pv_bond_cpns
//---------------------------------------
number asset_swap_helper.pv_bond_cpns(disc_func disc)
{
	number pv_cpn100 = 0;
	for(integer i = 0; i < v_size(bond_pmt_d_in_); i++) {
		if(bond_pmt_d_in_[i] <= settle_)
			continue;
		number df 	= disc.disc((bond_pmt_d_in_[i] - trade_) / 365.0);
		pv_cpn100	+= bond_c100_in_[i] * df;
	}
	//number pv_cpn100 	= disc.disc((bond_pmt_dates_-trade_)/365.0) * bond_cpns100_;
	number df_s 		= disc.disc((settle_ - trade_)/365.0) ;
	pv_cpn100 			/= df_s;
	return pv_cpn100/100.0*no_bond_;
	/*if(!null(user_not_pcnt_)){
		return pv_cpn100/100.0*user_not_pcnt_/100 * no_bond_;
	}
	else {
		return pv_cpn100/100.0*no_bond_;
	}*/
}

number asset_swap_helper.accrued_adj()		{ return init_ok_ ? aa_: null<number>; }
number asset_swap_helper.backend()			{ return init_ok_ ? be_: null<number>; }
number asset_swap_helper.pv_backend()		{ return init_ok_ ? be_*df_ms_: null<number>; }
number asset_swap_helper.pv_fees()			{ return init_ok_ ? pv_fees_: null<number>; }
number asset_swap_helper.notional_float()	{ return init_ok_ ? no_: null<number>; }
number asset_swap_helper.notional_fixed()	{ return init_ok_ ? no_bond_: null<number>; }
number asset_swap_helper.premium_adj()		{ return init_ok_ ? pd_: null<number>; }
number asset_swap_helper.upfront()			{ return init_ok_ ? uf_: null<number>; }
number asset_swap_helper.pv_fixed_ex_fees()	{ return init_ok_ ? pv_bond_cpn_: null<number>; }
number asset_swap_helper.pv_float_ex_fees()	{ return init_ok_ ? pv_flt_ex_fees_: null<number>; }
number asset_swap_helper.pv_float()			{ return init_ok_ ? pv_flt_: null<number>; }
number asset_swap_helper.spread()			{ return init_ok_ ? sp_tot_:null<number>; }		
number asset_swap_helper.spread_upfront()	{ return init_ok_ ? sp_uf_: null<number>; }
number asset_swap_helper.spread_backend()	{ return init_ok_ ? sp_be_: null<number>; }
number asset_swap_helper.spread_swap()		{ return init_ok_ ? sp_ss_: null<number>;}
number asset_swap_helper.spread_accrued()	{ return init_ok_ ? sp_aa_: null<number>; }
number asset_swap_helper.spread_premium()	{ return init_ok_ ? sp_pd_: null<number>; }	
logical asset_swap_helper.is_fund_sprd() 	{ return ((uf_ + be_ > 0) && fund_sprd_ != 0) || ((uf_ + be_ < 0) && inv_sprd_ != 0);}	
//---------------------------------------
// spread_bpval
//---------------------------------------
number asset_swap_helper.spread_bpval(number nominal)			
{ 
	if(init_ok_){
		number scale = nominal / 100.0;
		return abs(pv_flt_ - pv_float(sp_tot_+ 0.0001)) * scale;
	}
	else {
		return null<number>;
	}
}
//---------------------------------------
// pv_float
//---------------------------------------
number asset_swap_helper.pv_float(number spread)			
{ 
	if(!init_ok_)
		return null<number>;
	
	number pv = 0;
	number df_s = disc_flt_.disc((settle_ - trade_) / 365.0);

	for(integer i = 0; i < v_size(vd_in_); i++) {
		if(vd_in_[i] <= settle_)
			continue;
		number df_n = disc_flt_.disc((vd_in_[i] - trade_) / 365.0);
		if (null(df_n))
			return null<number>;
		number cf = (vr_in_[i] + spread) * per_in_[i] * no_;
		pv += cf * df_n;
	}
	pv /= df_s;
	return pv;
	
}
//---------------------------------------
// spread_calc
//---------------------------------------
void asset_swap_helper.spread_calc(number pv_flt)
{
	sp_uf_ 				= 0;
	number sp_uf_f_diff = 0;
	sp_be_ 				= 0;
	number sp_be_f_diff = 0;		
	number base 		= sum_rdfdt_ * no_;

	if(uf_ != 0){
		sp_uf_ = ((-uf_ + base) / no_ - sum_rdfdt_) / sum_dfdt_;	
		if(is_fund_sprd()){ 
			number sp_uf_f = ((-uf_ + base) / no_ - sum_rdfdt_f_) / sum_dfdt_f_;
			sp_uf_f_diff = sp_uf_f - sp_uf_;
			sp_uf_ = sp_uf_f;
		}																					
	}

	if(be_ != 0){
		sp_be_ = ((-be_ * df_ms_ + base) / no_ - sum_rdfdt_) / sum_dfdt_;
		if(is_fund_sprd()){ 
			number sp_be_f = ((-be_ * df_ms_ + base) / no_ - sum_rdfdt_f_) / sum_dfdt_f_;
			sp_be_f_diff = sp_be_f - sp_be_;
			sp_be_ = sp_be_f;
		}
	}
	sp_aa_ = ((-aa_ + base) / no_ - sum_rdfdt_) / sum_dfdt_;
	sp_pd_ = ((-pd_ + base) / no_ - sum_rdfdt_) / sum_dfdt_;
	number pv_bond_cpn_fx = pv_bond_cpn_*fx_;
	sp_ss_ =(pv_bond_cpn_fx / no_ - sum_rdfdt_) / sum_dfdt_;
	sp_tot_ = (pv_flt / no_ - sum_rdfdt_) / sum_dfdt_;

	//if(sp_uf_f_diff != 0)
		sp_tot_ += (sp_uf_f_diff + sp_be_f_diff);
							
	return ;
}

//---------------------------------------
// priced_instrument
//---------------------------------------
void asset_swap_helper.priced_instrument(number spread, out number dp100) 
{
	if(!init_ok_) {
		dp100 = null<number>;
		return;
	}
	
	number A = sum_rdfdt_ + sum_dfdt_ * spread;	

	number pv_bond_cpn_fx = pv_bond_cpn_ * fx_;
	
	switch (type_) {
	case asw_type.ASW_PAR_VALUE:{	
						
		if(is_fund_sprd()){
			dp100 = pv_bond_cpn_fx + (1 - A) * 100;			
			number uf = dp100 - 100*fx_;
			number base = sum_rdfdt_ * no_;			
			number sp_uf = ((-uf + base) / no_ - sum_rdfdt_) / sum_dfdt_;	
			number sp_uf_f = ((-uf + base) / no_ - sum_rdfdt_f_) / sum_dfdt_f_;			
			A = sum_rdfdt_ + sum_dfdt_ * (spread - sp_uf_f + sp_uf);
			dp100 = pv_bond_cpn_fx + (1 - A) * 100;
		}
		else
			dp100 = pv_bond_cpn_fx + (1 - A) * 100;
		break;
	}
	case asw_type.ASW_MARKET_CLEAN:{								
		if(is_fund_sprd()){			
			number base = sum_rdfdt_ * no_;				
			number sp_aa = ((-aa_ + base) / no_ - sum_rdfdt_) / sum_dfdt_;	
			number sp_aa_f = ((-aa_ + base) / no_ - sum_rdfdt_f_) / sum_dfdt_f_;				
			A = sum_rdfdt_ + sum_dfdt_ * (spread - sp_aa_f + sp_aa);
			dp100 = (pv_bond_cpn_fx + aa_ * (A + df_ms_ - 1) + df_ms_ * 100) / (A + df_ms_);

			number bed = (dp100 - aa_ - 100) * df_ms_;	
			number sp_bed = ((-bed + base) / no_ - sum_rdfdt_) / sum_dfdt_;	
			number sp_bed_f = ((-bed + base) / no_ - sum_rdfdt_f_) / sum_dfdt_f_;
			A = sum_rdfdt_ + sum_dfdt_ * (spread - sp_aa_f + sp_aa_ - sp_bed_f + sp_bed);
			dp100 = (pv_bond_cpn_fx + aa_ * (A + df_ms_ - 1) + df_ms_ * 100) / (A + df_ms_);
		}
		else	
			dp100 = (pv_bond_cpn_fx + aa_ * (A + df_ms_ - 1) + df_ms_ * 100) / (A + df_ms_);

		break;
	}
	case asw_type.ASW_MARKET_DIRTY:{
		if(is_fund_sprd()){
			dp100 = (pv_bond_cpn_fx + df_ms_ * 100) / (A + df_ms_);			
			number base = sum_rdfdt_ * no_;
			number bed = (dp100 -100) * df_ms_;
			number sp_bed = ((-bed + base) / no_ - sum_rdfdt_) / sum_dfdt_;	
			number sp_bed_f = ((-bed + base) / no_ - sum_rdfdt_f_) / sum_dfdt_f_;				
			A = sum_rdfdt_ + sum_dfdt_ * (spread - sp_bed_f + sp_bed);
			dp100 = (pv_bond_cpn_fx + df_ms_ * 100) / (A + df_ms_);
		}
		else
			dp100 = (pv_bond_cpn_fx + df_ms_ * 100) / (A + df_ms_);
		break;
	}
	case asw_type.ASW_NO_ADJ:
		dp100 = null<number>;
	}

	dp100 /= fx_;
	
	return ;
}

//---------------------------------------
// cash_flows and amounts
//---------------------------------------
vector(date) asset_swap_helper.cash_flow_dates() 			{ return init_ok_ ? cf_d_: null<vector(date)>;}
vector(number) asset_swap_helper.float_cash_flow_amounts() 	{ return init_ok_ ? cf_: null<vector(number)>; }

void asset_swap_helper.cash_flows(	out vector(date) 	bd,
									out vector(number) 	bcf, 
									out vector(number) 	fixcf,
									out vector(date) 	fltd,									 
									out vector(number) 	fltcf,
									out vector(number) 	fee) 
{ 
	if(init_ok_){
		
		//vector(date) bd_tmp 	= bond_pmt_dates_;
		//vector(number) bcf_tmp 	= bond_cpns100_;
		//integer s 				= v_size(bond_pmt_dates_);
		
		resize(bd,0);
		resize(bcf,0);
		resize(fixcf,0);
	
		push_back(bd, settle_);
		push_back(bcf,cp100_+ai100_);
		push_back(fixcf,0);

		for(integer i = 0; i < v_size(bond_pmt_d_in_); i++) {
			if(bond_pmt_d_in_[i] <= settle_)
				continue;
			push_back(bd,bond_pmt_d_in_[i]);
			push_back(bcf,bond_c100_in_[i]);
			push_back(fixcf,-bond_c100_in_[i]); 					 
		}
		
		fltcf = cf_;
		resize(fltd,0);
		for(integer i = 0; i < v_size(vd_in_); i++) {
			if(vd_in_[i] <= settle_)
				continue;
			push_back(fltd,vd_in_[i]);
		}

		integer s = v_size(cf_);
		resize(fee,s);
		fee = 0;
		fee[0] 	= uf_;
		fee[s-1]= be_;

	}
	else {
		resize(fltd,0);
		resize(fltcf,0);
		resize(fee,0);
		resize(fixcf,0);	
		resize(bcf,0);
		resize(bd,0);
	}
}


//-------------------------------------------------------
// asw_cashflows
//-------------------------------------------------------
/*void swap_fixflt_asw.asw_cashflows(	out number flt_nom,
									out number upfront,
									out number backend,
									out number ai,
									out number pd,
									out number cp)										
{
	

	error_info ee = error_info(true,true);
	number ai100 		= b_.accrued(false,true,ee);
	CORE_INT.instr_fail_check(null(ai100), "invalid accrued",b_.name(),ee);
				
	number cp100 		= b_.clean_price(false,true,ee);
	CORE_INT.instr_fail_check(null(cp100), "invalid clean price",b_.name(),ee);

	ai  = ai100/100 * bond_nominal_;
	cp 	= cp100/100 * bond_nominal_ ;	
	pd	= cp - bond_nominal_;
	
	if(null(user_not_pcnt_)){
		switch (asw_type_) {
		case ASW_PAR_VALUE:
			flt_nom 	= bond_nominal_;
			upfront 	= ai + pd;		//dp-flt_nom
			backend 	= 0;			//flt_nom - bond_nominal_
			break;
		case ASW_MARKET_CLEAN:
			flt_nom 	= cp;
			upfront 	= ai;			//dp-flt_nom
			backend 	= pd;			//flt_nom - bond_nominal_
			break;
		case ASW_MARKET_DIRTY:
			flt_nom 	= cp + ai;
			upfront 	= 0;			//dp-flt_nom
			backend 	= ai + pd;		//flt_nom - bond_nominal_
			break;
		case ASW_NO_ADJ:
			flt_nom 	= bond_nominal_;		    
			upfront 	= 0;
			backend 	= 0;
			break;
		}
	}
	else {		
		flt_nom = user_not_pcnt_/100 * bond_nominal_;
		upfront = ai + cp - flt_nom;						//dp-flt_nom
		backend 	= flt_nom - bond_nominal_;
	}
	
}*/

//-------------------------------------------------------
// class asset_swap_helper 
//-------------------------------------------------------
class asset_swap_helper_ccy : public asset_swap_helper 
{
	public:
		void backend(out number, out number);
		void upfront(out number, out number);
		void accrued_adj(out number, out number);
		void notional_fixed(out number, out number);
		void notional_float(out number, out number);
		void premium_adj(out number, out number);
		void pv_fixed_ex_fees(out number, out number);
		void pv_float(out number, out number);
	
		asset_swap_helper_ccy(asw_type,date,date,number ,number,number,number,disc_func,disc_func,number	option(nullable),
							number	option(nullable) ,vector(date) ,vector(number) 	,vector(date) ,
							vector(number) ,vector(number) );
	protected:
};


//---------------------------------------
// asset_swap_helper  constructor
//---------------------------------------
asset_swap_helper_ccy.asset_swap_helper_ccy (	asw_type 		type,
												date 			trade,
												//date 			start,
												date 			settle,
												number 			cp100,
												number			ai100,
												number			bond_nominal,
												number 			fx,
												disc_func		disc_fix,
												disc_func		disc_flt,
												number	option(nullable) fund_sprd,
												number	option(nullable) inv_sprd,
												vector(date) 	bond_pmt_dates,
												vector(number) 	bond_cpns100,
												vector(date) 	flt_pmt_dates,
												vector(number) 	flt_accr_periods,
												vector(number) 	flt_rates)
												:asset_swap_helper(type,trade,settle,cp100,ai100,bond_nominal,fx,disc_fix,disc_flt,fund_sprd,inv_sprd,
																   bond_pmt_dates,bond_cpns100,flt_pmt_dates,flt_accr_periods,flt_rates)
{
	
}

void asset_swap_helper_ccy.backend(out number backend_fix_ccy, out number backend_flt_ccy)
{
	backend_flt_ccy = asset_swap_helper.backend();
	backend_fix_ccy = backend_flt_ccy / fx_mat_;
}

void asset_swap_helper_ccy.upfront(out number upfront_fix_ccy, out number upfront_flt_ccy)
{
	upfront_flt_ccy = asset_swap_helper.upfront();
	upfront_fix_ccy = upfront_flt_ccy / fx_;
}

void asset_swap_helper_ccy.accrued_adj(out number accrued_adj_fix_ccy, out number accrued_adj_flt_ccy)
{
	accrued_adj_flt_ccy = asset_swap_helper.accrued_adj();
	accrued_adj_fix_ccy = accrued_adj_flt_ccy / fx_;
}
void asset_swap_helper_ccy.notional_float(out number notional_fix_ccy, out number notional_flt_ccy)
{
	notional_flt_ccy = asset_swap_helper.notional_float();
	notional_fix_ccy = notional_flt_ccy / fx_;
}
void asset_swap_helper_ccy.notional_fixed(out number notional_fix_ccy, out number notional_flt_ccy)
{
	notional_fix_ccy = asset_swap_helper.notional_fixed();
	notional_fix_ccy = notional_fix_ccy * fx_;
}
void asset_swap_helper_ccy.premium_adj(out number premium_adj_fix_ccy, out number premium_adj_flt_ccy)
{
	premium_adj_flt_ccy = asset_swap_helper.premium_adj();
	premium_adj_fix_ccy = premium_adj_flt_ccy / fx_;
}

void asset_swap_helper_ccy.pv_fixed_ex_fees(out number pv_fixed_fix_ccy, out number pv_fixed_flt_ccy)
{
	pv_fixed_fix_ccy = asset_swap_helper.pv_fixed_ex_fees();
	pv_fixed_flt_ccy = pv_fixed_fix_ccy * fx_;
}	
void asset_swap_helper_ccy.pv_float(out number pv_float_fix_ccy, out number pv_float_flt_ccy)
{
	pv_float_flt_ccy = asset_swap_helper.pv_float();
	pv_float_fix_ccy = pv_float_flt_ccy / fx_;
}

module CORE_INT
{
	//---------------------------------------------------------
	//fixflt_asw_helper_spr
	//version1: ql_swap_float_def and tenor_surface is created inside function
	//----------------------------------------------------------
	number fixflt_asw_mat_helper_spr(date 					td,
									date 					sd,
									date 					bond_mat_adj,
									day_count_method 		bond_dc,
									bd_convention 			bond_bd,
									integer					bond_freq,
									logical 				bond_eom,
									..calendar 				bond_calendar,
									number 					bond_roll_day,
									string					bond_currency,
									number 					bond_y,
									day_count_method 		flt_dc,
									bd_convention 			flt_bd,
									integer 				flt_freq,//assume reset_freq == cpn_freq
									logical 				flt_eom,
									..calendar 				flt_calendar,
									 string					flt_currency,
									..ir_index 				flt_ir_index,																
									disc_func 				disc_func,	
									fwd_func  				flt_fwd_func,
									number option(nullable) curr_fixing )
	{
		date start_date 			= sd;
		date_code settle_code 		= date_code("BD2");//not used in calculations so this dflt is ok
		date  maturity 				= bond_mat_adj;
		
		date flt_eff_date 			= start_date;		
		..coupon_freq  flt_cfreq 	= coupon_freq_from_int(flt_freq)  ;
		tenor_surface flt_ts 		= tenor_surface(td, disc_func,[flt_fwd_func],[flt_ir_index],false);
		//float leg
		notional_exchg_style ne 	= notional_exchg_style.NE_NONE;
		ql_swap_float_def flt_def 	= ql_swap_float_def("float_def",flt_currency,settle_code,flt_cfreq,
														null<reset_freq>,flt_dc,flt_bd,flt_calendar,flt_eom,
														null<flt_comp_avg_type>, null<flt_sprd_comp_method>,null<flt_avg_method>,
														flt_ir_index,flt_reset_style.RS_NORM,ne);
		QL_FAIL_COND(null(flt_def),"invalid float leg (def)");
		
		date  flt_first_cpn_date 	= null<date>;
		date  flt_last_reg_date 	= null<date>;				
		number flt_nominal 			= 100;
		number flt_sprd 			= 0;
		number fx_rate				= 1;
		ql_float_leg flt  			= ql_float_leg(	flt_def,"float_leg",td,sd,start_date,maturity,null<tenor_code>,
													false,flt_nominal,flt_eff_date,flt_first_cpn_date,flt_last_reg_date,flt_sprd,disc_func,	
													flt_ts,bond_roll_day,fx_rate,null<vector(date)>,null<vector(number)>,curr_fixing,false);

		coupon_freq  fix_cpn_freq 	= coupon_freq_from_int(bond_freq)  ;
		ql_swap_fix_def fix_def 	= ql_swap_fix_def(	"fix_def", bond_currency, settle_code, fix_cpn_freq,bond_dc,
														bond_bd,bond_calendar,bond_eom,ne);

		date  fix_first_cpn_date 	= null<date>;//b.next_cpn_date(ee);
		date  fix_last_reg_date 	= null<date>;//bond_last_reg_adj;//b.last_reg_cpn_date(ee);	
		date  fix_eff_date 			= start_date;
		number fix_nominal 			= 100;
		number fix_rate				= 0;//will not be used
		ql_fix_leg fix 				= ql_fix_leg(	fix_def,"fix_leg",td,sd,start_date,maturity,null<tenor_code>,
													true,fix_nominal ,fix_eff_date,fix_first_cpn_date,
													fix_last_reg_date,fix_rate,disc_func,bond_roll_day,fx_rate);

		ql_fixed_income_swap sw 	= ql_fixed_income_swap(fix,flt,"temp");

		
		number r = sw.solver(0,SS_FIX_RATE_LEG1,false);
		QL_FAIL_COND(null(r),"error solving for fixed coupon rate");

		/*fix 			= ql_fix_leg(fix_def,"fix_leg",td,sd,start_date,bond_mat_adj,null<tenor_code>,
									true,fix_nominal,fix_eff_date,fix_first_cpn_date,fix_last_reg_date,fix_r,disc_func,0,1);
		sw 				= ql_fixed_income_swap(fix,flt,"");*/
		
		number spread = bond_y-r;
		return spread;
	}

	
	//---------------------------------------------------------
	//fixflt_asw_helper_spr
	//version1: ql_swap_float_def and tenor_surface is created inside function
	//----------------------------------------------------------
	number fixflt_asw_helper_spr(date 					td,
								date 					sd,
								number 					bond_nominal,
								date 					bond_mat_adj,
								date option(nullable)	bond_last_reg_adj,
								vector(date) 			bond_dates,//post settle
								vector(number) 			bond_cpns100,
								number 					bond_roll_day,
								string					bond_currency,
								number 					ai100,
								number 					cp100,
								day_count_method 		flt_dc,
								bd_convention 			flt_bd,
								integer 				flt_freq,//assume reset_freq == cpn_freq
								logical 				flt_eom,
								..calendar 				flt_calendar,
								 string					flt_currency,
								..ir_index 				flt_ir_index,								
								number 	option(nullable)	fx_rate,
								disc_func option(nullable) 	disc_func_fix,
								disc_func 				disc_func_flt,	
								fwd_func  				flt_fwd_func,
								 number option(nullable) curr_fixing,								
								asw_type				type)
								//out number	upfront_pmt, 
								//out number pv_flt_leg,  
								//out number	pv_fix_leg )
	{		
		logical ccy_swap 		= !equal_casei(bond_currency,flt_currency);
		
		if(ccy_swap) {
			QL_FAIL_COND(null(fx_rate),"invalid fx rate");
			QL_FAIL_COND(null(disc_func_fix),"invalid disc_func (fixed leg)");
		}
		
		if(null(fx_rate))
			fx_rate = 1;
		if(null(disc_func_fix))
			disc_func_fix = disc_func_flt;
		
		number flt_nominal 			= bond_nominal * fx_rate;	
		date start_date 			= sd;
		date flt_eff_date 			= start_date;
		date_code settle_code 		= date_code("BD2");
		..coupon_freq  flt_cfreq 	= coupon_freq_from_int(flt_freq)  ;
		tenor_surface flt_ts 		= tenor_surface(td, disc_func_flt,[flt_fwd_func],[flt_ir_index],false);
		//float leg
		notional_exchg_style ne 	= ccy_swap? notional_exchg_style.NE_BOTH: notional_exchg_style.NE_NONE;
		ql_swap_float_def flt_def 	= ql_swap_float_def("float_def",flt_currency,settle_code,flt_cfreq,
												null<reset_freq>,flt_dc,flt_bd,flt_calendar,flt_eom,
												null<flt_comp_avg_type>, null<flt_sprd_comp_method>,null<flt_avg_method>,
												flt_ir_index,flt_reset_style.RS_NORM,ne);

		QL_FAIL_COND(null(flt_def),"invalid float leg (def)");
		
		date  flt_first_cpn_date 	= null<date>;
		date  flt_last_reg_date 	= null<date>;
	
		ql_float_leg flt  			= ql_float_leg(	flt_def,"float_leg",td,sd,start_date,bond_mat_adj,null<tenor_code>,
													false,flt_nominal,flt_eff_date,flt_first_cpn_date,flt_last_reg_date,0,disc_func_flt,	
													flt_ts,bond_roll_day,1,null<vector(date)>,null<vector(number)>,curr_fixing,false);

		swap_flt_leg_data flt_data 	= new swap_flt_leg_data(flt,false);
			
		vector(date) 	flt_pmt_dates 	= flt_data.pmt_date_ir(sd , cut_off_type.PMT_DATE_IR_FWD );	
		vector(number) 	flt_rates 		= flt_data.fix_rate(sd , cut_off_type.PMT_DATE_IR_FWD );
		vector(number) 	flt_accr_periods = flt_data.accr_per(sd , cut_off_type.PMT_DATE_IR_FWD );

		//asw calc
		number	fund_sprd 	= 0;
		number	inv_sprd 	= 0;
		asset_swap_helper asw;
		if(ccy_swap) {
			asset_swap_helper_ccy aswc 	= new asset_swap_helper_ccy (type,td, sd, cp100, ai100,bond_nominal,fx_rate,
																	 disc_func_fix,disc_func_flt, fund_sprd,  inv_sprd,
																	bond_dates,bond_cpns100, flt_pmt_dates, flt_accr_periods, flt_rates);
			asw = aswc;
			
		}
		else {
			asw = new asset_swap_helper (	type,td, sd, cp100, ai100,bond_nominal,disc_func_flt, fund_sprd,  inv_sprd,
											bond_dates,bond_cpns100, flt_pmt_dates, flt_accr_periods, flt_rates);
		}

		return asw.spread();
	}
	//---------------------------------------------------------
	//fixflt_asw_helper_spr
	//version2: ql_swap_float_def and tenor_surface is input
	//----------------------------------------------------------
	/*number fixflt_asw_helper_spr(date 					td,
								date 					sd,
								number 					bond_nominal,
								date 					bond_mat_adj,
								date option(nullable)	bond_last_reg_adj,
								vector(date) 			bond_dates,//post settle
								vector(number) 			bond_cpns100,
								number 					bond_roll_day,
								number 					ai100,
								number 					cp100,
								ql_swap_float_def 		flt_def,
								tenor_surface 			flt_ts,
								logical 				ccy_swap,
								number 	option(nullable)	fx_rate,
								disc_func option(nullable) 	disc_func_fix,
								disc_func 				disc_func_flt,	
								fwd_func  				flt_fwd_func,	
								number option(nullable) curr_fixing,								
								asw_type				type)
	{
		if(ccy_swap) {
			QL_FAIL_COND(null(fx_rate),"invalid fx rate");
			QL_FAIL_COND(null(disc_func_fix),"invalid disc_func (fixed leg)");
		}
		
		if(null(fx_rate))
			fx_rate = 1;
		if(null(disc_func_fix))
			disc_func_fix = disc_func_flt;
		
		number flt_nominal 	= bond_nominal * fx_rate;	
		date start_date 	= sd;
		date flt_eff_date 	= start_date;

		//float leg
		date  flt_first_cpn_date = null<date>;
		date  flt_last_reg_date = null<date>;
	
		ql_float_leg flt  = ql_float_leg(flt_def,"float_leg",td,sd,start_date,bond_mat_adj,null<tenor_code>,
										false,flt_nominal,flt_eff_date,flt_first_cpn_date,flt_last_reg_date,0,disc_func_flt,	
										flt_ts,bond_roll_day,1,null<vector(date)>,null<vector(number)>,curr_fixing,false);

		swap_flt_leg_data flt_data = new swap_flt_leg_data(flt,false);
			
		vector(date) 	flt_pmt_dates 	= flt_data.pmt_date_ir(sd , cut_off_type.PMT_DATE_IR_FWD );	
		vector(number) 	flt_rates 		= flt_data.fix_rate(sd , cut_off_type.PMT_DATE_IR_FWD );
		vector(number) 	flt_accr_periods = flt_data.accr_per(sd , cut_off_type.PMT_DATE_IR_FWD );

		//asw calc
		number	fund_sprd 	= 0;
		number	inv_sprd 	= 0;
		asset_swap_helper asw;
		if(ccy_swap) {
			asset_swap_helper_ccy aswc 	= new asset_swap_helper_ccy (type,td, sd, cp100, ai100,bond_nominal,fx_rate,
																	 disc_func_fix,disc_func_flt, fund_sprd,  inv_sprd,
																	bond_dates,bond_cpns100, flt_pmt_dates, flt_accr_periods, flt_rates);
			asw = aswc;
			
		}
		else {
			asw = new asset_swap_helper (	type,td, sd, cp100, ai100,bond_nominal,disc_func_flt, fund_sprd,  inv_sprd,
											bond_dates,bond_cpns100, flt_pmt_dates, flt_accr_periods, flt_rates);
		}

		return asw.spread();
	}*/
	
	//---------------------------------------------------------
	//fixflt_asw_helper
	//----------------------------------------------------------
	void fixflt_asw_helper(	date 					td,
							date 					sd,
							date 					start_date,
							logical 				is_orig,
							number 					bond_nominal,
							date 					bond_mat_adj,
							date option(nullable)	bond_last_reg_adj,
							date 					bond_next_cpn_adj,
							vector(date) 			bond_dates,//post settle
							vector(number) 			bond_cpns100,
							number 					bond_roll_day,
							number 					ai100,
							number 					cp100,
							number  option(nullable)	ai100_start,
							number 	 option(nullable)	cp100_start,
							ql_swap_fix_def 		fix_def,
							ql_swap_float_def 		flt_def,
							tenor_surface 			flt_ts,
							number option(nullable)	fx_rate,
							number option(nullable)	fx_rate_orig,
							number option(nullable) quote_spread,
							number option(nullable) quote_spread_orig,
							disc_func option(nullable) disc_func_fix,
							disc_func 				disc_func_flt,	
							fwd_func  				flt_fwd_func,	
							number option(nullable) curr_fixing,								
							asw_type				type,
							string 					name,
							out number 				upfront,//if is_orig=true this will upfront for the start date bond
							out number 				backend,//if is_orig=true this will backend for the start date bond
							out ql_fix_leg 			fix,
							out ql_float_leg 		flt,
							out ql_fixed_income_swap sw)
	{
		logical ccy_swap = equal_casei(fix_def.currency(),flt_def.currency());

		if(ccy_swap) {
			QL_FAIL_COND(null(fx_rate),"invalid fx rate");
			QL_FAIL_COND(null(disc_func_fix),"invalid disc_func (fixed leg)");
		}
		else {
			fx_rate = fx_rate_orig = 1;
		}

		number flt_sprd, fx_u;
		if(is_orig) {
			QL_FAIL_COND(null(ai100_start),"invalid accrued (as of start date)");	
			QL_FAIL_COND(null(cp100_start),"invalid clean price (as of start date)");
			//QL_FAIL_COND(ccy_swap && null(fx_rate_orig),"invalid fx rate (as of start date)");
			//QL_FAIL_COND(null(quote_spread_orig),"invalid spread (as of start date)");
			number prem_disc, accr_adj, swap_not;
			CORE_INT.asw_uf_be(type, cp100_start, ai100_start, bond_nominal,prem_disc, accr_adj,swap_not, upfront,backend);
			fx_u 		= fx_rate_orig;
			//flt_sprd 	= quote_spread_orig;
			flt_sprd 	= null(quote_spread_orig) ? 0 : quote_spread_orig;
		}
		else {
			//upfront/backend:
			//returns value with positive sign hence belongs on the fixed leg if we assume we buy the assetswap (i.e. buy the bond)	
			number prem_disc, accr_adj, swap_not;
			CORE_INT.asw_uf_be(type, cp100, ai100, bond_nominal,prem_disc, accr_adj,swap_not, upfront,backend);
			fx_u 		= fx_rate;			
			flt_sprd 	= null(quote_spread) ? 0 : quote_spread;
		}
		
		if(null(disc_func_fix))
			disc_func_fix = disc_func_flt;

		number flt_nominal 	= bond_nominal * fx_u;
		logical pay_fixed 	= true;
		
		//float leg
		date  flt_first_cpn_date 	= null<date>;
		date  flt_last_reg_date 	= null<date>;
		date flt_eff_date 			= start_date;					
		flt  						= ql_float_leg(flt_def,"float_leg",td,sd,start_date,bond_mat_adj,null<tenor_code>,
													!pay_fixed,flt_nominal,flt_eff_date,flt_first_cpn_date,flt_last_reg_date,flt_sprd,disc_func_flt,	
													flt_ts,bond_roll_day,1,null<vector(date)>,null<vector(number)>,curr_fixing,false);

		//fix leg
		date fix_first_cpn_date 	= bond_next_cpn_adj;	
		date fix_last_reg_date		= bond_last_reg_adj;	
		if(!null(fix_last_reg_date) && fix_last_reg_date < fix_first_cpn_date)// bond is in odd last period
			fix_last_reg_date = null<date>;
		date fix_eff_date 			= start_date;
		
		vector(date) fee_dates;	
		vector(number) fee;
		if(!null(upfront) && !CORE_INT.is_equal_zero(upfront)) {
			push_back(fee_dates,start_date);
			push_back(fee,upfront);
		}
		if(!null(backend) && !CORE_INT.is_equal_zero(backend)) {
			push_back(fee_dates,bond_mat_adj);
			push_back(fee,backend);
		}

		fix  = ql_fix_leg(	fix_def,"fix_leg",td,sd,start_date,bond_mat_adj,null<tenor_code>,
							pay_fixed,bond_nominal,fix_eff_date,fix_first_cpn_date,fix_last_reg_date,
							bond_cpns100/100.0*bond_nominal,disc_func_fix,bond_roll_day,fx_u,fee_dates,fee);		
		QL_FAIL_COND(null(fix),"invalid fixed leg");
		
		sw = ql_fixed_income_swap(fix,flt,name);
		QL_FAIL_COND(null(sw),"invalid swap");

		/*if(null(quote_spread)) {
			number flt_s	= asw.spread();//sw.solver(0,swap_solver_code.SS_FLT_SPRD_LEG1,false);
			QL_FAIL_COND(null(flt_s),"error solving for spread", E_INIT);

			flt  = ql_float_leg(flt_def,"float_leg",td,sd,start_date,bond_mat_adj,null<tenor_code>,
								!pay_fixed,flt_nominal,flt_eff_date,flt_first_cpn_date,flt_last_reg_date,flt_s,disc_func_flt,	
								flt_ts,bond_roll_day,1,null<vector(date)>,null<vector(number)>,curr_fixing,false);
			
			sw = ql_fixed_income_swap(fix,flt,name);

		}*/
	}
}