/*	
	Core curve functions
	Developer: Algorithmica Research, Magnus Nyström			
*/
option(null: hard);	

/*-----------------------------------------------------------------------
  Name :  	check_quote
  Notes :           
  Changes:
  			Date    		Who     Detail
  			13-Jul-2016 	MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
logical check_quote(curve option(nullable) 		c, 
					out string  err_name)
option (category: 'Yield Curve')
option(com_name: "check_quote_ext")
{
	try {
		err_name = "";
		if(null(c)) {
			err_name = "null_curve";
			return false;
		}
		vector(instrument) v_i = c.instruments();
		for(integer i = 0; i < v_size(v_i ); i++){
			if(null(v_i[i].quote())) {
				err_name = v_i[i].name();
				return false;
			}
		}
		return true;
	}
	catch {
		return false;
	}
}
/*-----------------------------------------------------------------------
  Name :  	check_quote
  Notes :           
  Changes:
  			Date    		Who     Detail
  			13-Jul-2016 	MN      Original code (moved from fiws) 
  ----------------------------------------------------------------------*/
logical check_quote(curve option(nullable) c)
option (category: 'Yield Curve')
{
	string err_name;
	return check_quote(c, err_name);
}

module CORE_INT
{
	/*-----------------------------------------------------------------------
	  Name :  	check_curve
	  Notes :           
	  Changes:
	  			Date    		Who     Detail
	  			13-Jul-2016 	MN      Original code (moved from fiws) 
	  ----------------------------------------------------------------------*/
	logical check_curve(curve option(nullable) 	crv,
						logical 				check_quotes, 					
						out string  err_msg)
	{
		try {
			if(null(crv)){
				err_msg = "null curve";
				return false;
			}
			string crv_name = crv.name();
		
			if(!..check_curve(crv)) {			
				err_msg = strcat('invalid ', crv_name);
				return false;
			}
		
			if(check_quotes && !..check_quote(crv,err_msg)){			
				err_msg = strcat(['invalid quote(s) in ', crv_name, ' [', err_msg, ']']);
				return false;
			}
			return true;
		}
		catch {
			return false;
		}
	}
	
	//-------------------------------------------------------
	// check_instr_type
	//-------------------------------------------------------
	logical check_instr_type(	curve option(nullable) 	c, 
								instr_type 				it, 
								out string err_name)
	{		
		if(!..check_curve(c))
			return false;
		return check_instr_type(c.instruments(),it, err_name);
	}
	
	//-------------------------------------------------------
	// check_instr_type
	//-------------------------------------------------------
	logical check_instr_type(	curve option(nullable) 	c, 
								instr_type 				it)
	{
		string err_name;
		return check_instr_type(c, it, err_name);
	}
	//-------------------------------------------------------
	// check_instr_type
	//-------------------------------------------------------
	logical check_instr_type(	curve option(nullable) 	c, 
								vector(instr_type) 		v_it, 
								out string err_name)
	{		
		if(!..check_curve(c))
			return false;
		return check_instr_type(c.instruments(),v_it, err_name);
	}
	
	//-------------------------------------------------------
	// check_instr_type
	//-------------------------------------------------------
	logical check_instr_type(	curve option(nullable) 	c, 
								vector(instr_type) 		v_it)
	{
		string err_name;
		return check_instr_type(c, v_it, err_name);
	}

	//-------------------------------------------------------
	// max_curve_date
	//-------------------------------------------------------
	date max_curve_date(curve crv)
	{
		vector(instrument) vi = crv.instruments();
		integer size = v_size(vi);
		return vi[size-1].maturity();	
	}
	//-------------------------------------------------------
	// max_curve_t
	//-------------------------------------------------------
	number max_curve_t(curve crv)
	{
		date max = max_curve_date( crv);	
		return (max - crv.trade_date())/365.0;
	}

	//-------------------------------------------------------
	// max_curve_t
	//-------------------------------------------------------
	number max_curve_t(cashflows_base crv)
	{
		vector(number) p = crv.get_all_periods();
		return p[v_size(p)-1];	
	}
	
	//-------------------------------------------------------
	// max_curve_date
	//-------------------------------------------------------
	date max_curve_date(cashflows_base crv)
	{
		number max_t = max_curve_t(crv);
		integer tt = integer(trunc(max_t*365.0));
		date max_d = crv.trade_date() + tt;
		return max_d ;	
	}	
}


//-------------------------------------------------------
//class  fixing_curve 
// instruments in curve must be db instruments
//-------------------------------------------------------	
class fixing_curve 
option (category: 'Yield Curve')
{
	public:
			
	number 				fixing_rate(date, string , out string);
	number 				fixing_rate(date, tenor_code , out string);
	number 				fixing_rate_ip(date, string ,string, date, out string);
	number 				fixing_rate_ip(date, tenor_code ,tenor_code, date, out string);
	vector(number)		fixing_rate(vector(date), string, logical, out string);
	vector(number)		fixing_rate(vector(date), tenor_code, logical, out string);
	
	curve_name			name();	
	vector(ir_index) 	index();
	vector(instrument) 	instruments();
	integer				n_instruments();
	ir_index 			find_ir_index(string );
	ir_index 			find_ir_index(tenor_code );
	instrument 			find_inst(string , date option(nullable) d = null<date>);
	instrument 			find_inst(tenor_code , date option(nullable) d = null<date>);
	curve				inst_curve(date option(nullable) d = null<date>);

	//fixing_curve(curve_name, date);
	fixing_curve(curve);
	fixing_curve(vector(instrument), string option(nullable));
	
	protected:

	void 				init();
	void 				check_index();
	string 				lookup_inst(string );
	string 				lookup_inst(tenor_code );
	number				fixing_rate_int(date 						fix_date, 
									string option(nullable)		tenor_s,
									tenor_code option(nullable)	tenor_c, 
									out string err_msg);
	number 				fixing_rate_ip_int(	date 		fix_date, 
									string option(nullable) tenor_sh_s,  
									string option(nullable) tenor_lo_s,
									tenor_code option(nullable) tenor_sh_c,  
									tenor_code option(nullable) tenor_lo_c, 
									date 		mat, 									 
									out string 	err_msg);

	vector(number) 		fixing_rate_int(vector(date) fix_date, 
											string 	option(nullable) 	tenor_s,
											tenor_code 	option(nullable) 	tenor_c, 
											logical 	allow_nulls, 
											out string 	err_msg);
	
	date 				d_create_;
	vector(instrument) 	inst_;
	curve_name 			cn_;						
	vector(ir_index) 	ib_;
	vector(string) 		ib_tenor_;
	vector(tenor_code) 	ib_tenor_code_;
	logical				db_curve_;
};
/*
notes:

1. when fixing curve is created from a curve or a curvename it is possible that a curve created for a 
different fixing date contains different instruments

2. when fixing curve is created from a vector of instruments then these instruments are assumed
to be valid for all fixing dates

*/

/*
todo:
check all instruments in curve: curr, same dc, same bd, same cal, etc
*/
//-------------------------------------------------------
//constructor
//-------------------------------------------------------	
/*fixing_curve.fixing_curve(	curve_name 	cn,
							date 		d)
							: d_create_(d), cn_(cn), db_curve_(true)
{		
	curve crv 	= curve(cn_,d_create_,"");
	QL_REQUIRE(!null(crv),"invalid curve");
	QL_REQUIRE(!crv.empty(),"empty curve");//allow empty curve? --> for a historical date we may have a curve
	inst_ 		= crv.instruments();
}*/
//-------------------------------------------------------
//constructor
//-------------------------------------------------------	
fixing_curve.fixing_curve(	curve 	crv_in)
							: d_create_(crv_in.trade_date()), cn_(crv_in.name())
{	
	if(null(cn_))
		db_curve_ = false;
	else {
		curve crv 	= curve(cn_,d_create_,"");//create curve from name to see if it is a curve that can be created from the db
		db_curve_ 	= !null(crv);
	}

	if(!db_curve_) {						//in this case the crv_in probably have been created in code and cannot be recreated from db											
		inst_ 	= crv_in.instruments();		//and we treat the curve the same way as if the fixing curve was created from a vector(instrument)	
		QL_REQUIRE(v_size(inst_) > 0,"invalid curve");
		init();
	}
	else if(crv_in.empty()) 
		QL_FAIL("empty curve"); //allow empty curve? --> for a historical date we may have a curve
	else {
		inst_ = crv_in.instruments();
		init();
	}
}
//-------------------------------------------------------
//constructor
//-------------------------------------------------------	
fixing_curve.fixing_curve(	vector(instrument) 			inst,
							string	option(nullable)	name)
							: inst_(clone_vector(inst)), cn_(name), db_curve_(false)
{		
	QL_REQUIRE(v_size(inst_) > 0,"invalid instrument vector");

	d_create_ = inst_[0].trade_date();
	
	init();	
}

//-------------------------------------------------------
// init
// run only when # instruments > 0 
//-------------------------------------------------------		
void fixing_curve.init()
{		
	string err_name;	
	QL_REQUIRE(CORE_INT.check_instr_type( inst_,instr_type.DEPOSIT_INDEX,err_name),'invalid curve (only index deposits allowed)',  err_name, true);			
		
	integer size = v_size(inst_);
	resize(ib_,size);
	resize(ib_tenor_,size);
	resize(ib_tenor_code_,size);

	for(integer i=0; i< size; i++) {
		ib_[i] 			= inst_[i].ir_index();
		QL_REQUIRE(!null(ib_[i]), 'failed to create ir_index object',inst_[i].name(),true );			
		ib_tenor_[i] 		= ib_[i].tenor();
		ib_tenor_code_[i] 	= ib_[i].tenor_code();
	}
	QL_REQUIRE(v_size(v_find_uniquei(ib_tenor_)) == size,'duplicate fixing instruments found');
	//QL_REQUIRE(v_size(v_find_unique(ib_tenor_code_)) == size,'duplicate fixing instruments found');	
	check_index();				
}

//-------------------------------------------------------
// check_index
//-------------------------------------------------------		
void fixing_curve.check_index()
{
	string err_name;
	string ccy = ib_[0].currency();
	QL_REQUIRE(!null(ccy),"invalid currency");
	QL_REQUIRE(CORE_INT.check_eq_currency(inst_, ccy, err_name),'invalid curve (wrong instrument currency)', err_name, true);

	integer size 	= v_size(inst_);	
	number spot_d 	= ib_[0].spot_days();
	logical is_lib 	= equal_casei(ib_[0].type() , "LIBOR");

	for(integer i=1; i< size; i++) {
		number _spot_d = ib_[i].spot_days();		
		QL_REQUIRE(spot_d == _spot_d,"invalid index spot days", inst_[i].name(), true);
		
		logical _is_lib = equal_casei(ib_[i].type() , "LIBOR");		
		QL_REQUIRE( _is_lib == is_lib, "invalid index type", inst_[i].name(), true);		
	}	
}

vector(instrument) 	fixing_curve.instruments()	{ return !null(inst_) ? clone_vector(inst_) : null<vector(instrument)>;}
integer 			fixing_curve.n_instruments(){ return v_size(inst_);}
curve_name 			fixing_curve.name()			{ return cn_;}
vector(ir_index) 	fixing_curve.index()		{ return ib_;}

//-------------------------------------------------------
// find_ir_index
//-------------------------------------------------------
ir_index fixing_curve.find_ir_index(string tenor)
{
	integer idx;
	if(CORE_INT.match(tenor,ib_tenor_, idx)) 
		return ib_[idx];
	else
		return null<ir_index>;
}
//-------------------------------------------------------
// find_ir_index
//-------------------------------------------------------
ir_index fixing_curve.find_ir_index(tenor_code tenor)
{
	integer idx;
	if(CORE_INT.match(tenor,ib_tenor_code_, idx)) 
		return ib_[idx];
	else
		return null<ir_index>;
}
//-------------------------------------------------------
// find_inst
//-------------------------------------------------------
instrument fixing_curve.find_inst(	string 					tenor, 
									date option(nullable) 	d)
{
	if(null(d)) d = d_create_;
	integer idx;
	if(CORE_INT.match(tenor,ib_tenor_, idx)) {
		if(d == d_create_)
			return inst_[idx];
		else {			
			string n = inst_[idx].name();				
			return instrument(n, d)	;
		}
	}				
	else
		return null<instrument>;
}
//-------------------------------------------------------
// find_inst
//-------------------------------------------------------
instrument fixing_curve.find_inst(	tenor_code 				tenor, 
									date option(nullable) 	d)
{
	if(null(d)) d = d_create_;
	integer idx;
	if(CORE_INT.match(tenor,ib_tenor_code_, idx)) {
		if(d == d_create_)
			return inst_[idx];
		else {			
			string n = inst_[idx].name();				
			return instrument(n, d)	;
		}
	}				
	else
		return null<instrument>;
}
//-------------------------------------------------------
// inst_curve
//-------------------------------------------------------
curve fixing_curve.inst_curve(date option(nullable) d)
{
	if(null(d)) d = d_create_;
	integer size = v_size(inst_);
	if(size == 0)
		return null<curve>;	
	if(d == d_create_)
		return curve(inst_, cn_);
	else {
	vector(instrument) vi;
		for(integer i=0;i<size;i++) {
			push_back(vi, instrument(inst_[i].name(), d));
		}
		return curve(vi,cn_);
	}
}
//-------------------------------------------------------
// lookup_inst
//-------------------------------------------------------
string fixing_curve.lookup_inst(string tenor)
{
	if(v_size(inst_) == 0)
		return null<string>;

	integer idx;
	if(CORE_INT.match(tenor,ib_tenor_, idx)) 
		return inst_[idx].name();	
	else	
		return null<string>;	
}
//-------------------------------------------------------
// lookup_inst
//-------------------------------------------------------
string fixing_curve.lookup_inst(tenor_code tenor)
{
	if(v_size(inst_) == 0)
		return null<string>;

	integer idx;
	if(CORE_INT.match(tenor,ib_tenor_code_, idx)) 
		return inst_[idx].name();	
	else	
		return null<string>;	
}
//-------------------------------------------------------
//fixing_rate
// tenor = ON, TN, SN, 1W,2W,3W, 1M,...12M, 1Y,...
//-------------------------------------------------------		
number fixing_curve.fixing_rate(date 						fix_date, 
								tenor_code 					tenor, 								
								out string   err_msg)
{
	return fixing_rate_int(fix_date,  null<string>,tenor,  err_msg);
}
//-------------------------------------------------------
//fixing_rate
// tenor = ON, TN, SN, 1W,2W,3W, 1M,...12M, 1Y,...
//-------------------------------------------------------		
number fixing_curve.fixing_rate(date 						fix_date, 
								string 						tenor, 								
								out string   err_msg)
{
	return fixing_rate_int(fix_date,  tenor,null<tenor_code>,  err_msg);
}

//-------------------------------------------------------
//fixing_rate
// tenor = ON, TN, SN, 1W,2W,3W, 1M,...12M, 1Y,...
//-------------------------------------------------------		
number fixing_curve.fixing_rate_int(date 						fix_date, 
									string option(nullable)		tenor_s,
									tenor_code option(nullable)	tenor_c, 
									out string   err_msg)
{
	err_msg = "";
	if(d_create_ == fix_date || !db_curve_){
		instrument i = null(tenor_s) ? find_inst(tenor_c, null<date>) : find_inst(tenor_s, null<date>);
		//string n = lookup_inst(tenor);
		if(null(i)) { 
			err_msg = strcat(["fixing instrument not available in current fixing curve (", str(fix_date), ")"]);
			return null<number>;
		}
		error_info ee = new error_info(true,false);
		number y = i.yield(false,ee);
		if(ee.is_error()) { 
			err_msg = strcat([ee.message(), "(", str(fix_date), ")"]);
			return null<number>;
		}
		return y;
		
	}
	else {
		curve n_crv = curve(cn_,fix_date);
		if(null(check_curve(n_crv))){
			err_msg = strcat(["invalid curve for current fixing date (", str(fix_date), ")"]);
			return null<number>;
		}
	
		string err_name;
		if(!CORE_INT.check_instr_type( inst_,instr_type.DEPOSIT_INDEX,err_name)){
			err_msg = strcat(["invalid curve for current fixing date (only index deposits allowed) (", str(fix_date), ")"]);
			return null<number>;
		}			
		vector(instrument) n_vi = n_crv.instruments();

		string n = null(tenor_s) ? lookup_inst(tenor_c) : lookup_inst(tenor_s);
		if(!null(n)) {//tenor exists in the original curve
			vector(integer) vfi = v_find(n_vi.name(), [n]); 
			if(v_size(vfi) == 1){//name also exists in current curve
				error_info ee = new error_info(true,false);
				number y = instrument(n, fix_date).yield(false,ee);
				if(ee.is_error()) {
					err_msg = strcat([ee.message(), "(", str(fix_date), ")"]);
					return null<number>;
				}
				return y;
			}
			else { 
				err_msg = strcat(["fixing instrument not available in current fixing curve (", str(fix_date), ")"]);
				return null<number>;
			}
		}
		else {//tenor does not exist in original curve so we add all new tenors in the current curve and check again					
			integer size = v_size(n_vi);
			logical add = false;
			for(integer i=0; i< size; i++) {
				vector(integer) v_f = v_find(inst_.name(), [n_vi[i].name()]);
				if(v_size(v_f) >= 1) continue;//no need to add if instrument already exists

				//init_tags(n_vi[i]);
				
				ir_index idx = n_vi[i].ir_index();
				QL_REQUIRE(!null(idx), 'failed to create ir_index object',n_vi[i].name(),true );

				string n = lookup_inst(idx.tenor_code());
				if(null(n)) {
					push_back(ib_,idx); 
					push_back(ib_tenor_,idx.tenor());
					push_back(ib_tenor_code_,idx.tenor_code());
					add = true;				
				}
			}
			if(!add) {
				err_msg = strcat(["fixing instrument not available in current fixing curve (", str(fix_date), ")"]);
				return null<number>;
			}
			else {
				check_index();
				string n = null(tenor_s) ? lookup_inst(tenor_c) : lookup_inst(tenor_s);
				if(null(n)) { 
					err_msg = strcat(["fixing instrument not available in current fixing curve (", str(fix_date), ")"]);
					return null<number>;
				}
				error_info ee = new error_info(true,false);
				number y = instrument(n, fix_date).yield(false,ee);
				if(ee.is_error()) {
					err_msg = strcat([ee.message(), "(", str(fix_date), ")"]);
					return null<number>;
				}
				return y;
			}
		}
	}
}
//-------------------------------------------------------
//fixing_rate
// tenor = ON, TN, SN, 1W,2W,3W, 1M,...12M, 1Y,...
//-------------------------------------------------------		
number fixing_curve.fixing_rate_ip(	date 		fix_date, 
									string 		tenor_sh,  
									string 		tenor_lo, 
									date 		mat, 									 
									out string  	err_msg)
{
	return fixing_rate_ip_int(fix_date, tenor_sh, tenor_lo, null<tenor_code> ,null<tenor_code>,  mat,  err_msg);
}
//-------------------------------------------------------
//fixing_rate
// tenor = ON, TN, SN, 1W,2W,3W, 1M,...12M, 1Y,...
//-------------------------------------------------------		
number fixing_curve.fixing_rate_ip(	date 		fix_date, 
									tenor_code 	tenor_sh,  
									tenor_code 	tenor_lo, 
									date 		mat, 									 
									out string  	err_msg)
{
	return fixing_rate_ip_int(fix_date, null<string>,null<string>, tenor_sh, tenor_lo,   mat,  err_msg);
}
//-------------------------------------------------------
//fixing_rate
// tenor = ON, TN, SN, 1W,2W,3W, 1M,...12M, 1Y,...
//-------------------------------------------------------		
number fixing_curve.fixing_rate_ip_int(	date 		fix_date, 
									string option(nullable) tenor_sh_s,  
									string option(nullable) tenor_lo_s,
									tenor_code option(nullable) tenor_sh_c,  
									tenor_code option(nullable) tenor_lo_c, 
									date 		mat, 									 
									out string 	err_msg)
{
	err_msg = "";
	number r_sh;
	if(null(tenor_sh_s))
		r_sh = fixing_rate(fix_date, tenor_sh_c, err_msg);
	else
		r_sh = fixing_rate(fix_date, tenor_sh_s, err_msg);
	if(null(r_sh))
		return null<number>;

	number r_lo;
	if(null(tenor_lo_s))
		r_lo = fixing_rate(fix_date, tenor_lo_c, err_msg);
	else
		r_lo = fixing_rate(fix_date, tenor_lo_s, err_msg);	
	if(null(r_lo))
		return null<number>;

	vector(integer) vf 	= null(tenor_sh_s) ? v_find(ib_tenor_code_, tenor_sh_c) : v_find(ib_tenor_, [tenor_sh_s]);
	
	date value_sh 		= ib_[vf[0]].value_date(fix_date);
	date m_sh 			= ib_[vf[0]].maturity(value_sh);
	vf 					= null(tenor_lo_s) ? v_find(ib_tenor_code_, tenor_lo_c): v_find(ib_tenor_, [tenor_lo_s]);
	//date value_lo 		= ib_[vf[0]].value_date(fix_date);
	date m_lo 			= ib_[vf[0]].maturity(value_sh);
	if(mat < m_sh) {
		err_msg = "maturity shorter that short tenor";
		return null<number>;
	}
	else if(mat > m_lo) {
		err_msg = "maturity longer that long tenor";
		return null<number>;
	}	
	else 
		return CORE_INT.ip_lin(	fix_date,m_sh,m_lo, r_sh, r_lo,mat);				
}

//-------------------------------------------------------
//fixing_rate
// tenor = ON, TN, SN, 1W,2W,3W, 1M,...12M, 1Y,...
//-------------------------------------------------------		
vector(number) fixing_curve.fixing_rate(vector(date) fix_date, 
										string 		tenor, 										
										logical 	allow_nulls, 
										out string 	err_msg)
{
	return fixing_rate_int(fix_date,  tenor, null<tenor_code>, allow_nulls, err_msg);
}
//-------------------------------------------------------
//fixing_rate
// tenor = ON, TN, SN, 1W,2W,3W, 1M,...12M, 1Y,...
//-------------------------------------------------------		
vector(number) fixing_curve.fixing_rate(vector(date) fix_date, 
										tenor_code 		tenor, 										
										logical 	allow_nulls, 
										out string  	err_msg)
{
	return fixing_rate_int(fix_date, null<string> , tenor, allow_nulls, err_msg);
}
//-------------------------------------------------------
//fixing_rate
// tenor = ON, TN, SN, 1W,2W,3W, 1M,...12M, 1Y,...
//-------------------------------------------------------		
vector(number) fixing_curve.fixing_rate_int(vector(date) fix_date, 
											string 	option(nullable) 	tenor_s,
											tenor_code 	option(nullable) 	tenor_c, 
											logical 	allow_nulls, 
											out string 	err_msg)
{
	err_msg = "";
	integer size  = v_size(fix_date);
	if(size == 0) { 
		err_msg = "invalid fix date vector (size == 0)";
		return null<vector(number)>;
	}

	vector(number) r;
	resize(r,size);
	string err_msg;

	for(integer i=0; i< size; i++) {
						
		r[i] = null(tenor_s) ? fixing_rate(fix_date[i], tenor_c, err_msg) : fixing_rate(fix_date[i], tenor_s, err_msg);
		if(!allow_nulls && null(r[i])) {
			return null<vector(number)>;
		}	
	}
	return r;	
}

//-------------------------------------------------------
//fixing_rate_ip
//-------------------------------------------------------
/*number fixing_curve.fixing_rate_ip(date mat)
{
	instrument depo = curve_.instruments()[0];		
	disc_func df_pre = bootstrap(curve_,ip_linear(),RT_SIMPLE); 
	date_code dc = date_code(strcat("BD", str(spot_days_,0)));
	date settle = dc.apply_fwd(fix_d_, depo.calendar());        
        curve curve_chop = curve_.chop_short (settle+1);
        disc_func df_ext = bootstrap_ext(df_pre,settle,curve_chop,ip_linear(),RT_SIMPLE,true);
        return df_ext.zero_rate(fix_d_,settle,mat,RT_SIMPLE, depo.accr_dc_method());
}*/

//-------------------------------------------------------
//fixing_curve
//-------------------------------------------------------	
/*fixing_curve fixing_curve(	curve_name 					cn,
							date 						d)	
option (category: 'Yield Curve')	
option(com_name: "fixing_curve_crv_n")						
{		
	return new fixing_curve(cn,d);
}*/

//-------------------------------------------------------
//fixing_curve
//-------------------------------------------------------	
fixing_curve fixing_curve(curve c)	
option (category: 'Yield Curve')
option(com_name: "fixing_curve")
{		
	return new fixing_curve(c);
}
//-------------------------------------------------------
//fixing_curve
//-------------------------------------------------------	
fixing_curve fixing_curve(	vector(instrument) 			vi,	
							string	option(nullable)	name)	
option (category: 'Yield Curve')							
option(com_name: "fixing_curve_v")
{		
	return new fixing_curve(vi,name);
}
//-------------------------------------------------------
//fixing_curve
//-------------------------------------------------------	
fixing_curve fixing_curve(	vector(instrument) 		vi,															
							fixing_curve			crv) 
option (category: 'Yield Curve')
option(com_name: "fixing_curve_v_alt")								
{
	return new fixing_curve(vi,crv.name());
}


module WSP_UTIL
{
	//-------------------------------------------------------
	// init_curve
	//-------------------------------------------------------
	logical init_curve(	curve c,
						vector(instr_type) allowed_types,
						out vector(instrument) vi,
						out string err_string)
	{
		try{
			err_string = "";

			if(!..check_curve(c)){
				vi 			= null<vector(instrument)>;
				err_string 	="empty curve";
				return false;
			}
			
			vi = c.instruments();
				
			string name;			
			if(!CORE_INT.check_instr_type(vi, allowed_types, name)){
				vi 			= null<vector(instrument)>;
				err_string 	= strcat(["invalid type of instrument in curve [",name,"]"]);
				return false;
			}
			return true;
		}
		catch {
			err_string = err.message();		
			return false;
		}
	}
	
	//-----------------------------------------------------------------------------
	//  set_cruve_yld
	//-----------------------------------------------------------------------------
	logical set_curve_yld(	curve 						c,
							vector(number) 				user_y_pct,//yield in percentage form
							out curve option(nullable) 	new_c)
	{
		try {
			vector(instrument) i_v = c.instruments();
			number crv_size = v_size(i_v);
			if(v_size(user_y_pct) != crv_size)
				return false;
					
			for(integer i = 0; i < crv_size; i++){
				if(!null(user_y_pct[i]))								
					i_v[i] = i_v[i].set_yield(user_y_pct[i]/100);
			}							
			new_c = curve(i_v, c.name());
			return true ;
		}
		catch {
			return false;
		}
	}
		
	//-----------------------------------------------------------------------------
	//  set_curve_quote
	//-----------------------------------------------------------------------------
	logical set_curve_quote(curve 				c,
							vector(number) 		user_q,
							out curve 			new_c)
	{
		try {
			vector(instrument) i_v = c.instruments();
			number crv_size = v_size(i_v);
			if(v_size(user_q) != crv_size)
				return false;

			error_info ee = new error_info(true,false);
			for(integer i = 0; i < crv_size; i++)
				i_v[i] = i_v[i].set_quote(user_q[i],ee);
										
			new_c = curve(i_v, c.name());
			return true ;
		}
		catch {
			return false;
		}
	}

	//-------------------------------------------------------
	//-------------------------------------------------------
	//-------------------------------------------------------
	// class input_curve_data
	//-------------------------------------------------------
	//-------------------------------------------------------
	//-------------------------------------------------------
	class input_curve_data
	{

	public:
		input_curve_data(date,vector(instrument) ,vector(number) option(nullable) ,vector(number) option(nullable),vector(string) option(nullable) );
		input_curve_data(date,vector(instrument) ,vector(instrument) ,vector(number) option(nullable) ,vector(number) option(nullable) ,
						 vector(number) option(nullable),vector(number) option(nullable),vector(string) option(nullable) );
		logical 			is_init();
		string				err_msg();
		date 				trade_date();
		vector(instrument) 	vi_orig_;
		vector(instrument) 	vi_orig_2_;		
		vector(number) 		usr_q_;
		vector(number) 		usr_q_2_;
		vector(number) 		usr_sp_;
		vector(number) 		usr_sp_2_;
		vector(string) 		incl_;
		vector(instrument) 	vi_incl_;//initiated in init()
		vector(instrument) 	vi_incl_2_;//initiated in init()
		vector(number) 		orig_q_;
		vector(number) 		orig_q_2_;

		vector(logical) 	vi_orig_incl_;
		
		
	protected:
		input_curve_data(date,vector(number) option(nullable),vector(number) option(nullable),vector(number) option(nullable),
						 vector(number) option(nullable),vector(string) option(nullable) );
		void 				spread_calc(integer,vector(number) ,vector(number),out vector(instrument) );
		void 				init();
		logical 			is_init_;
		string				err_msg_;
		logical 			p_2_sides_;
		date 				d_;
	};

	//-------------------------------------------------------
	// constructor  <protected>
	//-------------------------------------------------------
	input_curve_data.input_curve_data(date d, vector(number) option(nullable) usr_q,
									  vector(number) option(nullable) usr_q2,
									  vector(number) option(nullable) usr_sp,
									  vector(number) option(nullable) usr_sp2,
									  vector(string) option(nullable) incl)
									: vi_orig_(null<vector(instrument)>), vi_orig_2_(null<vector(instrument)>),									  
									  usr_q_(usr_q), usr_q_2_(usr_q2),usr_sp_(usr_sp),usr_sp_2_(usr_sp2),
									  incl_(incl), vi_incl_(null<vector(instrument)>) ,
									  orig_q_(null<vector(number)>),orig_q_2_(null<vector(number)>),vi_orig_incl_(null<vector(logical)>),
									  is_init_(false), err_msg_("uninitiated"), d_(d)
	{}

	//-------------------------------------------------------
	// constructor
	//-------------------------------------------------------
	input_curve_data.input_curve_data(date d, vector(instrument) vi,
									  vector(number) option(nullable) usr_q,
									  vector(number) option(nullable) usr_sp,
									  vector(string) option(nullable) incl)
									: vi_orig_(vi), vi_orig_2_(null<vector(instrument)>), usr_q_(usr_q), usr_q_2_(null<vector(number)>),									  
									  usr_sp_(usr_sp), usr_sp_2_(null<vector(number)>), incl_(incl), vi_incl_(null<vector(instrument)>) ,
									  orig_q_(null<vector(number)>),orig_q_2_(null<vector(number)>),vi_orig_incl_(null<vector(logical)>),
									  is_init_(false), err_msg_("uninitiated"), p_2_sides_(false), d_(d)
	{
		init();
	}
	//-------------------------------------------------------
	// constructor  bid/ask
	//-------------------------------------------------------
	input_curve_data.input_curve_data(date d,vector(instrument) vi,
									  vector(instrument) vi2,
									  vector(number) option(nullable) usr_q,
									  vector(number) option(nullable) usr_q2,
									  vector(number) option(nullable) usr_sp,
									  vector(number) option(nullable) usr_sp2,
									  vector(string) option(nullable) incl)
									: vi_orig_(vi), vi_orig_2_(vi2),usr_q_(usr_q), usr_q_2_(usr_q2),usr_sp_(usr_sp),usr_sp_2_(usr_sp2),
									  incl_(incl), vi_incl_(null<vector(instrument)>) ,
									  orig_q_(null<vector(number)>),orig_q_2_(null<vector(number)>),vi_orig_incl_(null<vector(logical)>),
									  is_init_(false), err_msg_("uninitiated"), p_2_sides_(true), d_(d)
	{
		init();
	}
	
	
	//-------------------------------------------------------
	// init
	//-------------------------------------------------------
	void input_curve_data.init()
	{
		try{
			if(null(vi_orig_))
				return;
			
			integer size 	= v_size(vi_orig_);

			resize(orig_q_,size);
			orig_q_ = null<number>;
			
			resize(orig_q_2_,size);
			orig_q_2_ = null<number>;

			resize(vi_orig_incl_,size);
			vi_orig_incl_ = false;

			if(size != v_size(usr_q_))  {
				resize(usr_q_,size);
				usr_q_ = null<number>;
			}
			if(p_2_sides_ && size != v_size(vi_orig_2_))  {
				QL_FAIL("invalid instrument vector (vector 2)");
			}
			
			if(size != v_size(usr_q_2_))  {
				resize(usr_q_2_,size);
				usr_q_2_ = null<number>;
			}
			if(size != v_size(usr_sp_))  {				
				resize(usr_sp_,size);
				usr_sp_ = null<number>;
			}

			if(size != v_size(usr_sp_2_))  {				
				resize(usr_sp_2_,size);
				usr_sp_2_ = null<number>;
			}
			
			if(size != v_size(incl_))  {				
				resize(incl_,size);
				incl_ = "x";
			}

			error_info ee 	= error_info(true,true);
			
			for(integer i=0; i < size; i++) {
				orig_q_[i] = vi_orig_[i].quote(ee);
				spread_calc(i,  usr_q_,usr_sp_,vi_orig_);
				if(p_2_sides_) {
					orig_q_2_[i] = vi_orig_2_[i].quote(ee);
					spread_calc(i,  usr_q_2_,usr_sp_2_,vi_orig_2_);
				}
			}
			resize(vi_incl_,0);
			resize(vi_incl_2_,0);
			for(integer i=0; i < size; i++) {
				if(!null(incl_[i]) && equal_casei(incl_[i], "x"))  {
					vi_orig_incl_[i] = true;
					push_back(vi_incl_, vi_orig_[i]);
					if(p_2_sides_)
						push_back(vi_incl_2_, vi_orig_2_[i]);					
				}	
			}
			
			is_init_ = true;
			err_msg_ = "";
			return;
			
		}	
		catch {
			err_msg_ = err.message();
			is_init_ = false;
		}
	}
	logical input_curve_data.is_init() { return is_init_;}
	string input_curve_data.err_msg() { return err_msg_;}
	date 	input_curve_data.trade_date() { return d_;}
	//-------------------------------------------------------
	// spread_calc
	//-------------------------------------------------------
	void input_curve_data.spread_calc(integer i, vector(number) usr_q, vector(number) usr_sp, out vector(instrument) vi_orig)
	{
		logical usrq = !null(usr_q[i]);
		logical usrs = !null(usr_sp[i]) && usr_sp[i] != 0;

		error_info ee 	= error_info(true,true);
		number sp = usr_sp[i];
		if(usrs) {
			quote_style q_style = vi_orig[i].quote_style_e(ee);
			switch(q_style){
			case quote_style.YIELD:
			case quote_style.SPREAD:
			case quote_style.DISC_MARGIN:
				sp = usr_sp[i]/10000;
				break;
			case quote_style.YIELD_PCT:
			
				sp = usr_sp[i]/100;
				break;
			case quote_style.SPREAD_BP:
			case quote_style.CLEAN_PCT_EXINFL:
			case quote_style.CLEAN_PCT:
			case quote_style.EDFUT:
			case quote_style.DIRTY_PCT:
			case quote_style.DISC_MARGIN_BP:
			case quote_style.FX_PIPS:
			case quote_style.FX_RATE://consider changing this i.e. sprd is pips
				break;
			default:
				usr_sp[i] = null<number>;
				usrs = false;
			}
		}
		if(usrq && usrs ) {
			vi_orig[i] = vi_orig[i].set_quote(usr_q[i] + sp);
		}
		else if(usrq){
			vi_orig[i] = vi_orig[i].set_quote(usr_q[i]);
		}
		else if(usrs) {
			number q = vi_orig[i].quote();
			if(!null(q))
				vi_orig[i] = vi_orig[i].set_quote(q + sp);
		}
		
		return ;
	}
	//-------------------------------------------------------
	//-------------------------------------------------------
	//-------------------------------------------------------
	// class input_curve_data_ext
	//-------------------------------------------------------
	//-------------------------------------------------------
	//-------------------------------------------------------
	/*
	 constructor: crv_orig_ is created
	 init: crv_incl_ is created
	 */
	class input_curve_data_ext : public input_curve_data
	{
		public:
			input_curve_data_ext(date,curve_name option(nullable) ,quote_side,logical,vector(instr_type) option(nullable)  );
		
			input_curve_data_ext(date,curve_name option(nullable) ,quote_side,logical,
								 vector(number) option(nullable) ,vector(number) option(nullable) ,
								 vector(string) option(nullable),vector(instr_type) option(nullable) );
			
			input_curve_data_ext(date,curve_name option(nullable) ,quote_side,quote_side,logical,
								 vector(number) option(nullable) ,vector(number) option(nullable) ,vector(number) option(nullable) ,
								 vector(number) option(nullable) ,vector(string) option(nullable),vector(instr_type) option(nullable) );
			
			void output_helper(	out vector(string),out vector(date),out vector(number),out vector(number),out vector(number),out vector(number),
								out vector(string) ,out integer );
			void output_helper(	out vector(string),out vector(date),out vector(number),out vector(number),out vector(number),out vector(number),out vector(number),out vector(number),out vector(number),out vector(number),
								out vector(string) ,out integer, out integer );
			void output_helper(	out vector(string),out vector(date),out vector(text_rgb),out vector(text_rgb),out vector(text_rgb),
								out vector(number),out vector(number),out vector(number),out vector(number),
								out vector(string) ,out integer, out integer );

			curve_name	name();
			curve 		crv_orig_;
			curve 		crv_incl_;
			curve 		crv_orig_2_;
			curve 		crv_incl_2_;
			curve_name 	name_;
			quote_side 	qs_;
			quote_side 	qs_2_;
		protected:
			void 		init(date, logical,vector(instr_type) option(nullable));
			
			
	};

	curve_name	input_curve_data_ext.name() { return name_;}

	//-------------------------------------------------------
	// constructor
	//-------------------------------------------------------
	input_curve_data_ext.input_curve_data_ext(	date d,
												curve_name option(nullable) name,
												quote_side qs,
												logical check_quotes,
												vector(instr_type) option(nullable)  allowed_types)
												: input_curve_data(d,null<vector(number)>,null<vector(number)>,null<vector(number)>,null<vector(number)>,null<vector(string)>),
												  crv_orig_(null<curve>), crv_incl_(null<curve>) ,
												  crv_orig_2_(null<curve>), crv_incl_2_(null<curve>),
												  name_(null_curve_name(name)), qs_(qs),qs_2_(null<quote_side>)
	{		
		this.init(d, check_quotes, allowed_types);
	}
	
	//-------------------------------------------------------
	// constructor
	//-------------------------------------------------------
	input_curve_data_ext.input_curve_data_ext(	date d,
												curve_name option(nullable) name,
												quote_side qs,
												logical check_quotes,
												vector(number) option(nullable) usr_q,
												vector(number) option(nullable) usr_sp,
												vector(string) option(nullable) incl,
												vector(instr_type) option(nullable)  allowed_types)
												: input_curve_data(d,usr_q,null<vector(number)>,usr_sp,null<vector(number)>,incl),
												  crv_orig_(null<curve>), crv_incl_(null<curve>) ,
												  crv_orig_2_(null<curve>), crv_incl_2_(null<curve>),
												  name_(null_curve_name(name)), qs_(qs),qs_2_(null<quote_side>)
	{		
		this.init(d, check_quotes, allowed_types);
	}
	//-------------------------------------------------------
	// constructor
	//-------------------------------------------------------
	input_curve_data_ext.input_curve_data_ext(	date d,
												curve_name option(nullable) name,
												quote_side qs,
												quote_side qs2,
												logical check_quotes,
												vector(number)  option(nullable) usr_q,
												vector(number)  option(nullable) usr_q2,
												vector(number)  option(nullable) usr_sp,
												vector(number)  option(nullable) usr_sp2,
												vector(string)  option(nullable) incl,
												vector(instr_type) option(nullable)  allowed_types)
												: input_curve_data(d,usr_q,usr_q2,usr_sp,usr_sp2,incl),
												  crv_orig_(null<curve>), crv_incl_(null<curve>) ,
												  crv_orig_2_(null<curve>), crv_incl_2_(null<curve>),
												  name_(null_curve_name(name)), qs_(qs),qs_2_(qs2)
	{		
		this.init(d, check_quotes, allowed_types);
	}
	
	//-------------------------------------------------------
	// init
	//-------------------------------------------------------
	void input_curve_data_ext.init(date d,
								   logical check_quotes,
								   vector(instr_type) option(nullable)  allowed_types)
	{
		try {
			p_2_sides_ = !null(qs_2_);
			
			QL_REQUIRE(!null(name_),"invalid curve name");
			crv_orig_ 	= curve(name_, d, qs_);
			
			logical ok 	= CORE_INT.check_curve(crv_orig_,check_quotes,err_msg_);
			QL_REQUIRE(ok,err_msg_);

			if(!null(allowed_types)){
				ok = WSP_UTIL.init_curve(crv_orig_,allowed_types, vi_orig_, err_msg_);
				QL_REQUIRE(ok,err_msg_);
			}
			else {
				vi_orig_ 	= crv_orig_.instruments();
			}

			if(p_2_sides_) {
				crv_orig_2_ 	= curve(name_, d, qs_2_);
			
				logical ok 	= CORE_INT.check_curve(crv_orig_2_,check_quotes,err_msg_);
				QL_REQUIRE(ok,err_msg_);
				vi_orig_2_ 	= crv_orig_2_.instruments();
			}
			
			input_curve_data.init();
			QL_REQUIRE(is_init_,err_msg_);
			
			crv_orig_ 	= curve(vi_orig_, name_);
			crv_incl_ 	= v_size(vi_incl_) > 0 ? curve(vi_incl_, crv_orig_.name()): null<curve>;

			if(p_2_sides_) {
				crv_orig_2_ 	= curve(vi_orig_2_, name_);
				crv_incl_2_ 	= v_size(vi_incl_2_) > 0 ? curve(vi_incl_2_, crv_orig_2_.name()): null<curve>;
			}
		
			
		}	
		catch {
			err_msg_ = err.message();
			is_init_ = false;
		}
	}


	//-------------------------------------------------------
	// output_helper
	//-------------------------------------------------------
	void input_curve_data_ext.output_helper(out vector(string) 	name,
											out vector(date) 	mat,
											out vector(number) 	q_orig,
											out vector(number) 	q,
											out vector(number)	usr_q,
											out vector(number)	usr_sp,
											out vector(string) 	incl,
											out integer s_orig)
	{
		vector(instrument) vi_orig 	= this.crv_orig_.instruments();
		s_orig 						= v_size(vi_orig);

		if(s_orig>0) {
			name 	= vi_orig.name();
			mat 	= vi_orig.maturity();
			q_orig	= clone_vector(this.orig_q_);
			q 		= vi_orig.quote();
			usr_q	= clone_vector(this.usr_q_);
			usr_sp	= clone_vector(this.usr_sp_);
			incl 	= clone_vector(this.incl_);
		}
		else {
			resize(name,0);
			resize(mat,0);
			resize(q_orig,0);
			resize(q,0);
			resize(usr_q,0);
			resize(usr_sp,0);
			resize(incl,0);
		}
	}
	//-------------------------------------------------------
	// output_helper
	//-------------------------------------------------------
	void input_curve_data_ext.output_helper(out vector(string) 	name,
											out vector(date) 	mat,
											out vector(number) 	q_orig,
											out vector(number) 	q_orig2,
											out vector(number) 	q,
											out vector(number) 	q2,
											out vector(number)	usr_q,
											out vector(number)	usr_q2,
											out vector(number)	usr_sp,
											out vector(number)  usr_sp2,
											out vector(string) 	incl,
											out integer s_orig,
											out integer s_orig2)
	{
		output_helper(name,mat,q_orig,q,usr_q,usr_sp,incl,s_orig);

		vector(instrument) vi_orig;
		if(p_2_sides_) {
		 	vi_orig 	= this.crv_orig_2_.instruments();
			s_orig2 	= v_size(vi_orig);

			if(s_orig > 0)
				QL_REQUIRE(s_orig2 == s_orig,"invalid size of second price vector");
		}
		else {
			s_orig2 	= 0;
		}
		
		if(s_orig2>0) {
			q_orig2	= clone_vector(this.orig_q_2_);
			q2 		= vi_orig.quote();			
			usr_q2	= clone_vector(this.usr_q_2_);
			usr_sp2	= clone_vector(this.usr_sp_2_);
		}
		else {
			resize(q_orig2,0);
			resize(q2,0);
			resize(usr_q2,0);
			resize(usr_sp2,0);
		}
	}
	//-------------------------------------------------------
	// output_helper  - work in progress
	//-------------------------------------------------------
	void input_curve_data_ext.output_helper(out vector(string) 	name,
											out vector(date) 	mat,
											out vector(text_rgb) 	q1q2,
											out vector(text_rgb) 	q,
											out vector(text_rgb) 	q2,
											out vector(number)	usr_q,
											out vector(number)	usr_q2,
											out vector(number)	usr_sp,
											out vector(number)	usr_sp2,
											out vector(string) 	incl,
											out integer s_orig,
											out integer s_orig2)
	{
		vector(number) _q, _q2;
		vector(number) 	_q_orig, _q_orig2;
		output_helper(name,mat,_q_orig, _q_orig2,_q,_q2, usr_q,usr_q2, usr_sp,usr_sp2,incl,s_orig,s_orig2);

		resize(q, s_orig);
		resize(q2, s_orig2);
		integer dec = 2;
		
		for(integer i=0;i<s_orig;i++) {
			
			q[i] = null(_q[i]) ? text_rgb("") : text_rgb(str(_q[i],dec));
		}
		for(integer i=0;i<s_orig2;i++) {
			
			q2[i] = null(_q2[i]) ? text_rgb("") : text_rgb(str(_q2[i],dec));
		}

		integer s_q1q2 = s_orig > 0 ? s_orig : s_orig2;
		resize(q1q2, s_q1q2);
		
		for(integer i=0;i<s_orig2;i++) {

			string s;
			if(s_orig > 0 && s_orig2 > 0) {
				string s1 = null(_q_orig[i]) ? "" : str(_q_orig[i],2);
				string s2 = null(_q_orig2[i]) ? "" : str(_q_orig2[i],2);
				s = strcat([s1," / ",s2]);
				
			}
			else if(s_orig > 0) {
				string s1 = null(_q_orig[i]) ? "" : str(_q_orig[i],2);
				s = strcat([s1," / "]);
			}
			else if(s_orig2 > 0) {
				string s2 = null(_q_orig2[i]) ? "" : str(_q_orig2[i],2);
				s = strcat([" / ",s2]);
			}
			
			q1q2[i] = text_rgb(s) ;
		}
	}




	
	//-------------------------------------------------------
	//  fwd_curve_filter
	//------------------------------------------------------- 
	vector(logical) fwd_curve_filter(date 				settle_date_fwd,
									vector(instrument)  vi_orig,
									vector(instr_type) 	allowed_types)
	{
		
		integer s = v_size(vi_orig);
		vector(logical) f[s];
		f = true;
		for(integer i = 0; i < s; i++){
			if(!vi_orig[i].is_valid() || settle_date_fwd >= vi_orig[i].maturity()){
				f[i] = false;
				continue;
			}
			if(!CORE_INT.check_instr_type([vi_orig[i]], allowed_types)){
				f[i] = false;
				continue;
			}							
		}
		return f;
	}

			
	//-------------------------------------------------------
	//  fwd_curve_carry_repo
	//  reporate used for reinvestment
	//------------------------------------------------------- 
	vector(instrument) fwd_curve_carry_repo(date 				trade_date,
											date 				trade_date_fwd,
											date 				settle_date_fwd,
											vector(instrument)  vi_orig,//sorted
											vector(number)		repo_rates,
											day_count_method 	repo_dc ,
											rate_type 			repo_rt,
											vector(instr_type) 	allowed_types)
	{
		integer s = v_size(vi_orig);
		QL_REQUIRE(v_size(repo_rates) == s,"internal error (fwd_curve_carry_repo)");
		
		vector(logical) f 		= fwd_curve_filter(settle_date_fwd,vi_orig,allowed_types);
		vector(integer) ok_i 	= v_find(f, true);
			
		integer size = v_size(ok_i);
		QL_REQUIRE(size > 0,"invalid forward curve");
			
		vector(instrument) vi = sub_vector(vi_orig, ok_i);
		vector(instrument) vi_fwd[size];
		vector(number) r = sub_vector(repo_rates, ok_i);

		error_info ee = new error_info(true,false);
		for(integer i = 0; i < size; i++){
			number y_fwd = vi[i].yield_fwd(settle_date_fwd,r[i],null<number>, repo_dc, repo_rt,
										   null<disc_func>, cpn_reinv_type.FREE,ee);
			vi_fwd[i] = vi[i].move_date(trade_date_fwd, settle_date_fwd);							
			vi_fwd[i] = vi_fwd[i].set_yield (y_fwd);
		}

		return vi_fwd;	
	}
	//-------------------------------------------------------
	//  fwd_curve_carry_repo
	//  reporate (from repo_df) used for reinvestment
	//------------------------------------------------------- 
	vector(instrument) fwd_curve_carry_repo(date 				trade_date,
											date 				trade_date_fwd,
											date 				settle_date_fwd,
											vector(instrument)  vi_orig,//sorted
											disc_func 			repo_df,
											vector(instr_type) 	allowed_types)
	{
		day_count_method dcm = day_count_method.DC_ACT_360;
		rate_type rtt = rate_type.RT_SIMPLE;
			
		integer s = v_size(vi_orig);
		vector(number) repo_rates[s];
		for(integer i = 0; i < s; i++){
			if(settle_date_fwd < vi_orig[i].maturity()){
				repo_rates[i] = repo_df.simple_rate(trade_date,vi_orig[i].settle_date(),settle_date_fwd,dcm);	
			}
		}
		return fwd_curve_carry_repo(trade_date,trade_date_fwd,settle_date_fwd, vi_orig,repo_rates,dcm ,rtt,allowed_types);
	}
	//-------------------------------------------------------
	//  fwd_curve_carry_df
	//	impfwd used for reinvestment
	//------------------------------------------------------- 
	vector(instrument) fwd_curve_carry_df(	date 				trade_date,
											date 				trade_date_fwd,
											date 				settle_date_fwd,
											vector(instrument)  vi_orig,//sorted
											disc_func 			df,
											vector(instr_type) 	allowed_types)
	{
		
		vector(logical) f 		= fwd_curve_filter(settle_date_fwd,vi_orig,allowed_types);
		vector(integer) ok_i 	= v_find(f, true);
			
		integer size = v_size(ok_i);
		QL_REQUIRE(size > 0,"invalid forward curve");
			
		vector(instrument) vi = sub_vector(vi_orig, ok_i);
		vector(instrument) vi_fwd[size];

		day_count_method dcm = day_count_method.DC_ACT_360;
		rate_type rtt = rate_type.RT_SIMPLE;
		
		error_info ee = new error_info(true,false);
		for(integer i = 0; i < size; i++){
			number y_fwd = vi[i].yield_fwd(settle_date_fwd,null<number>,null<number>, dcm, rtt,
										   df, cpn_reinv_type.FWD,ee);
			vi_fwd[i] = vi[i].move_date(trade_date_fwd, settle_date_fwd);							
			vi_fwd[i] = vi_fwd[i].set_yield (y_fwd);
		}

		return vi_fwd;	
	}
	//-------------------------------------------------------
	//  fwd_curve_imp_df
	//  note: no carry calculation
	//------------------------------------------------------- 
	vector(instrument) fwd_curve_imp_df(date 				trade_date,
										date 				settle_date_fwd,
										vector(instrument)  vi_orig,
										disc_func 			df,
										vector(instr_type) 	allowed_types )	
	{
		
		vector(logical) f 		= fwd_curve_filter(settle_date_fwd,vi_orig,allowed_types);
		vector(integer) ok_i 	= v_find(f, true);
			
		integer size = v_size(ok_i);
		QL_REQUIRE(size > 0,"invalid forward curve");
			
		vector(instrument) vi = sub_vector(vi_orig, ok_i);
		vector(instrument) vi_fwd[size];
		
		error_info ee = new error_info(true,false);
		for(integer i = 0; i < size; i++){
			
			vi_fwd[i] = vi[i].set_settle_date(settle_date_fwd);				
			number y = vi_fwd[i].yield(df,false,ee);									
			vi_fwd[i] = vi_fwd[i].set_yield(y,ee);
		}
		
		return vi_fwd;	
	}
}

module CORE_INT
{
	//-------------------------------------------------------
	// y_shift_curve
	//-------------------------------------------------------
	curve y_shift_curve(curve crv,
					  vector(number) shift)
	{
		integer size_shift 		= v_size(shift);
		if(size_shift == 0)
			return crv;
		
		vector(instrument) v_i 	= crv.instruments();
		integer size_all 		= v_size(v_i);
		if(size_all == 0)
			return crv;
		
		if(size_shift == 1){
			number s = shift[0];
			if(null(s) || CORE_INT.is_equal_zero(s))
				return crv;
			
			resize(shift, size_all) ;
			shift = s;
		}
		else {
			QL_REQUIRE(size_shift == size_all,"vector size mismatch");
		}

		for(integer i=0;i<size_all;i++){
			number s = shift[i];
			if(null(s) || CORE_INT.is_equal_zero(s))
				continue;

			instr_type it 	= v_i[i].instr_type();
			number y 		= v_i[i].yield();
			if(null(y))
				continue;
				
			switch(it){
			case instr_type.SWAP_FIXFLT:
			{
				
				swap_fixflt c 	= swap_fixflt(v_i[i]);
				c 				= c.set_rate_fix_leg(y + s );
				v_i[i] 			= c;
				break;
			}
			case instr_type.SWAP_FIXOIS:
			{
				
				swap_fixois c 	= swap_fixois(v_i[i]);
				c 				= c.set_rate_fix_leg(y + s );
				v_i[i] 			= c;
				break;
			}
			
			case instr_type.DEPOSIT:
			case instr_type.DEPOSIT_INDEX:
			case instr_type.FRA:
			case instr_type.IR_FUTURE:
			case instr_type.BILL:				
			case instr_type.FIXED_CPN_BOND:
			case instr_type.ZERO_CPN_BOND:
			case instr_type.RIBA:
			case instr_type.NOIS:
			//case instr_type.SWAP_FIXFIX:
			//case instr_type.SWAP_ZEROFLT: //CHECK
			//case instr_type.SWAP_ZEROOIS: //CHECK
			case instr_type.IL_REAL_BILL:		
			case instr_type.IL_REAL_FIXED_CPN_BOND:
			case instr_type.IL_REAL_ZERO_CPN_BOND:		
			//case instr_type.IL_SWAP_REAL_FIXFLT:
			//case instr_type.IL_SWAP_REAL_FIXOIS:	
			//case instr_type.IL_SWAP_ZERO://CHECK
			//case instr_type.IL_SWAP_FIX_Y_Y://CHECK
				v_i[i] = v_i[i].set_quote_from_yield(y + s);
				break;
			//case instr_type.BILL_FUTURE:
			//case instr_type.CALLABLE_FIX_CPN_BOND:
			//case instr_type.BOND_FUTURE:
			//case instr_type.RATING_ADJ_BOND:
			//case instr_type.FRN:
			//case instr_type.CALLABLE_FRN:
			//case instr_type.OVERNIGHT_FUTURE:
			//case instr_type.FX_SPOT:
			//case instr_type.FX_SWAP:
			//case instr_type.FX_FORWARD:
			//case instr_type.SWAP_FLTFLT:
			//case instr_type.SWAP_FLTFLT2S:
			//case instr_type.SWAP_FLTOIS:
			//case instr_type.SWAP_OISOIS:
			//case instr_type.SWAP_FLT_Y_Y:
			//case instr_type.SWAP_OIS_Y_Y:
			//case instr_type.EQUITY:	
			//case instr_type.GENERIC:
			default:
				QL_FAIL("invalid instrument type");		
			}						
		}
		return curve(v_i);
	}


	tuple instr_issuer_mat_tuple
	{		
		string issuer ;
		date maturity;
		integer i;

	};

	vector(instrument) sort_issuer_mat(vector(instrument) vi)
	{		
		vector(instr_issuer_mat_tuple) mydata;

		integer size = v_size(vi);
		if(size == 0)
			return vi;
		
		for(integer i=0;i<size;i++) {
			push_back(mydata, new instr_issuer_mat_tuple(vi[i].issuer(),vi[i].maturity(), i));
		}

		sort(mydata,mydata);

		return sub_vector(vi,mydata.i);
		
	}

	//-----------------------------------------------------------------------------
	// disc_func_rel                                                          
	//-----------------------------------------------------------------------------
	disc_func disc_func_rel(disc_func df_base, disc_func df_rel, number start,number end, number cut)
	{
		if(cut <= 0 || end <= 0 || end > cut)
			return null<disc_func>;

		if(end == cut)
			return disc_func_restrict(df_rel,0,cut) ;
		
		vector(number) per  = [end];
		vector(number) df	= [df_rel.disc(end)];
		number dfe_base		= df_base.disc(end);				
		number day_frac		= 1.0/365.0;
		integer i 			= 1;
		
		while(per[i-1] < cut){
			number p		= end + i*day_frac;						
			number rel_df	= df[i-1]/dfe_base;
			dfe_base		= df_base.disc(p);		
			number dff		= rel_df * dfe_base;
			
			push_back(per, p);
			push_back(df, dff);

			//number r = (df[i-1]/df[i]-1)*360;
			
			i++;
		}
		
		disc_func df_tmp = disc_func_interp(per,df, ip_linear(), rate_type.RT_SIMPLE);
		if(null(df_tmp))
			return null<disc_func>;
		
		disc_func df_res = disc_func_splice(df_base,df_tmp,start);//<-- ??? or end - day_frac
		if(null(df_res))
			return null<disc_func>;
		
		return disc_func_restrict(df_res,0,cut);
	}

	//-----------------------------------------------------------------------------
	// disc_func_fwd_bump                                                          
	//-----------------------------------------------------------------------------
	disc_func disc_func_fwd_bump(disc_func df_base, number start, number end, number df_sprd, number cut)
	{
		number dfs_base		= df_base.disc(start);
		number dfe_base		= df_base.disc(end);
		number sa			= 1.0/df_sprd-1.0;
		number df			= 1.0/(dfs_base/dfe_base + sa);
		number dfe 			= dfs_base*df;
		disc_func df_tmp 	= disc_func_interp([start, end],[dfs_base,dfe], ip_linear(), rate_type.RT_SIMPLE);	
		disc_func df_rel 	= disc_func_splice(df_base,df_tmp,start);
		return disc_func_rel(df_base, df_rel, start, end, cut);
	}

	//-----------------------------------------------------------------------------
	// disc_func_excl_toy_int
	// currently only single period instruments are supported
	//-----------------------------------------------------------------------------
	disc_func disc_func_excl_toy_int(   date 			d,
									  disc_func 	df_solve,
									  vector(number) instr_adj_d,									  
									  disc_func 	df_orig,
									  vector(number) sa,
									  vector(number) toy_s,
									  vector(number) toy_m)
	{
		QL_REQUIRE(v_size(instr_adj_d) == 2,"only single flow instrument currently supported");//for now

		integer size_cf 			= v_size(instr_adj_d);		
		vector(number) dfi_v 		= df_solve.disc(instr_adj_d);
		vector(number) dfi_orig_v 	= df_orig.disc(instr_adj_d);		
		
		number instr_s 				= instr_adj_d[0];		
		number dfis 				= dfi_v[0];		

		vector(number) ip_per[0];
		vector(number) ip_df[0];
		if(instr_s > 0) {
			ip_per = [instr_s];
			ip_df = [dfis];
		}
		
		integer size_toy	= v_size(toy_s);		
		number sum 			= 0;

		for(integer k=1;k<size_cf;k++) {

			number instr_cf_m 	= instr_adj_d[k];
			
			number dfi_df		= dfi_v[k]/dfis;
			number dfi_orig_comp= dfi_orig_v[0]/dfi_orig_v[k];
			
			for(integer i=0;i<size_toy;i++) {
				
				if(instr_cf_m <= toy_s[i] || instr_s >= toy_s[i])
					continue;

				number _dftoy_s = df_solve.disc(toy_s[i]);
				number _dftoy_m = df_solve.disc(toy_m[i]);
					
				number _df 		= dfi_df * _dftoy_s/_dftoy_m;
				
				sum				+= sa[i] / _df;
				
			}
			
			number comp_adj = dfi_orig_comp - sum;
			number dfim_adj = 1.0/comp_adj*dfis;

			push_back(ip_per, instr_cf_m);
			push_back(ip_df, dfim_adj);
			
			
		}
		
		return disc_func_interp(ip_per,ip_df, ip_linear(), rate_type.RT_CONT);
	}
	//-----------------------------------------------------------------------------
	// upd_instr_toy
	//-----------------------------------------------------------------------------	
	instrument upd_instr_toy(instrument i, disc_func df)
	{
		error_info 	ee			= new error_info(true,true);
		instr_type t = i.instr_type(ee);

		switch(t){
		case instr_type.DEPOSIT:
		case instr_type.DEPOSIT_INDEX:{
			deposit dep	= deposit(i);
			number cpn 	= dep.par_rate(df, ee);
			return dep.set_quote_from_yield(cpn,true,ee);
		}
		break;
		case instr_type.FRA:{
			fra f = fra(i);
			number rate = f.par_rate(df, ee);
			return f.set_rate(rate,null<date>, true,ee);
		}
		break;
		case instr_type.IR_FUTURE:{
			ir_future f = ir_future(i);
			number rate = f.par_rate(df, ee);
			return f.set_rate(rate,null<date>, true,ee);
		}
		break;
		case instr_type.SWAP_FIXFLT:{//ignoring float leg (assuming effect is negligible for bootstrap of disc_func)
			swap_fixflt sw = swap_fixflt(i);
			number rate = sw.par_rate_fix_leg(df, ee);
			return sw.set_rate_fix_leg(rate,ee);
		}
		break;
		case instr_type.SWAP_FIXOIS:{//ignoring float leg (assuming effect is negligible for bootstrap of disc_func)
			swap_fixois sw = swap_fixois(i);
			number rate = sw.par_rate_fix_leg(df, ee);
			return sw.set_rate_fix_leg(rate,ee);
		}
		break;
		case instr_type.BILL:
		case instr_type.FIXED_CPN_BOND:{			
			number y = i.yield(df, true, ee);
			return i.set_quote_from_yield(y,ee);
		}
		break;
		default:
			QL_FAIL("unsupported instrument type in toy removal handler");
		}
	}
	//-----------------------------------------------------------------------------
	// disc_func_excl_toy
	// currently only single period instruments are supported
	//-----------------------------------------------------------------------------	
	disc_func disc_func_excl_toy(	date d,
									out curve c,
									curve toy_c,
									interpolator boot_ip,
									rate_type	boot_rt)
	{
		error_info 	ee			= new error_info(true,true);
		disc_func df_orig 		= bootstrap(c, boot_ip, boot_rt);		
		vector(instrument) v_i 	= c.instruments();		
		integer size_instr 		= v_size(v_i);
		vector(instrument) toy 	= toy_c.instruments();
		integer size_toy 		= v_size(toy);

		//vector(date) instr_d = v_i[i].cash_flow_dates(true,false,false,ee);
		vectors_num instr_per_v = new vectors_num(size_instr);
		
		for(integer i=0;i<size_instr;i++) {

			vector(date) v 	= v_i[i].cash_flow_dates(true,false,false,ee);
			QL_REQUIRE(v_size(v) > 0,"invalid cash flows");
			
			instr_type t 	= v_i[i].instr_type(ee);
			switch(t) {
			case instr_type.FIXED_CPN_BOND:
			case instr_type.SWAP_FIXFLT:
			case instr_type.SWAP_FIXOIS: {
				date settle = v_i[i].settle_date(ee);
				if(v_size(v)==1) {
					v = concat([settle],v);
				}
				else if(v_size(v)==2) {
					if(settle < v[0])
						QL_FAIL("multiple cashflows currently work in progress");
					else if(settle > v[0])
						QL_FAIL("invalid cashflow dates (internal error)");
					//settle == v[0] --> fwd starting --> ok
				}
				else {
					QL_FAIL("multiple cashflows currently work in progress");
				}							
				break;
			}			
			case instr_type.FRA:
			case instr_type.IR_FUTURE:				
				break;
			case instr_type.DEPOSIT:
			case instr_type.DEPOSIT_INDEX:
			case instr_type.BILL:
				v = concat([v_i[i].settle_date(ee)],v);
				break;
			default:
				QL_FAIL("instrument type not supported");
			}
			instr_per_v.set(i,(v-d)/365.0);
		}
		for(integer i=0;i<size_toy;i++) {
			if(v_i[i].instr_type(ee) != instr_type.DEPOSIT)	{							
				QL_FAIL("toy instrument must be a deposit");
			}
		}
			
		//vector(date) m 			= v_i.maturity(ee);
		//vector(date) settle		= v_i.settle_date(ee);
		//vector(number) instr_s 	= (settle-d)/365.0;
		//vector(number) instr_m 	= (m-d)/365.0;

		vector(date) toy_settle = toy.settle_date(ee);
		vector(date) toy_mat 	= toy.maturity(ee);
		vector(number) toy_s 	= (toy_settle-d)/365.0;
		vector(number) toy_m 	= (toy_mat-d)/365.0;
		vector(number) per		= period(toy_settle,toy_mat,toy.dc_method(ee));

		vector(number) sa[size_toy];
		for(integer i=0;i<size_toy;i++) {
			sa[i] 	= toy[i].yield(false,ee)*per[i];
		}
		
		disc_func df = df_orig;
		
		for(integer i=0;i<2;i++) {//run twice because in reality a solver (but doing one update loop is enough to reach a good solution )
						
			for(integer j=0;j<size_instr;j++) {
				
				disc_func df_adj = disc_func_excl_toy_int(d,df, instr_per_v.get(j), df_orig, sa, toy_s, toy_m);				
				v_i[j] 			= upd_instr_toy(v_i[j], df_adj);
			}
			
			c = curve(v_i);
			df = bootstrap(c, boot_ip, boot_rt);
		}
		return df;
	}
}