/*	
	Option funcs b76 
	Developer: Algorithmica Research, Magnus Nyström			
*/
option(null: hard);

module CORE_INT_OPT
{
	//-------------------------------------------------------
	// black76_fk_init
	//-------------------------------------------------------
	void black76_fk_init(	out number 			f, 
							out number 			k, 					
							out number displacement)
	{
	
		if(!null(displacement) && !CORE_INT.is_equal_zero(displacement))	{	
			if(displacement < 0.0)	QL_FAIL("displacement must be non-negative");
		
			f += displacement;
			k += displacement;

			if(CORE_INT.zero_negative_check(k, CORE_INT.DBL_EPSILON))
				QL_FAIL("zero or negative strike (after displacement)");

			if(CORE_INT.negative_check(f, CORE_INT.DBL_EPSILON))
				QL_FAIL(throw_err("negative forward price (after displacement)"));
			
		}
		else {
			displacement = 0;

			if(CORE_INT.zero_negative_check(k, CORE_INT.DBL_EPSILON))
				QL_FAIL("zero or negative strike");

			if(CORE_INT.negative_check(f, CORE_INT.DBL_EPSILON))
				QL_FAIL(throw_err("negative forward price"));
		}
	}
}

//-------------------------------------------------------
// black76_prem
//-------------------------------------------------------
void black76_prem(	CORE_INT_OPT.bs_pre_calc bs,
					opt_type 		type,
					number 			f, 
					number 			k, 					
					out number 		fp, 
					out number 		sp)
{
	if(!bs.z_vol_){		
		if(type == opt_type.OPT_CALL) 
			fp = f*bs.Nd1()-k*bs.Nd2() ;		
		else if(type == opt_type.OPT_PUT) 
			fp = -f*bs.Nd1_rev()+k*bs.Nd2_rev() ;			
		else 
			QL_FAIL("invalid option type");
	}
	else {
		if(type == opt_type.OPT_CALL) 
			fp = max(0,f-k) ;			
		else if(type == opt_type.OPT_PUT) 
			fp = max(0,-f+k) ;
		else 
			QL_FAIL("invalid option type");
	}
	sp = bs.df_*fp ;
}

void black76_prem(	opt_type 		type,
					number 			f, 
					number 			k, 
					number 			v, 
					number 			df,
					number 			tau_v,
					number option(nullable) displacement,
					out number 		fp, 
					out number 		sp)
{
	if(CORE_INT.negative_check(tau_v, CORE_INT.DBL_EPSILON) || CORE_INT.negative_check(v, CORE_INT.DBL_EPSILON) ){//exactly 0 returns false
		fp = 0 ;
		sp = 0 ;
		return;
	}		

	CORE_INT_OPT.black76_fk_init(f, k, displacement);
	
	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(f,k, v,df,tau_v);
	
	black76_prem( bs,type,f, k, fp,sp)	;
}
//-------------------------------------------------------
// black76_prem
//-------------------------------------------------------
void black76_prem(	opt_type 		type,
					number 			f, 
					number 			k, 
					number 			v, 
					number 			cont_r,
					number 			tau,
					number 			tau_v,
					number option(nullable) displacement,
					out number 		fp, 
					out number 		sp)
option (category: 'Financial/Options/Vanilla/Functions')
{       
	number df = exp(-cont_r*tau);
	black76_prem(type,f, k, v, df,tau_v,displacement,fp,sp);
	return;
}

//-------------------------------------------------------
// black_76_delta_spot
//-------------------------------------------------------
number black_76_delta_spot(	opt_type 	type,
							number 		f, 
							number 		k, 
							number 		v, 
							//number 		cont_r,
							number option(nullable) cont_u_r,//to get spot delta
							number 		tau,
							number 		tau_v,
							number option(nullable) displacement )
option (category: 'Financial/Options/Vanilla/Functions')
{
	if(null(cont_u_r))
		return null<number>;
		
	CORE_INT_OPT.black76_fk_init(f, k, displacement);

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(f,k, v,1,tau_v);
	
	number dfq = exp(-tau*cont_u_r);
	
	if(type == opt_type.OPT_CALL) 
		return dfq * bs.Nd1();					
	else if(type == opt_type.OPT_PUT) 
		return -dfq * bs.Nd1_rev();		
	else 
		QL_FAIL("invalid option type");
}

//-------------------------------------------------------
// black_76_delta_fut
//-------------------------------------------------------
number black_76_delta_fut(	opt_type 	type,
							number 		f, 
							number 		k, 
							number 		v, 
							number 		df,
							number 		tau_v,
							number option(nullable) displacement )
{       		
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON){
		return 0;
	}

	CORE_INT_OPT.black76_fk_init(f, k, displacement);

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(f,k, v,df,tau_v);
	
	if(type == opt_type.OPT_CALL) 
		return bs.df_ * bs.Nd1();					
	else if(type == opt_type.OPT_PUT) 
		return -bs.df_ * bs.Nd1_rev();		
	else 
		QL_FAIL("invalid option type");
}
//-------------------------------------------------------
// black_76_delta_fut
//-------------------------------------------------------
number black_76_delta_fut(	opt_type 	type,
							number 		f, 
							number 		k, 
							number 		v, 
							number 		cont_r,
							number 		tau,
							number 		tau_v,
							number option(nullable) displacement )
option (category: 'Financial/Options/Vanilla/Functions')
{       		
	number df = exp(-cont_r*tau);
	return black_76_delta_fut(type,f, k,v,df,tau_v, displacement );
}
//-------------------------------------------------------
// black_76_delta_fwd   (forward contract, not futures)
//-------------------------------------------------------
number black_76_delta_fwd(	opt_type 	type,
							number 		f, 
							number 		k, 
							number 		v, 								
							number 		tau_v,
							number option(nullable) displacement )
option (category: 'Financial/Options/Vanilla/Functions')
{        
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON){
		return 0;
	}

	CORE_INT_OPT.black76_fk_init(f, k, displacement);

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(f,k, v,1,tau_v);

	if(type == opt_type.OPT_CALL) 
		return bs.Nd1();					
	else if(type == opt_type.OPT_PUT) 
		return -bs.Nd1_rev();		
	else 
		QL_FAIL("invalid option type");
}
//-------------------------------------------------------
// black_76_delta_fwd_c2p
//-------------------------------------------------------
number black_76_delta_fwd_c2p(number 		delta_call )
option (category: 'Financial/Options/Vanilla/Functions')
{
	return delta_call - 1.0;
}
//-------------------------------------------------------
// black_76_delta_fwd_p2c
//-------------------------------------------------------
number black_76_delta_fwd_p2c(number 		delta_put )
option (category: 'Financial/Options/Vanilla/Functions')
{
	return delta_put + 1.0;
}

//-------------------------------------------------------
// black_76_gamma_fut
//-------------------------------------------------------
number black_76_gamma_fut(	number 		f, 
							number 		k, 
							number 		v, 
							number 		df, 
							number 		tau_v,
							number  option(nullable) displacement)
{     
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;

	CORE_INT_OPT.black76_fk_init(f, k, displacement);

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(f,k, v,df,tau_v);

	return bs.df_ * bs.pdf_nd1()/(f*bs.vv_);	    		
}
//-------------------------------------------------------
// black_76_gamma_fut  
//-------------------------------------------------------
number black_76_gamma_fut(	number 	f, 
						number 	k, 
						number 	v, 
						number 	cont_r,
						number 	tau,
						number 	tau_v,
						number  option(nullable) displacement)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number df = exp(-cont_r*tau);
	return black_76_gamma_fut(f, k,v, df, tau_v,displacement);
}

//-------------------------------------------------------
// black_76_gamma_fwd
//-------------------------------------------------------
number black_76_gamma_fwd(	number 		f, 
							number 		k, 
							number 		v, 						
							number 		tau_v,
							number  option(nullable) displacement)
option (category: 'Financial/Options/Vanilla/Functions')
{     
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;

	CORE_INT_OPT.black76_fk_init(f, k, displacement);

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(f,k, v,1,tau_v);

	return bs.pdf_nd1()/(f*bs.vv_);	    		
}
//-------------------------------------------------------
// black_76_gamma_spot  
//-------------------------------------------------------
number black_76_gamma_spot(	number 	f, 
							number 	k, 
							number 	v, 
							number 	cont_r,
							number 	option(nullable) cont_u_r,
							number 	tau,
							number 	tau_v,
							number  option(nullable) displacement)
option (category: 'Financial/Options/Vanilla/Functions')
{
	
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;

	CORE_INT_OPT.black76_fk_init(f, k, displacement);

	if(null(cont_u_r)) cont_u_r = 0;
	number s = CORE_INT_OPT.bl_s_from_f(f, cont_r, tau,	 cont_u_r);

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(s,k, v,cont_r,cont_u_r,tau,tau_v);
	
	return bs.dfq_*bs.pdf_nd1()/(s*bs.vv_); //black_scholes_gamma
}
//-------------------------------------------------------
// black_76_vega  internal --> no displacement
//-------------------------------------------------------
number black_76_vega(CORE_INT_OPT.bs_pre_calc 	bs,
					 number 				f,
					 number 				tau_v)
{	
	return bs.pdf_nd1()*f*bs.df_*pow(tau_v, 0.5);
}
//-------------------------------------------------------
// black_76_vega  internal 
//-------------------------------------------------------
number black_76_vega(	number 	f, 
						number 	k, 
						number 	v, 
						number 	df, 
						number 	tau_v,
						number  option(nullable) displacement)
{     
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;
	
	CORE_INT_OPT.black76_fk_init(f, k, displacement);

	CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(f,k, v,df,tau_v);

	return black_76_vega(bs,f,tau_v);
						 
}
//-------------------------------------------------------
// black_76_vega  
//-------------------------------------------------------
number black_76_vega(	number 	f, 
						number 	k, 
						number 	v, 
						number 	cont_r,
						number 	tau,
						number 	tau_v,
						number  option(nullable) displacement)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number df = exp(-cont_r*tau);
	return black_76_vega(f, k,v, df, tau_v, displacement);
}
//-------------------------------------------------------
// black_76_imp_vol
//-------------------------------------------------------
number black_76_imp_vol(opt_type type,
						number 	f, 
						number 	k, 							
						number 	df, 
						number 	tau_v,
						number  option(nullable) displacement,
						number 	prem)//spot_prem
{ 	 
	CORE_INT_OPT.black76_fk_init(f, k, displacement);
	
	if (!CORE_INT_OPT.bl_check_bounds(type,f, k, prem/df))
		return -1;

	number sv = CORE_INT_OPT.bl_initial_sigma(type,f,k,tau_v,df,prem) ;
	if(sv <= CORE_INT.DBL_EPSILON)
		sv = 0.01 ;

	number  fp, sp, value, prime ;
	number sol_break 	= 0.000001 ;
	number nosol_break 	= 0.001 ;
	number max_i 		= 50 ;

	do {

		CORE_INT_OPT.bs_pre_calc bs = new CORE_INT_OPT.bs_pre_calc(f,k, sv,df,tau_v);
		
		black76_prem(bs,type,f,k,fp,sp);

		value = sp - prem;

		if(abs(value) <= sol_break)
			return sv ;

		prime = black_76_vega(bs, f, tau_v);
		
		if(CORE_INT.is_equal_zero(prime, CORE_INT.DBL_EPSILON ))
			return -1 ;
		
		sv -= value/prime ;
	
	} while(abs(value) > sol_break && sv > nosol_break && max_i-- > 0) ;
	
	if(abs(value) <= sol_break)
		return sv ;
	else
		return -1 ; 
}
//-------------------------------------------------------
// black_76_imp_vol
//-------------------------------------------------------
number black_76_imp_vol(	opt_type type,
							number 	f, 
							number 	k, 
							number 	cont_r,
							number 	tau,
							number 	tau_v,
							number  option(nullable) displacement,
							number 	prem)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number df = exp(-cont_r*tau);
	return black_76_imp_vol(type,f, k, df,tau_v,displacement,prem);
}