/*	
	Option funcs normal 
	Developer: Algorithmica Research, Magnus Nyström			
*/
option(null: hard);

//-------------------------------------------------------
// black_norm_prem
//-------------------------------------------------------
void black_norm_prem(	opt_type 		type,
						number 			f, 
						number 			k, 
						number 			v, 
						number 			df, 
						number 			tau_v, 
						out number 		fp, 
						out number 		sp)
{       
	if(CORE_INT.negative_check(tau_v, CORE_INT.DBL_EPSILON) || CORE_INT.negative_check(v, CORE_INT.DBL_EPSILON) ){//0 returns false
		fp = 0 ;
		sp = 0 ;
		return;
	}

	number vv = v * pow(tau_v, 0.5) ;
	
	if(vv >= CORE_INT.DBL_EPSILON){			
		number d1 = (f-k)/vv ;
		number phid1 = normal_pdf(d1);
	
		if(type == opt_type.OPT_CALL) {
			fp = (f-k)*normal_cdf(d1)+vv*phid1 ;
			sp = df*fp ;
		}
		else if(type == opt_type.OPT_PUT) {
			fp = (k-f)*normal_cdf(-d1)+vv*phid1 ;
			sp = df*fp ;
		}
		else 
			QL_FAIL("invalid option type");
	}	
	
	else {
		if(type == opt_type.OPT_CALL) {
			fp = max(0,f-k) ;
			sp = df*fp ;
		}	
		else if(type == opt_type.OPT_PUT) {	
			fp = max(0,-f+k) ;
			sp = df*fp ;			 
		}
		else 
			QL_FAIL("invalid option type");
	}	 
}

//------------------------------------------------------
// black_norm_prem
//-------------------------------------------------------
void black_norm_prem(	opt_type 		type,
						number 			f, 
						number 			k, 
						number 			v, 
						number 			cont_r,
						number 			tau, 
						number 			tau_v, 
						out number 		fp, 
						out number 		sp)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number df = exp(-cont_r*tau);
	black_norm_prem(type,f, k, v, df,tau_v,fp,sp);
}
//-------------------------------------------------------
// black_norm_vega
//-------------------------------------------------------
number black_norm_vega(	number 	f, 
						number 	k, 
						number 	v, 
						number 	df, 
						number 	tau_v)
{     
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;
	number vv = v * pow(tau_v, 0.5) ;	   
	number d1 = (f-k)/vv ;
	//number d11 = pow(tau_v, 0.5)/pow(2*CORE_INT.PI,0.5)*exp(-((d1*d1)/2));
	number phid1 = normal_pdf(d1);
	
	return df*pow(tau_v, 0.5)*phid1;					 
}

//-------------------------------------------------------
// black_norm_vega
//-------------------------------------------------------
number black_norm_vega(	number 	f, 
						number 	k, 
						number 	v, 
						number 	cont_r,
						number 	tau, 
						number 	tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{     
	number df = exp(-cont_r*tau);
	return black_norm_vega(f, k, v, df,tau_v);				 
}

//-------------------------------------------------------
// black_norm_delta
//-------------------------------------------------------
number black_norm_delta(opt_type 	type,
						number 		f, 
						number 		k, 
						number 		v, 
						number 		df, 
						number 		tau_v)
{        
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON){
		return 0;
	}

	number vv = v * pow(tau_v, 0.5) ;
	number d1 = (f-k)/vv ;

	number fd;
	if(type == opt_type.OPT_CALL) 
		fd = normal_cdf(d1);					
	else if(type == opt_type.OPT_PUT) 
		fd = -normal_cdf(-d1);		
	else 
		QL_FAIL("invalid option type");

	return df * fd;
}

//-------------------------------------------------------
// black_norm_delta
//-------------------------------------------------------
number black_norm_delta(	opt_type 	type,
						number 		f, 
						number 		k, 
						number 		v, 
						number 		cont_r,
						number 		tau, 
						number 		tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number df = exp(-cont_r*tau);
	return black_norm_delta(type,f, k, v,df,tau_v);
}

//-------------------------------------------------------
// black_norm_delta_fwd
//-------------------------------------------------------
number black_norm_delta_fwd(opt_type 	type,
							number 		f, 
							number 		k, 
							number 		v, 
							number 		df, 
							number 		tau_v)
{	
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON){
		return 0;
	}

	
	number vv = v * pow(tau_v, 0.5) ;
	number d1 = (f-k)/vv ;

	if(type == opt_type.OPT_CALL) 
		return normal_cdf(d1);					
	else if(type == opt_type.OPT_PUT) 
		return -normal_cdf(-d1);		
	else 
		QL_FAIL("invalid option type");
}
//-------------------------------------------------------
// black_norm_delta_fwd
//-------------------------------------------------------
number black_norm_delta_fwd(opt_type 	type,
							number 		f, 
							number 		k, 
							number 		v, 
							number 		cont_r,
							number 		tau, 
							number 		tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number df = exp(-cont_r*tau);
	return black_norm_delta_fwd(type,f,k,v, df, tau_v);
}
//-------------------------------------------------------
// black_norm_gamma
//-------------------------------------------------------
number black_norm_gamma(	number 		f, 
							number 		k, 
							number 		v, 
							number 		df, 
							number 		tau_v)
{     
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;
		
	number vv = v * pow(tau_v, 0.5) ;	   
	number d1 = (f-k)/vv ;
	number phid1 = normal_pdf(d1);
	
	//return df / (vv * pow(2*CORE_INT.PI,0.5))*exp(-((d1*d1)/2.0));
	return df * phid1 / vv;
}

//-------------------------------------------------------
// black_norm_gamma
//-------------------------------------------------------
number black_norm_gamma(	number 		f, 
							number 		k, 
							number 		v, 
							number 		cont_r,
							number 		tau, 
							number 		tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number df = exp(-cont_r*tau);    
	return black_norm_gamma(f, k, v,df,tau_v);	    		
}
//-------------------------------------------------------
// black_norm_gamma_fwd
//-------------------------------------------------------
number black_norm_gamma_fwd(number 		f, 
							number 		k, 
							number 		v,  
							number 		tau_v)
option (category: 'Financial/Options/Vanilla/Functions')
{
    
	if(tau_v <= CORE_INT.DBL_EPSILON || v <= CORE_INT.DBL_EPSILON)
		return 0;
		
	number vv = v * pow(tau_v, 0.5) ;	   
	number d1 = (f-k)/vv ;
	number phid1 = normal_pdf(d1);

	return phid1 / vv;    		
}
//-------------------------------------------------------
// black_norm_imp_vol
//-------------------------------------------------------
number black_norm_imp_vol(	opt_type 	type,
							number 		f, 
							number 		k, 									
							number 		df, 
							number 		tau_v,
							number 		prem)
{ 	
	if (!CORE_INT_OPT.bl_check_bounds(type,f, k, prem/df))
		return -1;
		 
	number sv 	= CORE_INT_OPT.bl_initial_sigma(type,f,k,tau_v,df,prem) ;
	sv 			= CORE_INT_OPT.bl_ln2normal_vol(sv,f,k) ;
	
	if(sv <= CORE_INT.DBL_EPSILON)
		sv = 0.01 ;	  

	number  fp, sp, value, prime ;
	number sol_break 	= 0.0000001 ;
	number nosol_break 	= 0.0001 ; 
	number max_i 		= 50 ;

	do {			
		black_norm_prem(type,f,k,sv,df,tau_v,fp,sp);

		value = sp - prem;

		if(abs(value) <= sol_break)
			return sv ;

		prime = black_norm_vega(f,k,sv,df,tau_v) ;
		
		if(CORE_INT.is_equal_zero(prime, CORE_INT.DBL_EPSILON ))
			return -1 ;
		
		sv -= value/prime ;
	
	} while(abs(value) > sol_break && sv > nosol_break && max_i-- > 0) ;
	
	if(abs(value) <= sol_break)
		return sv ;
	else
		return -1 ; 
}

//-------------------------------------------------------
// black_norm_imp_vol
//-------------------------------------------------------
number black_norm_imp_vol(	opt_type 	type,
							number 		f, 
							number 		k, 									
							number 		cont_r,
							number 		tau, 
							number 		tau_v,
							number 		prem)
option (category: 'Financial/Options/Vanilla/Functions')
{
	number df = exp(-cont_r*tau);
	return black_norm_imp_vol(type,f, k,df, tau_v,prem);
}