enum asset_swap_type option(category: "Financial/Asset Swap")
{
    AS_PAR_VALUE
	option(str: "par_value")
	option(desc:
	       "Par value, "
	       "i.e. notional = nominal amount for the bond."),
    AS_MARKET_CLEAN
	option(str: "market_clean")
	option(desc:
	       "Market value clean, "
	       "i.e. notional = clean price for the bond."),
    AS_MARKET_DIRTY
	option(str: "market_dirty")
	option(desc:
	       "Market value dirty, "
	       "i.e. notional = dirty price for the bond."),
    AS_NO_ADJ
	option(str: "no_adj")
	option(desc:
	       "No adjustments, "
	       "i.e. notional = nominal amount for the bond "
	       "and all difference in value between the bond "
	       "and the swap is reflected in the swap spread. ")
};

class asset_swap_result option(category: "Financial/Asset Swap")
{
public:
    number	accrued_adj();
    number	backend();
    number	notional();
    number	premium_adj();
    number	spread();
    number	upfront();

protected:
    number	aa_;
    number	be_;
    number	no_;
    number	pd_;
    number	sp_;
    number	uf_;
};

number asset_swap_result.accrued_adj()	{ return aa_; }
number asset_swap_result.backend()	{ return be_; }
number asset_swap_result.notional()	{ return no_; }
number asset_swap_result.premium_adj()	{ return pd_; }
number asset_swap_result.spread()	{ return sp_; }
number asset_swap_result.upfront()	{ return uf_; }

module __as_impl
{
    logical setup(instrument		b,
		  disc_func		disc,						// "Swap curve disc func"
		  disc_func 	option(nullable) disc_r,	// "Real rate swap curve disc func"
		  fwd_func		option(nullable) fwd_f,		// "Forward curve for swaps"
		  day_count_method	dc,
		  rate_type		rt,
		  number		freq,
		  out number		pv,
		  out number		df_ms,
		  out number		sum_rdfdt,
		  out number		sum_dfdt)
    {
	
	if(null(disc_r))	// If no cpi linked case, this will be the ordinary swap disc function.
		disc_r = disc;
	
	calendar	 c		= b.calendar();
	date		 d_m		= b.maturity();
	date		 d_t		= b.trade_date();
	date		 d_s		= b.settle_date();
	date		 d;
	number		 df_s;
	number		 df_d;

	if (d_s >= d_m) {
		return false;
	}

	d = c.adjust_date(d_m, BD_FOLLOWING);
	df_d = disc.disc((d - d_t) / 365);

	df_s = disc.disc((d_s - d_t) / 365);
	if (null(df_s))
	    return false;

	number df_s_r = disc_r.disc((d_s - d_t) / 365);
	if (null(df_s_r))
	    return false;


	integer		 delta_m	= round(12 / freq);

	if (delta_m * freq != 12)
	    throw(E_INVALID_ARG, "Invalid frequency");

	integer		 base_day	= day(d);
	date		 iter_d		= d;
	logical		 first		= true;

	sum_rdfdt	= 0;
	sum_dfdt	= 0;
	do {
	    date	 d_prev;
	    number	 df_prev, dt_d, r_d;

	    iter_d = add_months(iter_d, -delta_m);
	    if (days_in_month(iter_d) >= base_day)
			iter_d = date(year(iter_d), month(iter_d), base_day);

	    d_prev = c.adjust_date(iter_d, BD_MOD_FOLLOWING);
	    if (d_prev < d_s)
			d_prev = d_s;

	    df_prev = disc.disc((d_prev - d_t) / 365.0);
	    dt_d = period(d_prev, d, dc);

	    if(null(fwd_f))
			r_d = disc_to_rate(df_d / df_prev, rt, d_prev, d, dc);
		else
			r_d = fwd_f.fwd((d_prev - d_t) / 365.0);

	    if (null(df_d))
			return false;

	    if (first) {
			df_ms = df_d / df_s;
			first = false;
	    }

	    number	dfdt	= df_d / df_s * dt_d;

	    sum_rdfdt	+= dfdt * r_d;
	    sum_dfdt	+= dfdt;

	    d = d_prev;
	    df_d = df_prev;
	} while (d > d_s);

	number		df_m_r		= disc_r.disc((d_m - d_t) / 365);

	number factor = 1;
	if(b.is_il_bond() && !null(b.il_index_factor()))
		factor = b.il_index_factor();

	pv = b.dirty_price(disc_r) - 100 * factor * df_m_r / df_s_r;


	return !null(pv);
    }

    /**/

    class asr_impl : public asset_swap_result
    {
    public:
	asr_impl(asset_swap_type, instrument, number, number, number, number);
    };

    void asr_impl.asr_impl(asset_swap_type	type,
			   instrument		b,
			   number		pv,
			   number		df_ms,
			   number		sum_rdfdt,
			   number		sum_dfdt)
    {
	number	clean = b.clean_price();

	number factor = 1;
	if(b.is_il_bond() && !null(b.il_index_factor()))
		factor = b.il_index_factor();

	pd_ = clean - 100*factor;
	aa_ = b.accrued();

	switch (type) {
	case AS_PAR_VALUE:
	    no_ = 100 * factor;
	    uf_ = aa_ + pd_;
	    be_ = 0;
	    break;
	case AS_MARKET_CLEAN:
	    no_ = clean;
	    uf_ = aa_;
	    be_ = pd_;
	    break;
	case AS_MARKET_DIRTY:
	    no_ = b.dirty_price();
	    uf_ = 0;
	    be_ = aa_ + pd_;
	    break;
	case AS_NO_ADJ:
	    no_ = 100 * factor;
	    uf_ = 0;
	    be_ = 0;
	    break;
	}

	sp_ = ((pv - uf_ - be_ * df_ms) / no_ - sum_rdfdt) / sum_dfdt;
    }



	number asset_swap_inv(instrument b,
			      number			spread,
			      disc_func			disc,
				  fwd_func option(nullable)	fwd_f,
				  disc_func	 option(nullable) disc_r,
			      day_count_method		dc,
			      rate_type			rt,
			      number			freq,
			      asset_swap_type		type)
	{
	    number		pv;
	    number		df_ms;
	    number		sum_rdfdt;
	    number		sum_dfdt;
	
	    if (null(b.quote()) ||
		!__as_impl.setup(b, disc, disc_r, fwd_f, dc, rt, freq, pv, df_ms, sum_rdfdt, sum_dfdt))
		return null<number>;
	
	    number		A = sum_rdfdt + sum_dfdt * spread;
	    number		dirty;
	
	    switch (type) {
	    case AS_PAR_VALUE:
		dirty = pv + (1 - A) * 100;
		break;
	    case AS_MARKET_CLEAN:
		dirty =
		    (pv + b.accrued() * (A + df_ms - 1) + df_ms * 100) / (A + df_ms);
		break;
	    case AS_MARKET_DIRTY:
		dirty = (pv + df_ms * 100) / (A + df_ms);
		break;
	    case AS_NO_ADJ:
		return null<number>;
	    }
	
	    return b.price_to_yield(dirty);
	}


}

asset_swap_result asset_swap(instrument		b,
			     disc_func		disc,
			     day_count_method	dc,
			     rate_type		rt,
			     number		freq,
			     asset_swap_type	type)
    option(category: "Financial/Asset Swap")
{
	disc_func 	disc_r;
	fwd_func 	fwd_f;
    number		pv;
    number		df_ms;
    number		sum_rdfdt;
    number		sum_dfdt;

	if(!b.is_fix_cpn_bond())
		throw(E_INVALID_ARG, strcat(["Instrument ", b.name(), " is of invalid instrument type for asset swap"]));

    if (null(b.quote()) ||
	!__as_impl.setup(b, disc, disc_r, fwd_f, dc, rt, freq, pv, df_ms, sum_rdfdt, sum_dfdt))
	return null<asset_swap_result>;

    return new __as_impl.asr_impl(type, b, pv, df_ms, sum_rdfdt, sum_dfdt);
}


asset_swap_result asset_swap(instrument		b,
			     disc_func		disc,
				 fwd_func 	fwd_f,
			     day_count_method	dc,
			     rate_type		rt,
			     number		freq,
			     asset_swap_type	type)
    option(category: "Financial/Asset Swap")
{
 	disc_func disc_r;
    number		pv;
    number		df_ms;
    number		sum_rdfdt;
    number		sum_dfdt;

	if(!b.is_fix_cpn_bond())
		throw(E_INVALID_ARG, strcat(["Instrument ", b.name(), " is of invalid instrument type for asset swap"]));

    if (null(b.quote()) ||
	!__as_impl.setup(b, disc, disc_r, fwd_f, dc, rt, freq, pv, df_ms, sum_rdfdt, sum_dfdt))
	return null<asset_swap_result>;

    return new __as_impl.asr_impl(type, b, pv, df_ms, sum_rdfdt, sum_dfdt);
}


asset_swap_result asset_swap(instrument		il_b,
			     disc_func		disc,
				 disc_func 	disc_r,
			     day_count_method	dc,
			     rate_type		rt,
			     number		freq,
			     asset_swap_type	type)
    option(category: "Financial/Asset Swap")
{
	fwd_func 	fwd_f;
    number		pv;
    number		df_ms;
    number		sum_rdfdt;
    number		sum_dfdt;

	if(!il_b.is_il_bond())
		throw(E_INVALID_ARG, strcat(["Instrument ", il_b.name(), " is of invalid instrument type for index-linked asset swap"]));

    if (null(il_b.quote()) ||
	!__as_impl.setup(il_b, disc, disc_r, fwd_f, dc, rt, freq, pv, df_ms, sum_rdfdt, sum_dfdt))
	return null<asset_swap_result>;

    return new __as_impl.asr_impl(type, il_b, pv, df_ms, sum_rdfdt, sum_dfdt);
}


asset_swap_result asset_swap(instrument		il_b,
			     disc_func		disc,
				 fwd_func 	fwd_f,
				 disc_func 	disc_r,
			     day_count_method	dc,
			     rate_type		rt,
			     number		freq,
			     asset_swap_type	type)
    option(category: "Financial/Asset Swap")
{
	
    number		pv;
    number		df_ms;
    number		sum_rdfdt;
    number		sum_dfdt;

	if(!il_b.is_il_bond())
		throw(E_INVALID_ARG, strcat(["Instrument ", il_b.name(), " is of invalid instrument type for index-linked asset swap"]));

    if (null(il_b.quote()) ||
	!__as_impl.setup(il_b, disc, disc_r, fwd_f, dc, rt, freq, pv, df_ms, sum_rdfdt, sum_dfdt))
	return null<asset_swap_result>;

    return new __as_impl.asr_impl(type, il_b, pv, df_ms, sum_rdfdt, sum_dfdt);
}

number asset_swap_inv(instrument b,
		      number			spread,
		      disc_func			disc,
			  fwd_func 			fwd_f,
		      day_count_method		dc,
		      rate_type			rt,
		      number			freq,
		      asset_swap_type		type)
    option(category: "Financial/Asset Swap"){

  	disc_func	disc_r;

	if(!b.is_fix_cpn_bond())
		throw(E_INVALID_ARG, strcat(["Instrument ", b.name(), " is of invalid instrument type for asset swap"]));

	return __as_impl.asset_swap_inv(b,
		      spread,
		      disc,
			  fwd_f,
			  disc_r,
		      dc,
		      rt,
		      freq,
		      type);

}

number asset_swap_inv(instrument b,
		      number			spread,
		      disc_func			disc,
		      day_count_method		dc,
		      rate_type			rt,
		      number			freq,
		      asset_swap_type		type)
    option(category: "Financial/Asset Swap"){

  	disc_func	disc_r;
  	fwd_func 	fwd_f;

	if(!b.is_fix_cpn_bond())
		throw(E_INVALID_ARG, strcat(["Instrument ", b.name(), " is of invalid instrument type for asset swap"]));

	return __as_impl.asset_swap_inv(b,
		      spread,
		      disc,
			  fwd_f,
			  disc_r,
		      dc,
		      rt,
		      freq,
		      type);
}

number asset_swap_inv(instrument il_b,
		      number			spread,
		      disc_func			disc,
			  fwd_func 			fwd_f,
			  disc_func			disc_r,
		      day_count_method		dc,
		      rate_type			rt,
		      number			freq,
		      asset_swap_type		type)
    option(category: "Financial/Asset Swap"){

	if(!il_b.is_il_bond())
		throw(E_INVALID_ARG, strcat(["Instrument ", il_b.name(), " is of invalid instrument type for index-linked asset swap"]));

	return __as_impl.asset_swap_inv(il_b,
		      spread,
		      disc,
			  fwd_f,
			  disc_r,
		      dc,
		      rt,
		      freq,
		      type);

}

number asset_swap_inv(instrument il_b,
		      number			spread,
		      disc_func			disc,
			  disc_func			disc_r,
		      day_count_method		dc,
		      rate_type			rt,
		      number			freq,
		      asset_swap_type		type)
    option(category: "Financial/Asset Swap"){

  	fwd_func 	fwd_f;

	if(!il_b.is_il_bond())
		throw(E_INVALID_ARG, strcat(["Instrument ", il_b.name(), " is of invalid instrument type for index-linked asset swap"]));

	return __as_impl.asset_swap_inv(il_b,
		      spread,
		      disc,
			  fwd_f,
			  disc_r,
		      dc,
		      rt,
		      freq,
		      type);

}

