module __hz_impl
{
    instrument create_i2(	instrument				i,
			 				date					trade,
			 				number option(nullable)	yield)
    {
	
		instr_type it = i.instr_def().instr_type();
		if(		it != instr_type.FIXED_CPN_BOND 
			&&  it != instr_type.BILL 
			&&  it != instr_type.ZERO_CPN_BOND
			&&  it != instr_type.IL_REAL_BILL
			&&  it != instr_type.IL_REAL_FIXED_CPN_BOND
			&&  it != instr_type.IL_REAL_ZERO_CPN_BOND)	
			throw(E_INVALID_ARG, "invalid instrument type");
	
		string sc 		= i.settle_code();
		date_code dc 	= date_code(sc);
		date settle 	= dc.apply_fwd(trade,i.calendar());
	 	i 				= i.set_settle_date(settle);
	
		if (!null(yield))
		    i = i.set_yield(yield);
	
		return i;
    }

    class hz_func
    {
    public:
		hz_func(instrument		instr,
			date			d2,
			number option(nullable)	y2,
			rate_type		rt,
			day_count_method	dc);
	
		number		calc_rate();
		number		calc_y1(number r);
		number		calc_y2(number r);
	
	    private:
		virtual number	calc(number r) = 0;
	
	    protected:
		instrument	 b1;
		instrument	 b2;
		date		 d1;
		date		 d2;
		date		 settle1;
		date		 settle2;
		date		 last;
		rate_type	 rt;
		day_count_method dc;
		number		 p1;
		number		 p2;
		logical		 ok;
    };

    hz_func.hz_func(instrument				instr,
				    date					d2,
				    number option(nullable)	y2,
				    rate_type				rt,
				    day_count_method		dc)
	: b1(instr),
	  b2(create_i2(instr, d2, y2)),
	  d1(b1.trade_date()),
	  d2(b2.trade_date()),
	  settle1(b1.settle_date()),
	  settle2(b2.settle_date()),
	  last(null<date>),
	  rt(rt),
	  dc(dc),
	  p1(b1.dirty_price()),
	  p2(b2.dirty_price()),
	  ok(!null(p1) && !null(p2))
    {
	if (!ok)
	    return;

	if (settle1 >= settle2)
	    throw(E_INVALID_ARG, "settle1 >= settle2");

	vector(date)	dates	= b2.cash_flow_dates();

	if (v_size(dates) <= 0)
	    throw(E_UNSPECIFIC, "No cashflows");

	last = dates[0];
	if (last < settle2)
	    throw(E_UNSPECIFIC, "first cashflow before settle");
    }

     number hz_func.calc_rate()
     {
	 if (!ok)
	     return null<number>;

	 return zero_newton(this, &hz_func.calc, 0, 1e-10);
     }

     number hz_func.calc_y1(number r)
     {
	 if (!ok)
	     return null<number>;

	 number	df	= rate_to_disc(r, rt, settle1, settle2, dc);
	 number	p	= calc(r) * df + p1;

	 return b1.price_to_yield(p);
     }

     number hz_func.calc_y2(number r)
     {
	 if (!ok)
	     return null<number>;

	 number p	= p2 - calc(r);

	 return b2.price_to_yield(p);
     }

     /**/

     class hz_fix_func : public hz_func
     {
     public:
	 hz_fix_func(instrument			instr,
		     date			d2,
		     number option(nullable)	y2,
		     rate_type			rt,
		     day_count_method		dc);

    private:
	 virtual number calc(number r);

    protected:
	 number		cpns;
    };

    hz_fix_func.hz_fix_func(instrument			instr,
			    date			d2,
			    number option(nullable)	y2,
			    rate_type			rt,
			    day_count_method		dc)
	: hz_func(instr, d2, y2, rt, dc),
	  cpns(0)
    {}

    number hz_fix_func.calc(number r)
    {
	number	df = rate_to_disc(r, rt, settle1, settle2, dc);

	return p2 + cpns - p1 / df;
    }

    /**/

    class hz_fix_rate_func : public hz_fix_func
    {
    public:
	hz_fix_rate_func(instrument			instr,
			 date				d2,
			 number option(nullable)	y2,
			 rate_type			rt,
			 day_count_method		dc,
			 number				r_fix,
			 rate_type			rt_fix,
			 day_count_method		dc_fix);
    };

    hz_fix_rate_func.hz_fix_rate_func(instrument		instr,
				      date			d2,
				      number option(nullable)	y2,
				      rate_type			rt,
				      day_count_method		dc,
				      number			r_fix,
				      rate_type			rt_fix,
				      day_count_method		dc_fix)
	: hz_fix_func(instr, d2, y2, rt, dc)
    {
	if (!ok)
	    return;

	vector(date)	cf_d	= b1.cash_flow_dates();
	vector(number)	cf_amt	= b1.cash_flow_nom_amounts();
	integer		n	= v_size(cf_d);

	for (integer i = 0 ; i < n && cf_d[i] < last ; i++) {
	    number	 df;

	    if (cf_d[i] <= settle2)
		df = rate_to_disc(r_fix, rt_fix, cf_d[i], settle2, dc_fix);
	    else
		df = 1 / rate_to_disc(r_fix, rt_fix, settle2, cf_d[i], dc_fix);

	    cpns += cf_amt[i] / df;
	}
    }

    /**/

    class hz_fix_curve_spot_func : public hz_fix_func
    {
    public:
	hz_fix_curve_spot_func(instrument		instr,
			       date			d2,
			       number option(nullable)	y2,
			       rate_type		rt,
			       day_count_method		dc,
			       disc_func		disc);
    };

    hz_fix_curve_spot_func.
	hz_fix_curve_spot_func(instrument		instr,
			       date			d2,
			       number option(nullable)	y2,
			       rate_type		rt,
			       day_count_method	dc,
			       disc_func		disc)
	: hz_fix_func(instr, d2, y2, rt, dc)
    {
	if (!ok)
	    return;

	vector(date)	cf_d	= b1.cash_flow_dates();
	vector(number)	cf_amt	= b1.cash_flow_nom_amounts();
	integer		n	= v_size(cf_d);

	for (integer i = 0 ; i < n && cf_d[i] < last ; i++) {
	    number	dt	= (settle2 - cf_d[i]) / 365.0;
	    number	df;

	    if (dt >= 0)
		df = disc.disc(dt);
	    else
		df = 1 / disc.disc(-dt);

	    cpns += cf_amt[i] / df;
	}
    }

    /**/

    class hz_fix_curve_fwd_func : public hz_fix_func
    {
    public:
	hz_fix_curve_fwd_func(instrument		instr,
			      date			d2,
			      number option(nullable)	y2,
			      rate_type			rt,
			      day_count_method		dc,
			      disc_func			disc);
    };

    hz_fix_curve_fwd_func.hz_fix_curve_fwd_func(instrument		instr,
						date			d2,
						number option(nullable)	y2,
						rate_type		rt,
						day_count_method	dc,
						disc_func		disc)
	: hz_fix_func(instr, d2, y2, rt, dc)
    {
	if (!ok)
	    return;

	vector(date)	cf_d	= b1.cash_flow_dates();
	vector(number)	cf_amt	= b1.cash_flow_nom_amounts();
	integer		n	= v_size(cf_d);

	number		t2	= (settle2 - d1) / 365.0;
	number		df_2	= disc.disc(t2);

	for (integer i = 0 ; i < n && cf_d[i] < last ; i++) {
	    number	t	= (cf_d[i] - d1) / 365.0;
	    number	df_i	= df_2 / disc.disc(t);

	    cpns += cf_amt[i] / df_i;
	}
    }

    /**/

    class hz_free_func : public hz_func
    {
    public:
	hz_free_func(instrument			instr,
		     date			d2,
		     number option(nullable)	y2,
		     rate_type			rt,
		     day_count_method		dc);

    private:
	virtual number calc(number);

    private:
	integer		n_cpns;
	vector(date)	cf_d;
	vector(number)	cf_amt;
    };

    hz_free_func.hz_free_func(instrument		instr,
			      date			d2,
			      number option(nullable)	y2,
			      rate_type			rt,
			      day_count_method		dc)
	: hz_func(instr, d2, y2, rt, dc)
    {
	if (!ok)
	    return;

	integer		n;

	cf_d	= b1.cash_flow_dates();
	cf_amt	= b1.cash_flow_nom_amounts();
	n	= v_size(cf_d);

	for (integer i = 0 ; i < n ; i++)
	    if (cf_d[i] >= last) {
		n_cpns = i;
		break;
	    }
    }

    number hz_free_func.calc(number r)
    {
	number		 val	= p2;

	val -= p1 / rate_to_disc(r, rt, settle1, settle2, dc);

	for (integer i = 0 ; i < n_cpns ; i++) {
	    number	 df;

	    if (cf_d[i] <= settle2)
		df = rate_to_disc(r, rt, cf_d[i], settle2, dc);
	    else
		df = 1 / rate_to_disc(r, rt, settle2, cf_d[i], dc);

	    val += cf_amt[i] / df;
	}

	return val;
    }

    /**/

    public number hz_rate_fixed(instrument		instr,
				date			d2,
				number			y2,
				rate_type		rt,
				day_count_method	dc,
				number			r_fix,
				rate_type		rt_fix,
				day_count_method	dc_fix)
    {
	hz_func	f = new hz_fix_rate_func(instr, d2, y2, rt, dc,
					 r_fix, rt_fix, dc_fix);

	return f.calc_rate();
    }

    public number hz_y1_fixed(instrument		instr,
			      date			d2,
			      number			y2,
			      number			hz_rate,
			      rate_type			rt,
			      day_count_method		dc,
			      number			r_fix,
			      rate_type			rt_fix,
			      day_count_method		dc_fix)
    {
	hz_func f = new hz_fix_rate_func(instr, d2, y2,
					 rt, dc, r_fix, rt_fix, dc_fix);

	return f.calc_y1(hz_rate);
    }

    public number hz_y2_fixed(instrument		instr,
			      date			d2,
			      number			hz_rate,
			      rate_type			rt,
			      day_count_method		dc,
			      number			r_fix,
			      rate_type			rt_fix,
			      day_count_method		dc_fix)
    {
	hz_func f = new hz_fix_rate_func(instr, d2, null<number>,
					 rt, dc, r_fix, rt_fix, dc_fix);

	return f.calc_y2(hz_rate);
    }

    public number hz_rate_spot(instrument		instr,
			       date			d2,
			       number			y2,
			       rate_type		rt,
			       day_count_method		dc,
			       disc_func		disc)
    {
	hz_func f = new hz_fix_curve_spot_func(instr, d2, y2, rt, dc, disc);

	return f.calc_rate();
    }

    public number hz_y1_spot(instrument			instr,
			     date			d2,
			     number			y2,
			     number			hz_rate,
			     rate_type			rt,
			     day_count_method		dc,
			     disc_func			disc)
    {
	hz_func f = new hz_fix_curve_spot_func(instr, d2, y2, rt, dc, disc);

	return f.calc_y1(hz_rate);
    }

    public number hz_y2_spot(instrument			instr,
			     date			d2,
			     number			hz_rate,
			     rate_type			rt,
			     day_count_method		dc,
			     disc_func			disc)
    {
	hz_func f = new hz_fix_curve_spot_func(instr, d2, null<number>,
					       rt, dc, disc);

	return f.calc_y2(hz_rate);
    }

    public number hz_rate_fwd(instrument		instr,
			      date			d2,
			      number			y2,
			      rate_type			rt,
			      day_count_method		dc,
			      disc_func			disc)
    {
	hz_func f = new hz_fix_curve_fwd_func(instr, d2, y2, rt, dc, disc);

	return f.calc_rate();
    }

    public number hz_y1_fwd(instrument			instr,
			    date			d2,
			    number			y2,
			    number			hz_rate,
			    rate_type			rt,
			    day_count_method		dc,
			    disc_func			disc)
    {
	hz_func f = new hz_fix_curve_fwd_func(instr, d2, y2, rt, dc, disc);

	return f.calc_y1(hz_rate);
    }

    public number hz_y2_fwd(instrument			instr,
			    date			d2,
			    number			hz_rate,
			    rate_type			rt,
			    day_count_method		dc,
			    disc_func			disc)
    {
	hz_func f = new hz_fix_curve_fwd_func(instr, d2, null<number>,
					      rt, dc, disc);

	return f.calc_y2(hz_rate);
    }

    public number hz_rate_free(instrument		instr,
			       date			d2,
			       number			y2,
			       rate_type		rt,
			       day_count_method		dc)
    {
	hz_func f = new hz_free_func(instr, d2, y2, rt, dc);

	return f.calc_rate();
    }

    public number hz_y1_free(instrument			instr,
			     date			d2,
			     number			y2,
			     number			hz_rate,
			     rate_type			rt,
			     day_count_method		dc)
    {
	hz_func f = new hz_free_func(instr, d2, y2, rt, dc);

	return f.calc_y1(hz_rate);
    }

    public number hz_y2_free(instrument			instr,
			     date			d2,
			     number			hz_rate,
			     rate_type			rt,
			     day_count_method		dc)
    {
	hz_func f = new hz_free_func(instr, d2, null<number>, rt, dc);

	return f.calc_y2(hz_rate);
    }
}

import option(category: "Financial/Horizon Functions") __hz_impl;
