/*	
	Basic functions for "cpi" bootstrap     ALFA version
	Developed by Algorithmica Research, Magnus Nyström
	

	The seasonal factors are defined as in "Inflation Markets A Comprehensive and Cohesive Guide" 
	By Paul Canty and Markus Heider (Risk books)

	Multiplicative factors give the percentage deviation of the unadusted price index in a month
	relative to the seasonally adjusted series

	I(t) = T(t)*S(t)
	T=trend, S=season (normalized)


	Comments: BETA	

	ToDo: 	
			
*/
option(null: hard);

//-------------------------------------------------------
// create interpolation data
//-------------------------------------------------------
void interpolate( 	date 		last_known_dt,
					number 		last_known_v,
					date 		next_known_dt,
					number 		next_known_v,				
					vector(number) seas_f,
					date 		seas_base_date,
					integer 	index_freq,
					out vector(date) 	cpi_dates,
					out vector(number) 	cpi_values )
{
	integer mm 		= CORE_INT.months_between(last_known_dt, next_known_dt)-1;
	integer m_per 	= integer(floor(12/index_freq));
	integer m 		= integer(mm/m_per);
	resize(cpi_dates,m+2);
	resize(cpi_values,m+2);
	cpi_dates[0] = last_known_dt;
	cpi_values[0] = last_known_v;
	
	
	number per = (next_known_dt-last_known_dt)/365.0;
	number infl = (log(next_known_v)-log(last_known_v))*(1.0/per);
	
	for(integer i=0;i<m;i++){
		date fwd_date 	= add_months (last_known_dt, (i+1)*m_per);
		number v 		= __idx_forecast(seas_f,seas_base_date,index_freq,last_known_dt,last_known_v, fwd_date,infl, RT_CONT);
		cpi_dates[i+1] 	= fwd_date;
		cpi_values[i+1] = v;
	}

	cpi_dates[m+1] = next_known_dt;
	cpi_values[m+1] = next_known_v;	
}


//-------------------------------------------------------
// check_curve_type
//-------------------------------------------------------
void check_curve_type(curve il_zero_curve)
{
	QL_REQUIRE(check_curve(il_zero_curve),"invalid curve");
		
	string err_name;
	vector(instr_type) v_it = [instr_type.IL_SWAP_ZERO];		
	QL_REQUIRE(CORE_INT.check_instr_type(il_zero_curve, v_it, err_name),'invalid curve (wrong instrument type)', err_name, true);			
	
}
//-------------------------------------------------------
// first_swap
//-------------------------------------------------------
void first_swap(	swap_il_zero 	sw,
					vector(number) 	seas_f,
					date 			seas_base_date,
					date 			last_cpi_date,
					number 			last_cpi,
					out vector(date) cpi_dates_v,
					out vector(number) cpi_values_v)
{
	error_info ee = new error_info(true,true);
	
	date 	dt_lag1, dt_lag2;
	number  v_lag1, v_lag2;				
	sw.cpi_values(true,dt_lag1, dt_lag2,v_lag1, v_lag2,ee);

	logical is_noip = sw.is_noip_index_method(ee);
	if(!is_noip)
		QL_REQUIRE( !null(v_lag1) && !null(v_lag2),"missing historical cpi");
	else
		QL_REQUIRE( !null(v_lag1),"missing historical cpi");
	
	number  r = sw.infl_zero_rate(ee) ;
	date 	dt_fwd_lag1, dt_fwd_lag2;
	number  v_fwd_lag1, v_fwd_lag2;
	number imp_infl;
	sw.implied_index(r,seas_f,seas_base_date,false,dt_fwd_lag1, dt_fwd_lag2,v_fwd_lag1, v_fwd_lag2,imp_infl,ee);

	//interpolate intra-year cpi
	resize(cpi_dates_v,0);
	resize(cpi_values_v,0);
	
	date last_known_dt, next_known_dt;
	number last_known_v,next_known_v;
	if(last_cpi_date > dt_lag1){
		last_known_dt = last_cpi_date;
		last_known_v = last_cpi;
	}
	else {
		last_known_dt = dt_lag1;
		last_known_v = v_lag1;
	}
	if(!null(dt_fwd_lag2) && !null(v_fwd_lag2)){
		next_known_dt = dt_fwd_lag2;
		next_known_v = v_fwd_lag2;
	}
	else if(!null(dt_fwd_lag1) && !null(v_fwd_lag1)){ 
		next_known_dt = dt_fwd_lag1;
		next_known_v = v_fwd_lag1;
	}
	else {
		QL_FAIL("err");
	}

	integer index_freq = sw.index_freq(ee);
	interpolate( last_known_dt,last_known_v,next_known_dt, next_known_v,seas_f,
				 seas_base_date,index_freq,cpi_dates_v,cpi_values_v );

	integer s = v_size(cpi_dates_v);

	if(!null(dt_fwd_lag1) && !null(v_fwd_lag1)){			
		if(dt_fwd_lag1 > cpi_dates_v[s-1]) {
			push_back(cpi_dates_v,dt_fwd_lag1);
			push_back(cpi_values_v,v_fwd_lag1);			
		}
	}
	
}

//-------------------------------------------------------
// non_first_swap
//-------------------------------------------------------
void non_first_swap(swap_il_zero 	sw,
					vector(number) 	seas_f,
					date 			seas_base_date,
					date 			last_known_dt,
					number 			last_known_v,
					out vector(date) cpi_dates_v,
					out vector(number) cpi_values_v)
{
	error_info ee = new error_info(true,true);
		
	date 	dt_fwd_lag1, dt_fwd_lag2;
	number  v_fwd_lag1, v_fwd_lag2;
	number imp_infl;
	number r = sw.infl_zero_rate(ee) ;	
	
	sw.implied_index(r,seas_f,seas_base_date,true,dt_fwd_lag1, dt_fwd_lag2,v_fwd_lag1, v_fwd_lag2,imp_infl,ee);

	date next_known_dt;
	number next_known_v;
		
	if(!null(dt_fwd_lag2) && !null(v_fwd_lag2)){
		next_known_dt = dt_fwd_lag2;
		next_known_v = v_fwd_lag2;
	}
	else if(!null(dt_fwd_lag1) && !null(v_fwd_lag1)){ 
		next_known_dt = dt_fwd_lag1;
		next_known_v = v_fwd_lag1;
	}
	else {
		QL_FAIL("error in implied index calculation");
	}

	integer freq = sw.index_freq(ee);
	interpolate( last_known_dt,last_known_v,next_known_dt, next_known_v,seas_f,seas_base_date,freq,cpi_dates_v,cpi_values_v );

	integer s = v_size(cpi_dates_v);
			
	if(!null(dt_fwd_lag1) && !null(v_fwd_lag1)){			
		if(dt_fwd_lag1 > cpi_dates_v[s-1]) {
			push_back(cpi_dates_v,dt_fwd_lag1);
			push_back(cpi_values_v,v_fwd_lag1);
		}
	}
		
}
//-------------------------------------------------------
// bootstrap_cpi
//-------------------------------------------------------
void bootstrap_cpi(	curve 				il_zero_curve,
					vector(number) 		seas_f,
					date 				seas_base_date,
					out vector(date) 	cpi_dates_v,
					out vector(number) 	cpi_values_v,
					error_info option(nullable) error)
{
	try {
		error_info ee = new error_info(true,true);

		/*
		 1. check instrument types
		 2. check that all have same monthly index -TODO
		 3. check that all have same currency -TODO
		 */
		check_curve_type(il_zero_curve);
		vector(instrument) vi = il_zero_curve.instruments();
		integer size = v_size(vi);
		
		vector(swap_il_zero) sw_v[size];
		for(integer i=0;i<size;i++) {
			sw_v[i] = swap_il_zero(vi[i],ee);
		}

		date trade_date 	= sw_v[0].trade_date(ee);
		cpi_link cpi_l 		= sw_v[0].cpi_link(ee);
		date last_cpi_date 	= cpi_l.find_last_published(trade_date);
		number last_cpi 	= cpi_l.value(last_cpi_date);
		
		first_swap(sw_v[0], seas_f, seas_base_date, last_cpi_date, last_cpi,cpi_dates_v,cpi_values_v);
		cpi_link cpi_l_f 	= cpi_link_forecast(cpi_dates_v,cpi_values_v);
		
		for(integer i=1;i<size;i++) {
			
			sw_v[i].add_cpi_link_forecast(cpi_l_f);

			integer s 			= v_size(cpi_dates_v);
			date last_known_dt 	= cpi_dates_v[s-1];
			number last_known_v = cpi_values_v[s-1];
		
			non_first_swap(sw_v[i], seas_f, seas_base_date, last_known_dt,last_known_v,cpi_dates_v, cpi_values_v);
			cpi_l_f 			= cpi_link_forecast(cpi_l_f, cpi_dates_v,cpi_values_v);
		}

		vector(date) cpi_fcast_pubdates;
		cpi_l_f.data(cpi_dates_v,cpi_values_v,cpi_fcast_pubdates);
	}
	catch {
		CORE_INT.catch_error_info(error,err.type(),err.message(), "bootstrap_cpi");
		resize(cpi_dates_v,0);
		resize(cpi_values_v,0);		
	}
}





