/*	
	Basic swap functions for "dual" bootstrap
	Developed by Algorithmica Research, Magnus Nyström

	swap_curve_disc_ext and swap_curve_disc_b1_ext:
		used to create a basic swap market discount function (no bonds supported here)
		

	swap_curve_disc_parm used with swap_curve_disc_ext:
		can store 3 curves - short, middle, long
		short curve: 			can contain DEPOSIT, DEPOSIT_INDEX, SWAP_FIXOIS, SWAP_FIXFLT
		middle curve: 			can contain IR_FUTURE, FRA, SWAP_FIXFLT, SWAP_FIXOIS
		long curve:				can contain DEPOSIT, DEPOSIT_INDEX, IR_FUTURE, FRA, SWAP_FIXOIS, SWAP_FIXFLT


	swap_curve_b1_disc_parm used with swap_curve_disc_b1_ext:
		can store 5 curves - short, middle, long
		short curve: 			as swap_curve_disc_parm
		middle curve: 			as swap_curve_disc_parm
		long curve:				as swap_curve_disc_parm
		
		libor swap curve: 		can contain SWAP_FIXFLT (currently only spot starting)
		ois lib basis curve:	can contain SWAP_FLTOIS (currently only spot starting)
		[the last 2 curves are combined into a SWAP_FIXOIS
			1) fix = flt
			2) flt = ois+s
			--> fix-s = ois
		]

	Comments: 	- good to use many instruments (knots) in the 0Y-1Y area
				- step-function between cb meeting dates recommended
				- include turns
				- from 2Y the intruments are normally IBOR/OIS basis swaps
				- single period (imm) basis swaps active below 3Y

	ToDo: 	- edfut_cvx_adj?
			- allow for single period (imm) basis swaps
			- handling of turns
	
*/

option(null: hard);	

/*-----------------------------------------------------------------------
  Name :    class swap_curve_disc_ext
  Notes :           
  Changes:
  			Date    		Who     Detail
  			? 				MN      Original code (moved from fiws)
  ----------------------------------------------------------------------*/
class swap_curve_disc_ext : public swap_curve_ext
option (category: 'Yield Curve/003 Curve Building')
{
   	public:
		swap_curve_disc_ext(swap_curve_disc_parm, date, string);				
		swap_curve_disc_ext(swap_curve_disc_ext ) ;
		
		override disc_func 		disc_df(disc_z_model);
		override disc_func 		disc_df(disc_func,disc_z_model);	
		override swap_curve_type type();
		curve 					imp_short_curve(disc_func );
		curve 					imp_fra_curve(disc_func );
		curve 					imp_swap_curve(disc_func );
		curve 					imp_blended_curve(disc_func);
		vector(disc_func) 		instr_bumper_df_vect(disc_z_model ,number,out vector(integer) ,out disc_func,out disc_func,
													 out vector(instrument),out vector(disc_func) );

		matrix(number) 			sens_matrix_bump(number,number,fwd_func option(nullable) ,vector(integer) ,disc_func ,
												disc_func,vector(disc_func));
	protected:
			
	private:			
	
		void				init();
		void 				check_types();	
};

swap_curve_type		swap_curve_disc_ext.type() 	{ return type_; }

//-------------------------------------------------------
// constructor 
//-------------------------------------------------------
swap_curve_disc_ext.swap_curve_disc_ext(swap_curve_disc_parm	p,
										date				d,																				
										string 				name) 
										:swap_curve_ext(p,d,name, swap_curve_type.SCE_DISC)//init in swap_curve_ext is not run
{					
	init();		
}
//-------------------------------------------------------
// copy constructor
//-------------------------------------------------------
swap_curve_disc_ext.swap_curve_disc_ext(swap_curve_disc_ext s) 
										:swap_curve_ext(s)
{}
//-------------------------------------------------------
// init
//-------------------------------------------------------
void swap_curve_disc_ext.init()
{
	QL_REQUIRE(param_.type() == swap_curve_type.SCE_DISC, 'invalid type of param object (should be swap_curve_disc_parm)');	

	curve_ext.check();
	check_types();
	blend();						
}
//-------------------------------------------------------
// check_curves_disc
//-------------------------------------------------------
void swap_curve_disc_ext.check_types()
{	
	string err_name;	
	if(check_curve(param_.short_crv())){
		vector(instr_type) v_it = [instr_type.DEPOSIT, instr_type.DEPOSIT_INDEX, instr_type.SWAP_FIXOIS, instr_type.SWAP_FIXFLT];				
		QL_REQUIRE(CORE_INT.check_instr_type(param_.short_crv(), v_it, err_name),'invalid short discounting curve (invalid instrument type)', err_name, true);							
	}
	if(check_curve(param_.fra_crv())){
		vector(instr_type) v_it = [instr_type.IR_FUTURE, instr_type.FRA, instr_type.SWAP_FIXOIS, instr_type.SWAP_FIXFLT];
		QL_REQUIRE(CORE_INT.check_instr_type(param_.fra_crv(), v_it,err_name),'invalid middle discounting curve (invalid instrument type)', err_name, true);					
	}

	if(check_curve(param_.swap_crv())){
		vector(instr_type) v_it = [ instr_type.DEPOSIT, instr_type.DEPOSIT_INDEX,instr_type.IR_FUTURE, instr_type.FRA, instr_type.SWAP_FIXOIS, instr_type.SWAP_FIXFLT];
		QL_REQUIRE(CORE_INT.check_instr_type(param_.swap_crv(), v_it,err_name),'invalid swap curve (wrong instrument type)', err_name, true);									
	}
}	

//-------------------------------------------------------
// imp_short_curve
//-------------------------------------------------------
curve swap_curve_disc_ext.imp_short_curve(disc_func df)
{
	if(!check_curve(param_.short_crv()))
		return null<curve>;

	vector(instrument) vi = param_.short_crv().instruments();

	logical ok = CORE_INT.create_instr_from_df( vi, df );
	if(!ok)
		return null<curve>;
	
	return curve(vi,param_.short_crv().name());
}

//-------------------------------------------------------
// imp_fra_curve
//-------------------------------------------------------
curve swap_curve_disc_ext.imp_fra_curve(disc_func df)
{
	if(!check_curve(param_.fra_crv()))
		return null<curve>;

	vector(instrument) vi = param_.fra_crv().instruments();

	logical ok = CORE_INT.create_instr_from_df( vi, df, param_.irfut_cvx_adj() );
	if(!ok)
		return null<curve>;
	
	return curve(vi,param_.fra_crv().name());
}

//-------------------------------------------------------
// imp_swap_curve
//-------------------------------------------------------
curve swap_curve_disc_ext.imp_swap_curve(disc_func df)
{
	if(!check_curve(param_.swap_crv()))
		return null<curve>;
	
	vector(instrument) vi = param_.swap_crv().instruments();

	logical ok = CORE_INT.create_instr_from_df( vi, df);
	if(!ok)
		return null<curve>;

	return curve(vi,param_.swap_crv().name());
}

//-------------------------------------------------------
// imp_blended_curve
//-------------------------------------------------------
curve swap_curve_disc_ext.imp_blended_curve(disc_func df)
{
	if(!curve_ok())
		return null<curve>;

	vector(instrument) vi 	= curve_.instruments();

	logical ok = CORE_INT.create_instr_from_df( vi, df);
	if(!ok)
		return null<curve>;
	
	return curve(vi,curve_.name());
}
//-------------------------------------------------------
// disc_df
//-------------------------------------------------------
disc_func swap_curve_disc_ext.disc_df(	disc_z_model disc_model)							   
{	
	if(!curve_ok())
		return null<disc_func>;

	//if(flt_cal_.is_holiday (param_.d_))
	//	return null<disc_func>;
		
	return disc_model.df(curve_, null<vector(number)>);												
}
//-------------------------------------------------------
// disc_df
//-------------------------------------------------------
disc_func swap_curve_disc_ext.disc_df(disc_func df_base, disc_z_model disc_model)							   
{	
	if(!curve_ok())
		return null<disc_func>;

	//if(flt_cal_.is_holiday (param_.d_))
	//	return null<disc_func>;
		
	return disc_model.df(df_base, curve_, null<vector(number)>);												
}
//-------------------------------------------------------
// create swap_curve_disc_ext
//-------------------------------------------------------
swap_curve_disc_ext swap_curve_disc_ext(swap_curve_disc_parm	s,
										date					d,
										string 					name) 
										
{					
	return new swap_curve_disc_ext(s, d, name);	
}

//-------------------------------------------------------
// instr_bumper_df_vect
//-------------------------------------------------------
vector(disc_func) swap_curve_disc_ext.instr_bumper_df_vect(	disc_z_model 	disc_model,
															number 			y_shift,
															out vector(integer) incl,//input and output, size is equal to number of included instruments
															out disc_func	df_base,
															out disc_func	df_all_shift,
															out vector(instrument) bump_instr,//size is equal to number of included instruments 
															out vector(disc_func) df_v)//size is equal to number of instruments in the curve (=curve_)
{
	
	QL_REQUIRE(curve_ok(),"blended curve not created (disc)");

	return  CORE_INT.instr_bumper_df_vect(curve_,disc_model,y_shift,incl,df_base,df_all_shift,bump_instr,df_v);
}
//-------------------------------------------------------
// sens_matrix_bump
//-------------------------------------------------------
matrix(number) swap_curve_disc_ext.sens_matrix_bump(number 			y_shift,
													number			nominal,
													fwd_func option(nullable) f,//for fixflt
													vector(integer) incl,
													disc_func df_base,
													disc_func	df_all,
													vector(disc_func) df_incl_v
													
													/*out disc_func	df_base,
													out disc_func	df_all,													
													out vector(disc_func) df_all_v*/)
{
	//vector(instrument) instr_bump;
	//disc_func	df_base, df_all;
	//vector(disc_func) df_all_v ;
	//vector(disc_func) df_incl_v = instr_bumper_df_vect(disc_model,y_shift,incl,df_base, df_all, instr_bump,df_all_v);	
	curve imp_blend			= imp_blended_curve(df_base);
	/*if fixflt: implied instrument means we have solved for the implied fixed coupon from a disc_func i.e. npv = 0
	 however note that this is a swap_curve_disc_ext and therefore it is a classic libor swap without a fwd_func
	 normally we will have fixois swaps here
	 */
	QL_REQUIRE(check_curve(imp_blend),"invalid implied blended curve (disc)");
	vector(instrument) vi 	= sub_vector(imp_blend.instruments(), incl) ;	
	return CORE_INT.sens_matrix_instr_disc(nominal,vi,f,y_shift,df_base,df_all,df_incl_v);
}


