/*	
	Basic swap functions for "dual" bootstrap
	Developed by Algorithmica Research, Magnus Nyström		
	
	Comments: see swap_curve_disc_ext

*/

option(null: hard);

/*-----------------------------------------------------------------------
  Name :    class swap_curve_disc_b1_ext
  Notes :   this version uses ois/ibor basis swap curve plus a fix/flt swap curve on the long end
			
			--> creates a synthetic fixed leg vs o/n leg
			if fedfunds: the arithmetic average of fedfunds rates is ignored (i.e. effect of arithmetic avg vs compounding is negligible -->
			probably ok for a 3M period) 

			A fed funds/LIBOR basis swap is a floating-for-floating exchange of (netted) cash flows
			where the fed funds (FF) leg is tied to the fed effective rate and the LIBOR leg to 3-month
			LIBOR. Spreads are quoted against the FF leg and the payments are made quarterly
      
  Changes:
  			Date    		Who     Detail
  			2016-09-05		MN      
  ----------------------------------------------------------------------*/
class swap_curve_disc_b1_ext : public swap_curve_ext
option (category: 'Yield Curve/003 Curve Building')
{
   	public:
		swap_curve_disc_b1_ext(swap_curve_b1_disc_parm, date, logical,string);
		swap_curve_disc_b1_ext(	swap_curve_b1_disc_parm		p,
												swap_curve_fwd_f_parm 	fwd_parm,
												date					d,
												disc_z_model 			disc_model,
												fwd_z_model 			fwd_model,
												logical					extrap_ok,																				
												string 					name);
		swap_curve_disc_b1_ext(swap_curve_disc_b1_ext ) ;
		
		override disc_func 			disc_df(disc_z_model);		
		override swap_curve_type	type();
		override curve				swap_curve();
		
		integer 					ibor_freq() ;		
		day_count_method 			ibor_dc() 	;		
		bd_convention 				ibor_bd() 	;		
		logical 					ibor_eom() ;					
		calendar 					ibor_cal() ;
		fwd_func_tenor ibor_tenor();
		curve						adj_lib_swap_crv();
		curve						lib_swap_crv();
		vector(number) 				b_swap_sprd();
		vector(string) 				b_swap_name();

	private:				
		void				init(logical,logical);
		void 				init_check();
		void 				check_types();
		void				check_fixflt_instr();		
		void				check_b1_curve();
		void 				blend_b1(curve,logical);
		void 				blend_b1_ext(logical);
		cashflows 			cf(curve, vector(number) option(nullable));
		vector(number) 		basis_spread(curve,curve,logical,out curve,out vector(string));
		logical				basis_ibor_curve_match(curve,curve);
	
		cashflows 				cf_;	
		curve  					adj_long_crv_;
		curve					lib_swap_crv_;
		curve  					adj_lib_swap_crv_;
		vector(number)			b_swap_sprd_;
		vector(string)			b_swap_name_;
		swap_curve_b1_disc_parm b1_param_;
		ir_index				fwd_idx_;
		
		//for constructor that allows mismatch between libor-swap crv and ois-libor curve:
		swap_curve_fwd_f_parm 	fwd_parm_;
		disc_z_model 			disc_model_;
		fwd_z_model 			fwd_model_;
};

swap_curve_type		swap_curve_disc_b1_ext.type() 				{ return type_; }
curve				swap_curve_disc_b1_ext.swap_curve() 		{ return adj_long_crv_;}
curve				swap_curve_disc_b1_ext.adj_lib_swap_crv()	{ return adj_lib_swap_crv_;}
curve				swap_curve_disc_b1_ext.lib_swap_crv()		{ return lib_swap_crv_;}
vector(number)		swap_curve_disc_b1_ext.b_swap_sprd()		{ return CORE_INT.clone_vector(b_swap_sprd_);}
vector(string)		swap_curve_disc_b1_ext.b_swap_name()		{ return CORE_INT.clone_vector(b_swap_name_);}

integer				swap_curve_disc_b1_ext.ibor_freq() 			{ return integer(12/fwd_idx_.tenor_months());}
day_count_method	swap_curve_disc_b1_ext.ibor_dc() 			{ return fwd_idx_.day_count_method();}
bd_convention		swap_curve_disc_b1_ext.ibor_bd() 			{ return fwd_idx_.bd_convention();}
logical				swap_curve_disc_b1_ext.ibor_eom() 			{ return fwd_idx_.eom();}
calendar			swap_curve_disc_b1_ext.ibor_cal() 			{ return fwd_idx_.fixing_calendar();}
fwd_func_tenor swap_curve_disc_b1_ext.ibor_tenor()
{
	tenor_code t = fwd_idx_.tenor_code();
	return fwd_func_tenor_from_tc(t);
}

//fwd_func_tenor tenor 	= fwd_func_tenor.FT_3M;

//-------------------------------------------------------
// constructor 
//-------------------------------------------------------
swap_curve_disc_b1_ext.swap_curve_disc_b1_ext(	swap_curve_b1_disc_parm	p,
												date				d,
												logical				extrap_ok,																				
												string 				name) 
												:swap_curve_ext(p,d,name, swap_curve_type.SCE_DISC_B1),
												 fwd_parm_(null<swap_curve_fwd_f_parm>),
												 disc_model_(null<disc_z_model>),
												 fwd_model_(null<fwd_z_model>)
{					
	init(extrap_ok,false);		
}
//-------------------------------------------------------
// constructor for mismatch case
//-------------------------------------------------------
swap_curve_disc_b1_ext.swap_curve_disc_b1_ext(	swap_curve_b1_disc_parm		p,
												swap_curve_fwd_f_parm 	fwd_parm,
												date					d,
												disc_z_model 			disc_model,
												fwd_z_model 			fwd_model,
												logical					extrap_ok,																				
												string 					name) 
												:swap_curve_ext(p,d,name, swap_curve_type.SCE_DISC_B1),
												 fwd_parm_(fwd_parm),
												 disc_model_(disc_model),
												 fwd_model_(fwd_model)
{					
	init(extrap_ok, true);		
}

//-------------------------------------------------------
// copy constructor
//-------------------------------------------------------
swap_curve_disc_b1_ext.swap_curve_disc_b1_ext(swap_curve_disc_b1_ext s) 
										:swap_curve_ext(s), cf_(s.cf_), adj_long_crv_(s.adj_long_crv_),
										 lib_swap_crv_(s.lib_swap_crv_),adj_lib_swap_crv_(s.adj_lib_swap_crv_),
										 b_swap_sprd_(CORE_INT.clone_vector(s.b_swap_sprd_)),
										 b_swap_name_(CORE_INT.clone_vector(s.b_swap_name_)), b1_param_(s.b1_param_),
										 fwd_idx_(s.fwd_idx_),fwd_parm_(s.fwd_parm_),disc_model_(s.disc_model_),fwd_model_(s.fwd_model_)
{}

//-------------------------------------------------------
// init
//-------------------------------------------------------
void swap_curve_disc_b1_ext.init(logical extrap_ok, logical ext)
{
	QL_REQUIRE(param_.type() == swap_curve_type.SCE_DISC_B1, 'invalid type of param object (should be swap_curve_b1_disc_parm)');	
	b1_param_ = dynamic_cast<swap_curve_b1_disc_parm>(param_);

	if(!check_curve(b1_param_.ois_lib_basis_crv()))
		QL_FAIL( 'invalid ois/ibor basis swap curve');

	if(ext) {
		if(check_curve(b1_param_.lib_swap_crv())){
			lib_swap_crv_ = b1_param_.lib_swap_crv();
			fwd_parm_ = new swap_curve_fwd_f_parm(fwd_parm_);
			fwd_parm_.set_long_crv(lib_swap_crv_);
		}
		else if(check_curve(fwd_parm_.long_crv())) {
			lib_swap_crv_ = fwd_parm_.long_crv();			
		}
			
	}
	else {
		lib_swap_crv_ = b1_param_.lib_swap_crv();
	}
	
	if(!check_curve(lib_swap_crv_))
		QL_FAIL( 'invalid fix/flt swap curve');
	
	curve_ext.check();
	check_types();
	init_check();

	logical match = basis_ibor_curve_match(lib_swap_crv(),b1_param_.ois_lib_basis_crv());
	
	if(match){
		blend_b1(lib_swap_crv(),extrap_ok);
	}
	else {
		QL_REQUIRE(!null(fwd_parm_), 'invalid swap_curve_fwd_f_parm (required when mismatch between basis curve and swap curve)');
		QL_REQUIRE(!null(disc_model_), 'invalid disc_z_model (required when mismatch between basis curve and swap curve)');
		QL_REQUIRE(!null(fwd_model_), 'invalid fwd_z_model (required when mismatch between basis curve and swap curve)');
		blend_b1_ext(extrap_ok);		
	}
	
	cf_ = cf( curve_,null<vector(number)>);
	QL_REQUIRE(!null(cf_),'failed to create cashflows');	
}
//-------------------------------------------------------
// check_curves_disc
//-------------------------------------------------------
void swap_curve_disc_b1_ext.check_types()
{
	string err_name;
	
	if(check_curve(param_.short_crv())){
		vector(instr_type) v_it = [instr_type.DEPOSIT, instr_type.DEPOSIT_INDEX, instr_type.SWAP_FIXOIS, instr_type.SWAP_FIXFLT];				
		QL_REQUIRE(CORE_INT.check_instr_type(param_.short_crv(), v_it, err_name),'invalid short discounting curve (invalid instrument type)', err_name, true);							
	}

	if(check_curve(param_.fra_crv())){
		vector(instr_type) v_it = [instr_type.IR_FUTURE, instr_type.FRA, instr_type.SWAP_FIXOIS, instr_type.SWAP_FIXFLT];
		QL_REQUIRE(CORE_INT.check_instr_type(param_.fra_crv(), v_it,err_name),'invalid middle discounting curve (invalid instrument type)', err_name, true);						
	}

	if(check_curve(param_.swap_crv())){
		vector(instr_type) v_it = [ instr_type.DEPOSIT, instr_type.DEPOSIT_INDEX,instr_type.IR_FUTURE, instr_type.FRA, instr_type.SWAP_FIXOIS, instr_type.SWAP_FIXFLT];
		QL_REQUIRE(CORE_INT.check_instr_type(param_.swap_crv(), v_it,err_name),'invalid long discounting curve (invalid instrument type)', err_name, true);					
	}

	QL_REQUIRE(CORE_INT.check_instr_type(b1_param_.ois_lib_basis_crv(), instr_type.SWAP_FLTOIS,err_name),'invalid ois/ibor basis swap curve (mixed curve) (b1 disc)', err_name, true);	
	QL_REQUIRE(CORE_INT.check_instr_type(lib_swap_crv(), instr_type.SWAP_FIXFLT,err_name),'invalid fix/flt swap curve (invalid instrument type)', err_name, true);	
	
}
//-------------------------------------------------------
// check
//-------------------------------------------------------
void swap_curve_disc_b1_ext.init_check()
{	
	vector(instrument) vi 	= lib_swap_crv().instruments();
	swap_fixflt sw 			= swap_fixflt(vi[0]);
	fwd_idx_ 				= sw.ir_index_flt_leg();
			
	check_fixflt_instr();
	check_b1_curve();			
}
//-------------------------------------------------------
// check_fixflt_instr
//-------------------------------------------------------
void swap_curve_disc_b1_ext.check_fixflt_instr()
{
	string err_name;
	
	vector(instrument) vi 	= lib_swap_crv().instruments();
	instrument instr 		= vi[0];
	
	date settle  			= date_code(instr.settle_code()).apply_fwd(d_, instr.calendar());		
	QL_REQUIRE(CORE_INT.check_bswap_issue(vi, settle, err_name ),'invalid ois/ibor basis swap (mixed issue_dates) (b1 disc)', err_name, true);				

	string ccy = b1_param_.ccy();
	QL_REQUIRE(!null(ccy),"invalid currency");
	QL_REQUIRE(CORE_INT.check_eq_currency(vi, ccy, err_name),'invalid fix/flt swap curve (mixed currencies)', err_name, true);					
	
	day_count_method tmp = null<day_count_method>;
	QL_REQUIRE(CORE_INT.swap_dc_check(vi,tmp,true,err_name),'invalid fixed leg daycount_method in fix/flt swap curve (mixed)', err_name, true);
	tmp = null<day_count_method> ;
	QL_REQUIRE(CORE_INT.swap_dc_check(vi, tmp, false,err_name),'invalid float leg daycount_method in fix/flt swap curve (mixed)',err_name,true);

	integer f = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(vi,f,true,err_name),'invalid fixed leg coupon frequency in fix/flt swap curve (mixed)', err_name, true);
	QL_REQUIRE(f > 0 && f != 365,'invalid coupon_freq in fix/flt swap curve',instr.name(),true);
	
	f = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(vi,f,false,err_name),'invalid float leg coupon frequency in fix/flt swap curve (mixed)', err_name, true);
	QL_REQUIRE(f > 0 && f != 365,'invalid flt_coupon_freq in fix/flt swap curve',instr.name(),true);

	f = -1;
	QL_REQUIRE(CORE_INT.swap_rfreq_check(vi,f,false,err_name),'invalid float leg reset frequency in fix/flt swap curve (mixed)', err_name, true);

	tenor_code t = tenor_code(f);
	QL_REQUIRE(!null(t),'invalid tenor');
	QL_REQUIRE(CORE_INT.check_eq_index_tenor_code(vi, t, err_name),'invalid tenor in fix/flt swap curve', err_name, true);

}

//-------------------------------------------------------
// check_b1_curve
//-------------------------------------------------------
void swap_curve_disc_b1_ext.check_b1_curve()
{	
	string err_name;	
	
	curve basis_crv 		= b1_param_.ois_lib_basis_crv();
	vector(instrument) vi 	= basis_crv.instruments();
	instrument bi 			= vi[0];
	
	date settle  			= date_code(bi.settle_code()).apply_fwd(d_, bi.calendar());		
	QL_REQUIRE(CORE_INT.check_bswap_issue(vi, settle, err_name ),'invalid ois/ibor basis swap (mixed issue_dates) (b1 disc)', err_name, true);				

	integer flt_freq_ois = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(vi,flt_freq_ois,false,err_name),'invalid ois/ibor basis swap (mixed coupon frequencies) (b1 disc)', err_name, true);

	integer flt_freq_lib = -1;	
	QL_REQUIRE(CORE_INT.swap_cfreq_check(vi,flt_freq_lib,true,err_name),'invalid ois/ibor basis swap (mixed coupon frequencies) (b1 disc)', err_name, true);
	QL_REQUIRE(flt_freq_lib > 0 && flt_freq_lib != 365,'invalid float coupon freq (ois leg) in ois/ibor basis swap (b1 disc)',bi.name(),true);
	
	tenor_code t = tenor_code(flt_freq_lib);	
	QL_REQUIRE(CORE_INT.swap_index_tenor_code_check(vi, t, true, err_name),'invalid tenor in ois/ibor basis swap (b1 disc)', err_name, true);			

	string ccy = b1_param_.ccy();
	QL_REQUIRE(!null(ccy),"invalid currency");
	QL_REQUIRE(CORE_INT.check_eq_currency(vi, ccy, err_name),'invalid ois/ibor basis swap curve (mixed currencies) (b1 disc)', err_name, true);					
	QL_REQUIRE(check_quote(basis_crv,err_name),'invalid quote(s) in ois/ibor basis curve (b1 disc)',err_name,true);

	day_count_method dc_lib = null<day_count_method> ;	
	QL_REQUIRE(CORE_INT.swap_dc_check(vi, dc_lib, true, err_name),'invalid float daycount method in ois/ibor basis swap curve (mixed) (b1 disc)',err_name,true);

	//compatibility
	QL_REQUIRE(flt_freq_lib == ibor_freq(),'incompatible float coupon freq (mismatch between ois/ibor basis swap and fix/flt swap)', bi.name(), true);				
	QL_REQUIRE(dc_lib == ibor_dc(),'incompatible float daycount method (mismatch between ois/ibor basis swap and fix/flt swap)', bi.name(), true);			
	
}

//-------------------------------------------------------
// basis_ibor_curve_match 
//-------------------------------------------------------
logical	swap_curve_disc_b1_ext.basis_ibor_curve_match(	curve swap_crv,
														curve basis_crv)
{
	vector(instrument) sw_v = swap_crv.instruments();
	integer size_swap = v_size(sw_v);
	vector(instrument) b_v = basis_crv.instruments();
	integer size_fltois = v_size(b_v);

	if(size_fltois != size_swap) 
		return false;
	
	vector(date) s = sw_v.maturity();
	vector(date) o = b_v.maturity();		
				
	for(integer i=0;i<size_fltois;i++){
		if(s[i] != o[i]){
			return false;
		}
	}

	s = sw_v.issue_date();
	o = b_v.issue_date();		
				
	for(integer i=0;i<size_fltois;i++){
		if(s[i] != o[i]){
			return false;
		}
	}
	
	return true;	
}

//-------------------------------------------------------
// blend_b1 only for non-matched curves
//-------------------------------------------------------
void swap_curve_disc_b1_ext.blend_b1_ext(	logical extrap_ok)//fwd_parm contains the libor swap curve
{
	blend_b1(lib_swap_crv(),extrap_ok);
	cf_ 			= cf( curve_,null<vector(number)>);
	QL_REQUIRE(!null(cf_),'failed to create cashflows');
	disc_func df 	= disc_df(disc_model_);
	QL_REQUIRE(!null(df),'failed to create disc_func');
	
	vector(instrument) swap_instr 	= lib_swap_crv().instruments();
	vector(date) s 					= swap_instr.maturity();
	vector(instrument) fltois_instr	= b1_param_.ois_lib_basis_crv().instruments();
	vector(date) o 					= fltois_instr.maturity();
	
	//for now no fwd starting allowed
	//we know that in each curve all start dates are identical
	QL_REQUIRE(swap_instr[0].issue_date() == fltois_instr[0].issue_date(),'currently all swaps must have same start date');
	
	integer size_fltois			= v_size(o);
	vector(integer) found_v 	= v_find(o, s) ;
	vector(integer) idx 		= index_vector(size_fltois);
	vector(integer) not_found_v = v_complement (idx, found_v);
	integer add 				= v_size(not_found_v);

	instr_def fixflt_def_sek 	= swap_instr[0].instr_def();//FIX_ME for ext swap
	error_info err	 			= new error_info(true,true);
	number nominal 				= 100;
	fwd_func_tenor tenor = ibor_tenor();
	
	for(integer k=0;k<2;k++){//"poor mans" version for solving df/fwd interdependence
		
		swap_curve_fwd_ext  fwd_crv 		= swap_curve_fwd_ext(tenor, fwd_parm_, d_, df, "", null<fwd_func>,null<date>);
		fwd_func fwd						= fwd_crv.fwd(fwd_model_);
		QL_REQUIRE(!null(fwd),'failed to create fwd_func');
		vector(instrument) vi_add[add];
				
		for(integer i=0;i<add;i++){
			date mat 		= fltois_instr[not_found_v[i]].maturity();
			swap_fixflt sw 	= swap_fixflt_plain(fixflt_def_sek, strcat("fixflt-synt",str(i+1,0)),d_ ,null<date> ,
												null<string>,mat, null<string>,null<number>,nominal,false,
												df,fwd,null<number>,-1,err);
			QL_REQUIRE(!null(sw),'failed to create synthetic fix-flt swap (used to match the basis curve instrument)');
			vi_add[i] = sw;
		}
		curve lib_swap_crv_add = curve(concat(swap_instr,vi_add));
		fwd_parm_.set_long_crv(lib_swap_crv_add);
		
		blend_b1(lib_swap_crv_add,extrap_ok);
		cf_ 			= cf( curve_,null<vector(number)>);
		QL_REQUIRE(!null(cf_),'failed to create cashflows');
		df 				= disc_df(disc_model_);
		QL_REQUIRE(!null(df),'failed to create disc_func');
	}	
}

//-------------------------------------------------------
// blend_b1 only for matched curves
//-------------------------------------------------------
void swap_curve_disc_b1_ext.blend_b1(curve lib_swap_crv, logical extrap_ok)
{
	b_swap_sprd_ = basis_spread(lib_swap_crv,b1_param_.ois_lib_basis_crv(),extrap_ok,adj_lib_swap_crv_,b_swap_name_);													

	if(!null(param_.swap_crv())){
		date knot;
		adj_long_crv_ = CORE_INT.blend_spot(param_.swap_crv(), adj_lib_swap_crv(), false, knot, param_.blend_buf_days());
	}
	else
		adj_long_crv_ = adj_lib_swap_crv();

	curve_ = CORE_INT.blender_adj_long_curve(	param_.short_crv(), param_.fra_crv() , adj_long_crv_, param_.prio1(), 
									param_.prio2(),param_.merge_fra(), param_.merge_crv(), param_.no_overlap_fwd(),
									param_.blend_buf_days());

	if(null(curve_)) {
		curve_msg_ 	= "null curve (swap_curve_disc_b1_ext-blend)";		
		return;
	}

	curve_ = curve_.set_name(strcat(name_,":blended curve"));
	
	string n;
	logical nq = check_quote(curve_,n);
	
	if(!nq) {
		curve_msg_ = strcat(["null quotes in curve (swap_curve_disc_b1_ext-blend) [",n,"]"]);
		//curve_msg_ 	= "null quotes in curve (swap_curve_disc_b1_ext-blend)";		
		return;
	}

	
	curve_msg_ 	= check_curve(curve_) ? "" : "empty curve (swap_curve_disc_b1_ext-blend)";			
}
//-------------------------------------------------------
// init_bswap
//-------------------------------------------------------
/*void swap_curve_disc_b1_ext.init_bswap(curve basis_crv)
{	
	string n; 
	QL_REQUIRE(check_quote(param_.swap_crv_,n),'invalid quote(s) in swap curve (b2 fwd)',n,true);	
	QL_REQUIRE(check_quote(basis_crv,n),'invalid quote(s) in basis swap curve (b2 fwd)',n,true);
		
		
	b_swap_crv_ = CORE_INT.calc_swap_from_basis2(d_,df_disc_,param_.swap_crv_,basis_crv,add_basis_,"BasisIrs(b2)", x_pol_basis_);
	QL_REQUIRE(check_curve(b_swap_crv_),'invalid swap curve (created from basis curve) (b2 fwd)');
	
	//if(!null(first_flt_fix()))	
	//	b_swap_crv_ = swap_curve_fix_adj(b_swap_crv_);			
}
*/
//-------------------------------------------------------
// basis_spread
//-------------------------------------------------------
vector(number) swap_curve_disc_b1_ext.basis_spread(	curve 		swap_crv,
													curve 		basis_crv,
													logical		extrap_ok,
													out curve 	swap_adj_crv,
													out vector(string) 	basis_name)
{	
	vector(instrument) inst_swap 	= swap_crv.instruments();
	integer size_swap 				= v_size(inst_swap );
	QL_REQUIRE(size_swap > 0,'invalid swap curve');

	vector(instrument) inst_basis 	= basis_crv.instruments();
	integer size_basis 				= v_size(inst_basis );
	QL_REQUIRE(size_basis > 0,'invalid basis curve');
				
	vector(number) ttm_swap 		= inst_swap.maturity() - d_; 
	vector(number) ttm_basis		= inst_basis.maturity() - d_; 
	vector(number) spr_basis		= inst_basis.quote()/10000; 	
	number ttm_max_basis 			= ttm_basis[size_basis-1]; 
	number ttm_min_basis 			= ttm_basis[0]; 
			
	interpolation basis_ip = interpolation(ip_linear(0,0), ttm_basis,spr_basis);

	vector(number) s;	
	vector(logical) r;
	resize(basis_name,0);

	for(integer i = 0; i < size_swap; i++){
		
		if(ttm_swap[i] > ttm_max_basis || ttm_swap[i] < ttm_min_basis){
			if(extrap_ok){
				push_back(s,basis_ip.eval(ttm_swap[i]));
				push_back(r,true);				
				string mc = inst_swap[i].maturity_code();
				push_back(basis_name,strcat(["basis-extrap",mc]));
			}
			else 				
				push_back(r,false);
			
		}
		else {
			push_back(s,basis_ip.eval(ttm_swap[i]));
			push_back(r,true);
			vector(integer) v_f = v_find(ttm_basis, [ttm_swap[i]]);
			if(v_size(v_f) == 1)				
				push_back(basis_name,inst_basis[v_f[0]].name());			
			else {
				string mc = inst_swap[i].maturity_code();
				push_back(basis_name,strcat(["basis-interp",mc]));	 			
			}
		}
	}
	if(v_size(r) > 0) {
		vector(instrument) vi;
		for(integer i = 0; i < size_swap; i++){
			if(r[i])
				push_back(vi, inst_swap[i]);
		}
		swap_adj_crv = curve(vi,swap_crv.name());
	}
	else
		swap_adj_crv = swap_crv;

	return s;

/*	
	interpolation swap_ip = interpolation(ip_linear(0,0), ttm_swap, y_swap);

	for(integer i = 0; i < size_basis; i++){
		if(ttm_basis[i] > ttm_max_swap + 5)	
			break;

		if(ttm_basis[i] < ttm_min_swap - 5)	
			continue;

		number y_swap_ip = swap_ip.eval(ttm_basis[i]);	 						
		if(null(y_swap_ip)){
			CORE_INT.write_warning_message('invalid interpolation of swap yield (at least one instrument)');
			continue;
		}
		number r_sw = adj_ip_spr_basis2(d, df, swp, bswp, y_swap_ip, spr_basis[i]);
		if(null(r_sw)){
			CORE_INT.write_warning_message('invalid conversion of swap yield (at least one instrument)');
			continue;
		}
		instrument ii = CORE_INT.create_par_swap(i_c_n, strcat(inst_basis[i].name(),"-irs"),d,inst_basis[i].settle_date(),
												inst_basis[i].maturity(),r_sw,inst_basis[i].issue_date());
		if(null(ii)){
			CORE_INT.write_warning_message('failed to create interpolated basis-2 swap (at least one instrument)');
			continue;
		}
		push_back(sw, ii);
						
	}
	
	if(!x_pol_basis)
		return curve(sw, name);	
	*/

	/*number cut_off = 15;
	vector(logical) hit;
	hit_vector(ttm_swap, ttm_basis, cut_off, hit);
	
	interpolation basis_ip = interpolation(ip_linear(0,0), ttm_basis,spr_basis);

	for(integer i = 0; i < size_swap; i++){
		if(hit[i]) continue;

		number b_sprd_ip = basis_ip.eval(ttm_swap[i]);	 						
		if(null(b_sprd_ip)){
			CORE_INT.write_warning_message('invalid interpolation of basis-2 spread (at least one instrument)');
			continue;
		}
		number r_sw = adj_ip_spr_basis2(d, df, swp, bswp, y_swap[i], b_sprd_ip);
		if(null(r_sw)){
			CORE_INT.write_warning_message('invalid conversion of swap yield (at least one instrument)');
			continue;
		}

		string sw_name;
		if(ttm_swap[i] < ttm_min_basis || ttm_swap[i] > ttm_max_basis)
			sw_name = strcat(inst_swap[i].name(),"-xpol");
		else
			sw_name = strcat(inst_swap[i].name(),"-ipol");
		
		instrument ii = CORE_INT.create_par_swap(	i_c_n, sw_name,d,inst_swap[i].settle_date(),
												inst_swap[i].maturity(),r_sw,inst_swap[i].issue_date());
		if(null(ii)){
			CORE_INT.write_warning_message('failed to create interpolated basis-2 swap (at least one instrument)');
			continue;
		}
		push_back(sw, ii);
		
	}

	return curve(sw, name);	*/
}

//-------------------------------------------------------
// cf
//-------------------------------------------------------
cashflows swap_curve_disc_b1_ext.cf(curve 			crv,
									vector(number) option(nullable)	y_adj)
{		
	//date settle 			= date_code(settle_code_).apply_fwd(d_, flt_cal_);
	vector(instrument) src	= crv.instruments();
	return CORE_INT.cf_b1(src,adj_lib_swap_crv().instruments(),b_swap_sprd_, y_adj);
	
	/*integer size 			= v_size(src);

	if(null(y_adj)) {
		resize(y_adj,size);
		y_adj = 0;
	}
	else {
		QL_REQUIRE(size == v_size(y_adj),'invalid vector size in cashflow creation (internal error) (b1 disc)');
	}

	vector(cashflows)	dest[0];			
					
	for(integer i = 0; i < size ; i++){
		
		cashflows c;
			
		if(src[i].is_swap_fixflt()){
					
			vector(integer) v_f = v_findi(adj_lib_swap_crv().instruments().name(), [src[i].name()]);
			integer v_f_size = v_size(v_f);				
			QL_REQUIRE(v_f_size < 2,'invalid swap index (internal error)'); 
			if(v_f_size == 1) 					 
				//c = CORE_INT.cf_parswap_sprd(src[i], -b_swap_sprd_[v_f[0]],day(src[i].issue_date()), freq, dc, bd, cal, eom, y_adj[i]);
				c = CORE_INT.cf_parswap_sprd(src[i], -b_swap_sprd_[v_f[0]], y_adj[i]);	
			else 
				c = CORE_INT.cf_parswap(src[i], y_adj[i]);
			
		}
		else if(src[i].is_swap_fixois())
			c = CORE_INT.cf_oisswap(src[i], y_adj[i]);
		else if(src[i].is_fra())
			c = CORE_INT.cf_fra(src[i], y_adj[i]);
		else if(src[i].is_ir_future())
			c = CORE_INT.cf_mmfut(src[i], y_adj[i]);
		else if(src[i].is_deposit() || src[i].is_deposit_index())
			c = CORE_INT.cf_depo(src[i], y_adj[i]);					
		else
			QL_FAIL("invalid instrument type in curve (b1 disc)");
		
		if(null(c)) {
			string msg = strcat([	"<","invalid cashflow (b1 disc)" ,"[", src[i].name(),"]", ">"]);	;
			CORE_INT.write_warning_message( msg);
		}
		else
			push_back(dest, c);
	}

	return cashflows(dest);*/
}
//-------------------------------------------------------
// disc_df
//-------------------------------------------------------
disc_func swap_curve_disc_b1_ext.disc_df(disc_z_model disc_model)							   
{	
	if(!curve_ok())
		return null<disc_func>;
		
	return disc_model.df(cf_, null<vector(number)>);												
}

//-------------------------------------------------------
// create swap_curve_disc_ext
//-------------------------------------------------------
swap_curve_disc_b1_ext swap_curve_disc_b1_ext(	swap_curve_b1_disc_parm	s,
												date					d,
												logical					extrap_ok,	
												string 					name) 
										
{					
	return new swap_curve_disc_b1_ext(s, d, extrap_ok, name);	
}

//-------------------------------------------------------
// create swap_curve_disc_ext
//-------------------------------------------------------
swap_curve_disc_b1_ext swap_curve_disc_b1_ext(	swap_curve_b1_disc_parm	s,
												swap_curve_fwd_f_parm 	f,
												date					d,
												disc_z_model 			disc_model,
												fwd_z_model 			fwd_model,
												logical					extrap_ok,	
												string 					name) 
										
{					
	return new swap_curve_disc_b1_ext(s, f, d, disc_model, fwd_model, extrap_ok, name);	
}
 