/*	
	Basic swap functions for "dual" bootstrap
	Developed by Algorithmica Research, Magnus Nyström
			
	

		swap_curve_fwd_ext:
			used to create a basic swap market forward function (no bonds supported here)
		swap_curve_fwd_f_parm:
			can store 3 curves - short, middle, long
			short curve: 	either a fixing curve or a deposit curve (irrelevant tenors will be removed)
							can contain DEPOSIT, DEPOSIT_INDEX, SWAP_FIXFLT
			middle curve: 	can contain IR_FUTURE, FRA, SWAP_FIXFLT
			long curve:		can contain SWAP_FIXFLT 

	Comments: 	- when using imm-fra use 8 or 12 contracts (see Hamish book p.88)
				- consider using IBOR(3M)/OIS basis swaps to build fwd_func
				- 0Y-3Y: usually 3M because of imm-futures even if long end swap curve is vs 6M-IBOR				

	ToDo: - version with 3M/OIS basis swaps
	
*/

option(null: hard);	

/*-----------------------------------------------------------------------
  Name :    class swap_curve_fwd_ext
  Notes :           
  Changes:
  			Date    		Who     Detail
  			? 				MN      Original code (moved from fiws)
  ----------------------------------------------------------------------*/
class swap_curve_fwd_ext : public swap_curve_ext option (category: 'Yield Curve/003 Curve Building')
{
   	public:
		
		swap_curve_fwd_ext(fwd_func_tenor,swap_curve_fwd_f_parm, date, disc_func option(nullable),
						   string, fwd_func option(nullable), date option(nullable));
	
		swap_curve_fwd_ext(fwd_func_tenor,swap_curve_fwd_f_parm, date, disc_func ,
						   disc_func option(nullable),date,
						   string, fwd_func option(nullable), date option(nullable));
		
		//---legacy-------------------------
		swap_curve_fwd_ext(swap_curve_ext_parm, date, string);
		swap_curve_fwd_ext(swap_curve_ext_parm, date, string, fwd_func, date);

		swap_curve_fwd_ext(swap_curve_ext_parm, date, disc_func option(nullable), string);
		swap_curve_fwd_ext(swap_curve_ext_parm, date, disc_func option(nullable), string, fwd_func, date);	
				
		swap_curve_fwd_ext(	swap_curve_ext_parm, date, disc_func option(nullable), integer,day_count_method,
							bd_convention,calendar,logical,integer, string, string) ;	
		swap_curve_fwd_ext(	swap_curve_ext_parm, date, disc_func option(nullable), integer,day_count_method,
						   	bd_convention,calendar,logical,integer, string, string, fwd_func, date) ;		
		//------------------------------------
		
		swap_curve_fwd_ext(swap_curve_fwd_ext) ;
										
		virtual cashflows_fwd 	cf_fwd();
		virtual disc_func		disc_df();				
		virtual fwd_func		fwd(fwd_z_model);
		virtual fwd_func		fwd_spr(fwd_func,fwd_z_model);	
		virtual fwd_func		fwd_spr(disc_func,fwd_z_model);				
		virtual fwd_func 		fwd_comb(fwd_z_model,fwd_func,date);
		curve					fwd_comb_blended_curve(date );

		virtual curve			imp_short_curve(fwd_func);
		virtual curve			imp_fra_curve(fwd_func, logical = false);
		virtual curve			imp_swap_curve(fwd_func);
		virtual curve			imp_blended_curve(fwd_func, logical = false );
		
		virtual fwd_func		fwd_risk(fwd_z_model, disc_func option(nullable), number, logical, out disc_func );
		
		virtual vector(fwd_func) instr_bumper_ff_vect(fwd_z_model,number,out vector(integer) ,out fwd_func,
													  out fwd_func,out vector(instrument),out vector(instrument) ,out vector(fwd_func) );
		virtual vector(fwd_func) instr_bumper_ff_vect(fwd_z_model,number, vector(string)  option(nullable),out fwd_func,
													  out fwd_func,out vector(instrument),out vector(instrument) ,out vector(fwd_func) );
		virtual fwd_func 		instr_bumper_ff(fwd_z_model,number ,out fwd_func);
			
		virtual vector(fwd_func) wave_bumper_ff_vect(fwd_z_model,number, vector(string)  option(nullable),out fwd_func,out fwd_func,
													 out vector(instrument),out vector(instrument)  );
		virtual vector(fwd_func) wave_bumper_ff_vect(fwd_z_model,number,out vector(integer),out fwd_func,out fwd_func,
													 out vector(instrument),out vector(instrument)  );
		virtual vector(fwd_func) wave_bumper_ff_vect(vector(instrument), fwd_z_model,number,out fwd_func,out fwd_func);
				
		matrix(number) 			sens_matrix_bump(number,number,vector(integer),fwd_func,fwd_func,vector(fwd_func));
		matrix(number) 			sens_matrix_bump(vector(instrument),number,number,fwd_func,fwd_func,vector(fwd_func));
		matrix(number) 			sens_matrix_wave(number,number,vector(integer) ,fwd_func,fwd_func,vector(fwd_func));
		matrix(number) 			sens_matrix_wave(vector(instrument) ,number,number,fwd_func,fwd_func,vector(fwd_func) );
		
		virtual void 			set_disc_func(disc_func);
		disc_func				fwd2disc(fwd_func, interpolator, rate_type);
		virtual number			first_flt_fix(date option(nullable));		
		virtual integer			fix_freq();
		vector(number) 			first_flt_fix_vect(vector(instrument));

		//override curve			short_curve();
		override curve			fra_curve(logical=false);		//in swap_curve_ext
		override curve			swap_curve();		//in swap_curve_ext
		curve					swap_curve(disc_func option(nullable),fwd_func option(nullable),number option(nullable));		//in swap_curve_ext
		override swap_curve_type type();				
		
		override date			curve_end();		//in curve_ext
		date					curve_end(logical);
		
		virtual /*override*/ integer		flt_freq();			
		virtual /*override*/ day_count_method flt_dc() ;		
		virtual /*override*/ bd_convention	flt_bd() ;			
		virtual /*override*/ logical		flt_eom() ;			
		virtual /*override*/ calendar		flt_cal() ;			
		//virtual /*override*/ string 		spot_settle_code() ;	
		virtual /*override*/ integer	 	flt_reset_days(); 	
		virtual /*override*/ date 			flt_reset_date() ;	
		virtual /*override*/ date 			flt_reset_date(date) ;	
		
	protected:

		swap_curve_fwd_ext(swap_curve_ext_parm, date, disc_func option(nullable), string, swap_curve_type,
							fwd_func option(nullable), date option(nullable)); 
					
		virtual curve 			swap_curve_fix_adj(curve);		
		void					check_short_curve_types();
		void					check_fra_curve_types();
		void					check_swap_curve_types();					
		void					init_ir_index();	
		void					check_fra_instr(integer = -1,day_count_method option(nullable) = null<day_count_method>);		
		void					check_swap_instr(integer = -1,day_count_method option(nullable) = null<day_count_method>);
		void 					update_swap_instr();

		virtual cashflows_fwd 	cf_fwd(vector(instrument),vector(number) option(nullable),logical);
		logical					chop_short(date,curve , vector(number) option(nullable),out curve ,out vector(number) );
		curve 					chop_short(curve,vector(date),date );
		integer 				find_chop_short(vector(date),date);
		//logical 				name_in_fix_curve(string name);
		//matrix(number) 			sens_matrix_instr(number,vector(instrument) ,fwd_func,fwd_func,vector(fwd_func),number);
		//(fwd_func) 		wave_bumper_ff_vect_int(cashflows_fwd,fwd_z_model,number ,out fwd_func	,out fwd_func );
		disc_func 				calc_df(fwd_z_model);
		
		disc_func				df_disc_;							
		cashflows_fwd			cf_;	
		fwd_func 				ff_pre_;
		date					cut_off_;
		curve					curve_orig_;//used when we have ff_pre_		
		ir_index				fwd_idx_;
		logical					no_df_1d_;
		
	private:
		
		void 					init(fwd_func_tenor,disc_func option(nullable),date option(nullable));
		void					init_common();
		void					init_check();
		void 					init_legacy();
		void 					init_idx(fwd_func_tenor);
		void 					init_explicit(integer,day_count_method,bd_convention,calendar,logical,integer,string);
		void					blend();			
};



disc_func swap_curve_fwd_ext.fwd2disc(fwd_func fwd, interpolator ip, rate_type ip_rt) 	
{ 
	return CORE_INT.disc_func_fwd_roll(fwd,d_,curve_end(true),true,flt_freq(), flt_eom(),
										flt_cal(),flt_bd(),flt_dc(),ip,ip_rt);
}

curve				swap_curve_fwd_ext.fra_curve(logical incl_synt) { return swap_curve_ext.fra_curve(incl_synt);}
curve				swap_curve_fwd_ext.swap_curve() 				{ return swap_curve_ext.swap_curve();}
void 				swap_curve_fwd_ext.set_disc_func(disc_func df) 	{ df_disc_ = df;}
swap_curve_type		swap_curve_fwd_ext.type() 			{ return type_; }
disc_func 			swap_curve_fwd_ext.disc_df()		{ return df_disc_; }
cashflows_fwd 		swap_curve_fwd_ext.cf_fwd()			{ return cf_; }
integer				swap_curve_fwd_ext.flt_freq()		{ return fwd_idx_.tenor_code().frequency() ;}
day_count_method	swap_curve_fwd_ext.flt_dc() 		{ return fwd_idx_.day_count_method();}
bd_convention		swap_curve_fwd_ext.flt_bd() 		{ return fwd_idx_.bd_convention();}
logical				swap_curve_fwd_ext.flt_eom() 		{ return fwd_idx_.eom();}
calendar			swap_curve_fwd_ext.flt_cal() 		{ return fwd_idx_.fixing_calendar();}
//string 				swap_curve_fwd_ext.spot_settle_code() 	{ return strcat("BD",str(fwd_idx_.spot_days()));}	

//-------------------------------------------------------
// copy constructor 
//-------------------------------------------------------
swap_curve_fwd_ext.swap_curve_fwd_ext(swap_curve_fwd_ext s) 
							:swap_curve_ext(s), df_disc_(s.df_disc_), cf_(s.cf_), ff_pre_(s.ff_pre_), cut_off_(s.cut_off_), 
							 curve_orig_(s.curve_orig_), 
							 fwd_idx_(s.fwd_idx_),no_df_1d_(s.no_df_1d_)
{}

//-------------------------------------------------------
// constructor -protected
//-------------------------------------------------------
swap_curve_fwd_ext.swap_curve_fwd_ext(	swap_curve_ext_parm			p,
										date						d,								
										disc_func option(nullable) 	df,											
										string 						name,							
										swap_curve_type				type,
										fwd_func option(nullable) 	ff_pre, 
										date option(nullable) 		cut_off) 
										:swap_curve_ext(p, d,name, type), 
										 df_disc_(df),cf_(null<cashflows_fwd>), ff_pre_(ff_pre), cut_off_(cut_off),
										 curve_orig_(null<curve>), fwd_idx_(null<ir_index>), no_df_1d_(false) 
{}

//-------------------------------------------------------
// constructor (no disc_func in this constructor), use to get info on blending for example
// [LEGACY version]
//-------------------------------------------------------
swap_curve_fwd_ext.swap_curve_fwd_ext(	swap_curve_ext_parm		p,
										date					d,																												
										string 					name) 
										:swap_curve_ext(p,d, name, swap_curve_type.SCE_FWD), 
										 df_disc_(null<disc_func>), cf_(null<cashflows_fwd>), 
										 ff_pre_(null<fwd_func>), cut_off_(null<date>),
										 curve_orig_(null<curve>), fwd_idx_(null<ir_index>) 
{				
	init_legacy();		
}
//-------------------------------------------------------
// constructor (no disc_func in this constructor), use to get info on blending for example
// [LEGACY version]
//-------------------------------------------------------
swap_curve_fwd_ext.swap_curve_fwd_ext(	swap_curve_ext_parm		p,
										date					d,																												
										string 					name,
										fwd_func 				ff_pre, 
										date					cut_off) 
										:swap_curve_ext(p,d, name, swap_curve_type.SCE_FWD), 
										 df_disc_(null<disc_func>), cf_(null<cashflows_fwd>), 
										 ff_pre_(ff_pre), cut_off_(cut_off),
										curve_orig_(null<curve>), fwd_idx_(null<ir_index>) 
{				
	init_legacy();		
}
//-------------------------------------------------------
// constructor  [LEGACY version]
//-------------------------------------------------------
swap_curve_fwd_ext.swap_curve_fwd_ext(	swap_curve_ext_parm		p,
										date					d,								
										disc_func option(nullable)	df,											
										string 					name) 
										:swap_curve_ext(p,d, name, swap_curve_type.SCE_FWD), 
										 df_disc_(df), cf_(null<cashflows_fwd>),ff_pre_(null<fwd_func>), cut_off_(null<date>),
										 curve_orig_(null<curve>), fwd_idx_(null<ir_index>)
{				
	init_legacy();		
}
//-------------------------------------------------------
// constructor  [LEGACY version]
//-------------------------------------------------------
swap_curve_fwd_ext.swap_curve_fwd_ext(	swap_curve_ext_parm		p,
										date					d,								
										disc_func option(nullable) df,											
										string 					name,
										fwd_func 				ff_pre, 
										date					cut_off) 
										:swap_curve_ext(p,d, name, swap_curve_type.SCE_FWD), 
										 df_disc_(df), cf_(null<cashflows_fwd>),ff_pre_(ff_pre), cut_off_(cut_off),
										 curve_orig_(null<curve>), fwd_idx_(null<ir_index>)
{				
	init_legacy();			
}
//-------------------------------------------------------
// constructor  [LEGACY version]
//-------------------------------------------------------
swap_curve_fwd_ext.swap_curve_fwd_ext(	swap_curve_ext_parm		p,
										date					d,	
										disc_func option(nullable)	df,																		
										integer					flt_freq,
										day_count_method 		flt_dc,
										bd_convention 			flt_bd,
										calendar				flt_cal,
										logical					flt_eom,
										integer					flt_reset_days,
										string					settle_code,
										string 					name) 
										: swap_curve_ext(p, d, name, swap_curve_type.SCE_FWD), 
										  df_disc_(df), cf_(null<cashflows_fwd>),ff_pre_(null<fwd_func>), cut_off_(null<date>), 
										  curve_orig_(null<curve>), fwd_idx_(null<ir_index>)
{		
	init_explicit(flt_freq,flt_dc,flt_bd,flt_cal,flt_eom,flt_reset_days,settle_code);						
}
//-------------------------------------------------------
// constructor  [LEGACY version]
//-------------------------------------------------------
swap_curve_fwd_ext.swap_curve_fwd_ext(	swap_curve_ext_parm		p,
										date					d,	
										disc_func option(nullable)	df,																		
										integer					flt_freq,
										day_count_method 		flt_dc,
										bd_convention 			flt_bd,
										calendar				flt_cal,
										logical					flt_eom,
										integer					flt_reset_days,
										string					settle_code,
										string 					name,
										fwd_func 				ff_pre, 
										date					cut_off) 
										: swap_curve_ext(p, d,name, swap_curve_type.SCE_FWD), 
										  df_disc_(df), cf_(null<cashflows_fwd>),ff_pre_(ff_pre), cut_off_(cut_off), 
										  curve_orig_(null<curve>), fwd_idx_(null<ir_index>)
{		
	init_explicit(flt_freq,flt_dc,flt_bd,flt_cal,flt_eom,flt_reset_days,settle_code);						
}
		
//-------------------------------------------------------
// constructor
//-------------------------------------------------------
swap_curve_fwd_ext.swap_curve_fwd_ext(	fwd_func_tenor 					fwd_tenor,
										swap_curve_fwd_f_parm			p,
										date							d,								
										disc_func 						df,																					
										disc_func option(nullable)		df_synt_base,
										date 							synt_end,
										string 							name,
										fwd_func option(nullable) 		ff_pre, 
										date	option(nullable) 		cut_off) 
										:swap_curve_ext(p,d, name, swap_curve_type.SCE_FWD_F), 
										 df_disc_(df), cf_(null<cashflows_fwd>),ff_pre_(ff_pre), cut_off_(cut_off),
										 curve_orig_(null<curve>), fwd_idx_(null<ir_index>)
{				
	init(fwd_tenor,null(df_synt_base)? df : df_synt_base, synt_end);			
}
//-------------------------------------------------------
// constructor
//-------------------------------------------------------
swap_curve_fwd_ext.swap_curve_fwd_ext(	fwd_func_tenor 					fwd_tenor,
										swap_curve_fwd_f_parm			p,
										date							d,								
										disc_func option(nullable) 		df,											
										string 							name,
										fwd_func option(nullable) 		ff_pre, 
										date	option(nullable) 		cut_off) 
										:swap_curve_ext(p,d, name, swap_curve_type.SCE_FWD_F), 
										 df_disc_(df), cf_(null<cashflows_fwd>),ff_pre_(ff_pre), cut_off_(cut_off),
										 curve_orig_(null<curve>), fwd_idx_(null<ir_index>)
{				
	init(fwd_tenor,null<disc_func>, null<date>);			
}
//-------------------------------------------------------
// init_check
//-------------------------------------------------------
void swap_curve_fwd_ext.init_check()
{
	curve_ext.check();			//ccy check
	check_short_curve_types();	//type check: DEPOSIT, DEPOSIT_INDEX, SWAP_FIXFLT
	check_fra_curve_types();	//type check: IR_FUTURE, FRA, SWAP_FIXFLT
	check_swap_curve_types();	//type check: SWAP_FIXFLT
}
//-------------------------------------------------------
// init
//-------------------------------------------------------
void swap_curve_fwd_ext.init(	fwd_func_tenor 					fwd_tenor,
								disc_func option(nullable)		df_synt_base,
								date 	option(nullable)		synt_end )
{
	QL_REQUIRE(param_.type() == swap_curve_type.SCE_FWD_F,'invalid type of param object (should be swap_curve_fwd_f_parm)');	

	no_df_1d_ = null(df_disc_) && fwd_tenor == fwd_func_tenor.FT_1D;
	
	init_check();
	init_idx(fwd_tenor);//create fwd_idx_ and filter the short curve to include only 1 tenor

	check_fra_instr();//check tenors and dc's
	check_swap_instr();//check freqs, tenors, dc's, anc class
	update_swap_instr();//update df,fixing and clear tenor_surface
	
	logical incl_synt = !null(df_synt_base) && !null(synt_end);
	if(incl_synt) {
		swap_curve_fwd_f_parm param = dynamic_cast<swap_curve_fwd_f_parm>(param_);
		
		curve	df_basis_crv;
		disc_func_spr df_basis = CORE_INT.df_tenor_spread(	d_, df_synt_base, synt_end, param.short_crv(), 
															param.fra_crv(), param.swap_crv(),param.prio1(),param.prio2(),
															param.merge_fra(),param.blend_buf_days(), df_basis_crv);
		QL_REQUIRE(!null(df_basis), "failed to create basis spread function");

		
		synt_crv_ = CORE_INT.synt_curve_create(d_,param.fra_crv(), fwd_idx_, df_basis, df_basis_crv);

		param.extend_fra_crv(synt_crv_);
	}
	
	init_common();	//blend and create cashflows_fwd
}
//-------------------------------------------------------
// init_common
//-------------------------------------------------------
void swap_curve_fwd_ext.init_common()
{
	
	blend();					
		
	if(!null(ff_pre_)){	
		QL_REQUIRE(!null(cut_off_) && cut_off_ > d_,'invalid cut off date (fwd)');
		vector(number) y;
		logical ok = chop_short(cut_off_, curve_orig_, null<vector(number)>, curve_, y);	
		QL_REQUIRE(ok && check_curve(curve_),'invalid blended fwd curve (chopped curve) (fwd)');					
	}

	if(curve_ok()) {
		cf_ = cf_fwd(curve_.instruments(),null<vector(number)>,false);
		QL_REQUIRE(!null(cf_),'failed to create cashflows_fwd (fwd)');
	}
	else {
		cf_ = null<cashflows_fwd>;
	}
}
//-------------------------------------------------------
// init_idx		(curr)
//-------------------------------------------------------
void swap_curve_fwd_ext.init_idx(fwd_func_tenor fwd_tenor)
{
	swap_curve_fwd_f_parm param = dynamic_cast<swap_curve_fwd_f_parm>(param_);		
	
	if(!param.short_is_fix_crv()) {
		//deposit_index, deposit or fix-flt
		curve new_short_crv;
		fwd_idx_ = null<ir_index>;		
		logical upd = CORE_INT.fwd_init_depo(param.short_crv(), fwd_tenor, new_short_crv);		
		if(upd)
			param.set_short_crv(new_short_crv);
	}
	else {//only deposit_index
		
		string fwd_tenor_s = string(fwd_tenor);
		fixing_curve new_fixing_crv;
		logical upd = CORE_INT.fwd_init(d_, param.fixing_crv(), fwd_tenor, fwd_tenor_s,fwd_idx_,new_fixing_crv);
		if(upd) 
			param.set_fixing_crv(new_fixing_crv); 			
	}
	
	if(null(fwd_idx_)){
		if(check_curve(param.fra_crv())) {
			//fra, ir-fut or fix-flt
			instrument instr = param.fra_crv().instruments()[0];
			if(instr.instr_type() == instr_type.SWAP_FIXFLT){
				swap_fixflt sw = swap_fixflt(instr);
				fwd_idx_ = sw.ir_index_flt_leg();
			}
			else {
				fwd_idx_ = instr.ir_index();
			}
		}
		else if(check_curve(param.swap_crv())) {
			//only fix-flt
			instrument instr = param.swap_crv().instruments()[0];
			if(instr.instr_type() == instr_type.SWAP_FIXFLT){
				swap_fixflt sw = swap_fixflt(param.swap_crv().instruments()[0]);
				fwd_idx_ = sw.ir_index_flt_leg();
			}
			else if(instr.instr_type() == instr_type.SWAP_FIXOIS){
				swap_fixois sw = swap_fixois(param.swap_crv().instruments()[0]);
				fwd_idx_ = sw.ir_index_ois_leg();
			}
		}
	}		
}
//-------------------------------------------------------
// init_legacy  [LEGACY]
//-------------------------------------------------------
void swap_curve_fwd_ext.init_legacy()
{
	QL_REQUIRE(param_.type() == swap_curve_type.SCE_FWD,'invalid type of param object (should be swap_curve_fwd_parm)');		
	QL_REQUIRE(check_curve(param_.swap_crv()),'invalid swap curve (fwd)'); //for this constructor the swap curve is required		

	init_check();			
	init_ir_index();
	no_df_1d_ = null(df_disc_) && flt_freq() == 365;
	
	check_fra_instr();
	check_swap_instr();
	update_swap_instr();
	init_common();		
}
//-------------------------------------------------------
// init_explicit  [LEGACY]
//-------------------------------------------------------
void swap_curve_fwd_ext.init_explicit(	integer					flt_freq,
										day_count_method 		flt_dc,
										bd_convention 			flt_bd,
										calendar				flt_cal,
										logical					flt_eom,
										integer					flt_reset_days,
										string					settle_code)
{
	QL_REQUIRE(param_.type() == swap_curve_type.SCE_FWD,'invalid type of param object (should be swap_curve_fwd_parm)');

	logical sw_ok = check_curve(param_.swap_crv());
	logical fr_ok = check_curve(param_.fra_crv());
	QL_REQUIRE(sw_ok || fr_ok,'at least one of swap and fra/fut/sps curve is required (fwd)');		

	init_check();		

	string curr			= param_.ccy();
	integer spot_days 	= flt_reset_days;
	integer value_days 	= spot_days;
	string tenor 		= strcat(str(12/flt_freq),"M");

	fwd_idx_ = ir_index_ibor(strcat("Ibor-",curr),tenor,value_days,flt_reset_days, curr,
							flt_cal,"n/a", flt_dc,flt_bd,flt_eom,spot_days);			
	QL_REQUIRE(!null(fwd_idx_),'failed to create ir_index (fwd)');

	no_df_1d_ = null(df_disc_) && flt_freq == 365;
	
	check_fra_instr(flt_freq, flt_dc);
	check_swap_instr(flt_freq, flt_dc);
	update_swap_instr();
	init_common();	
}

//-------------------------------------------------------
// init_ir_index   [LEGACY]
//-------------------------------------------------------
void swap_curve_fwd_ext.init_ir_index()
{		
	if(!check_curve(param_.swap_crv()))
		return;
	
	vector(instrument) v_i = param_.swap_crv().instruments();
	
	integer freq = v_i[0].swap_reset_freq_leg2();	
	QL_REQUIRE(freq  > 0 && freq  != 365,'invalid float reset freq in swap curve (fwd)',v_i[0].name(),true);	
	 
	day_count_method dc = v_i[0].swap_dc_method_leg2() ;	
	QL_REQUIRE(!null(dc),'invalid float daycount method in swap curve (fwd)',v_i[0].name(),true);			
	
	bd_convention bd	= v_i[0].pmt_bus_day_d(BD_MOD_FOLLOWING);	
	QL_REQUIRE(!null(bd),'invalid pmt_bus_day in swap curve (fwd)',v_i[0].name(),true);	 	
	
	logical eom	= v_i[0].eom_d(true);				
	
	calendar cal 	= v_i[0].calendar(); // assume fixleg cal == flt leg cal
	string cal_name = v_i[0].calendar_name();
	QL_REQUIRE(!null(cal),'invalid calendar in swap curve (fwd)',v_i[0].name(),true);	
	
	integer rdays = integer(v_i[0].flt_reset_days_d(2));
	QL_REQUIRE(rdays >= 0,'invalid flt_reset_days in swap curve (fwd)',v_i[0].name(),true);	

	string curr			= param_.ccy();
	integer spot_days 	= rdays;
	integer value_days 	= spot_days;
	string tenor 		= strcat(str(12/freq),"M");

	fwd_idx_ = ir_index_ibor(strcat("Ibor-",curr),tenor,value_days,rdays, curr,cal,cal_name,dc,bd,eom,spot_days);			
	QL_REQUIRE(!null(fwd_idx_),'failed to create ir_index (fwd)');

}
//-------------------------------------------------------
// check_short_curve_types
//-------------------------------------------------------
void swap_curve_fwd_ext.check_short_curve_types()
{	
	if(check_curve(param_.short_crv())){
		string err_name;
		vector(instr_type) v_it = [instr_type.DEPOSIT, instr_type.DEPOSIT_INDEX,instr_type.SWAP_FIXFLT];		
		QL_REQUIRE(CORE_INT.check_instr_type(param_.short_crv(), v_it, err_name),'invalid short curve (wrong instrument type)', err_name, true);			
	}
}
//-------------------------------------------------------
// check_fra_curve_types
//-------------------------------------------------------
void swap_curve_fwd_ext.check_fra_curve_types()
{
	if(check_curve(param_.fra_crv())){
		string err_name;
		vector(instr_type) v_it = [instr_type.IR_FUTURE, instr_type.FRA, instr_type.SWAP_FIXFLT];
		QL_REQUIRE(CORE_INT.check_instr_type(param_.fra_crv(), v_it, err_name),'invalid fra/fut/sps curve (wrong instrument type)', err_name, true);						
	}
}
//-------------------------------------------------------
// check_fra_instr
//-------------------------------------------------------
void swap_curve_fwd_ext.check_fra_instr(integer freq, day_count_method option(nullable) dc)
{		
	if(!check_curve(param_.fra_crv()))
		return;

	if( freq == -1)
		freq = flt_freq();
	if(null(dc))
		dc = flt_dc();
	
	string err_name;
	vector(instrument) v_i = param_.fra_crv().instruments();		
	tenor_code t = tenor_code(freq);
	QL_REQUIRE(!null(t),'invalid tenor');
	QL_REQUIRE(CORE_INT.check_eq_index_tenor_code(v_i, t, err_name),'invalid tenor in fra/fut curve (mixed)', err_name, true);								
	QL_REQUIRE(CORE_INT.check_eq_dc_method(v_i, dc, err_name),'invalid daycount method in fra/fut curve (mixed)',err_name,true);						
}
//-------------------------------------------------------
// check_swap_curve_types
//-------------------------------------------------------
void swap_curve_fwd_ext.check_swap_curve_types()
{
	if(check_curve(param_.swap_crv())){
		string err_name;
		//do not change to allow more types (swap_curve_disc_b1_ext depends on this)
		QL_REQUIRE(CORE_INT.check_instr_type(param_.swap_crv(), [instr_type.SWAP_FIXFLT,instr_type.SWAP_FIXOIS],err_name),'invalid swap curve (wrong instrument type)', err_name, true);									
	}
}
//-------------------------------------------------------
// check_swap_instr
//-------------------------------------------------------
void swap_curve_fwd_ext.check_swap_instr(integer 			freq,
										   day_count_method option(nullable) dc)
{		
	if(!check_curve(param_.swap_crv()))
		return;

	if( freq == -1)
		freq = flt_freq();
	if(null(dc))
		dc = flt_dc();
	
	string err_name;
	vector(instrument) v_i = param_.swap_crv().instruments();

	string cl = null<string>;
	QL_REQUIRE(CORE_INT.check_eq_class(v_i,cl,err_name),'invalid swap curve (mixed class names)', err_name, true);	

	day_count_method tmp = null<day_count_method>;
	QL_REQUIRE(CORE_INT.swap_dc_check(v_i,tmp,true,err_name),'invalid fixed leg daycount_method in swap curve (mixed)', err_name, true);
	QL_REQUIRE(CORE_INT.swap_dc_check(v_i, dc, false,err_name),'invalid float leg daycount_method in swap curve (mixed)',err_name,true);

	integer f = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(v_i,f,true,err_name),'invalid fixed leg coupon frequency in swap curve (mixed)', err_name, true);
	f = -1;
	QL_REQUIRE(CORE_INT.swap_cfreq_check(v_i,f,false,err_name),'invalid float leg coupon frequency in swap curve (mixed)', err_name, true);	
	QL_REQUIRE(CORE_INT.swap_rfreq_check(v_i,freq,false,err_name),'invalid float leg reset frequency in swap curve (mixed)', err_name, true);
	
	tenor_code t = tenor_code(freq);
	QL_REQUIRE(!null(t),'invalid tenor');
	QL_REQUIRE(CORE_INT.check_eq_index_tenor_code(v_i, t, err_name),'invalid tenor in swap curve (mixed)', err_name, true);										
						
}

//-------------------------------------------------------
// update_swap_instr
//-------------------------------------------------------
void swap_curve_fwd_ext.update_swap_instr()
{
	if(!check_curve(param_.swap_crv()))
		return;
	
	vector(instrument) vi 	= param_.swap_crv().instruments();
	integer size 			= v_size(vi);
	error_info error 		= new error_info(true,false);
	
	for (integer i = 0 ; i < size ; i++) {

		
		
		if(vi[i].instr_type(error) == ..instr_type.SWAP_FIXFLT) {
		
			//logical is_fixflt = vi[i].is_swap_fixflt(error);
			//if(error.is_error())
				//continue;					
			if(null(df_disc_))
				continue;
			number first_fix 	= first_flt_fix(vi[i].settle_date());
			swap_fixflt sw 		= swap_fixflt(vi[i]);

			//try {
				sw.add_disc_func(df_disc_);
			//}
			//catch {
				//continue;
			//}
			if(!null(first_fix))
				sw.add_current_fixing(first_fix);

			sw.add_surface(null<tenor_surface>);
			
			vi[i] = sw;
		}
		else if(vi[i].instr_type(error) == ..instr_type.SWAP_FIXOIS) {
			//if(error.is_error())
				//continue;
			if(null(df_disc_))
				continue;
			swap_fixois sw 		= swap_fixois(vi[i]);

			//try {
				sw.add_disc_func(df_disc_);
			//}
			//catch {
			//	continue;
			//}
			//if(!null(first_fix))
				//sw.add_current_fixing(first_fix);

			sw.add_surface(null<tenor_surface>);
			
			vi[i] = sw;
		}
		
			
	}
	param_.set_swap_crv(curve(vi,param_.swap_crv().name()));
	
	return ;
}

//-------------------------------------------------------
// blend
//-------------------------------------------------------
void swap_curve_fwd_ext.blend()
{
	curve short_adj_c = param_.short_crv();
		
	if(type_ == swap_curve_type.SCE_FWD ){	
		if(!null(param_.short_crv()) && flt_freq() != 365) 	
			short_adj_c = CORE_INT_COMPAT.extract_tenor_depo(param_.short_crv(),12/flt_freq(),flt_cal(),flt_eom(),flt_bd());
	}	

	curve swap_adj_c ;
	if(!null(param_.swap_crv()) && !null(first_flt_fix(null<date>)))		
		swap_adj_c = swap_curve_fix_adj(param_.swap_crv());
	else
		swap_adj_c = param_.swap_crv();
	
	curve_ = curve_orig_ = CORE_INT.blender_adj_long_curve(	short_adj_c, param_.fra_crv() , swap_adj_c, param_.prio1(), 
													param_.prio2(),param_.merge_fra(), param_.merge_crv(), param_.no_overlap_fwd(),
													param_.blend_buf_days());
	
	if(null(curve_)) {
		curve_msg_ 	= "null curve (swap_curve_fwd_ext-blend)";		
		return;
	}

	curve_ = curve_orig_ = curve_.set_name(strcat(name_,":blended curve"));	
	
	string n;
	logical nq = check_quote(curve_,n);

	if(!nq) {
		curve_msg_ = strcat(["null quotes in curve (swap_curve_fwd_ext-blend) [",n,"]"]);
		//curve_msg_ 	= "null quotes in curve (swap_curve_fwd_ext-blend)";		
		return;
	}
		
	curve_msg_ 	= check_curve(curve_) ? "" : "empty curve (swap_curve_fwd_ext-blend)";		
}
//-------------------------------------------------------
// flt_reset_days
//-------------------------------------------------------
integer	 swap_curve_fwd_ext.flt_reset_days() 
{ 
	integer fd = fwd_idx_.fixing_days();
	QL_REQUIRE(fd >= 0,'invalid flt_reset_days in ir_index (days < 0 not supported) (fwd)');
	return fd;
}
//-------------------------------------------------------
// flt_reset_date
//-------------------------------------------------------
date swap_curve_fwd_ext.flt_reset_date() 
{ 
	date r = d_;
	while(!fwd_idx_.is_valid_fixing_date(r)) 
		r = flt_cal().move_bus_days(r, 1);
	return r;
}
//-------------------------------------------------------
// flt_reset_date
//-------------------------------------------------------
date swap_curve_fwd_ext.flt_reset_date(date settle) 
{ 
	return fwd_idx_.fixing_date(settle);
}
//-------------------------------------------------------
// first_flt_fix  (only for spot starting swaps)
//-------------------------------------------------------
number swap_curve_fwd_ext.first_flt_fix(date option(nullable) settle)
{
	date rd = null(settle) ? flt_reset_date() : flt_reset_date(settle);
	
	if(null(rd) || rd > d_ )
		return null<number>;
	
	if(type() == swap_curve_type.SCE_FWD_F){
		swap_curve_fwd_f_parm param = dynamic_cast<swap_curve_fwd_f_parm>(param_);
		string 	err_msg;
		if(null(param.fixing_crv()))
			return check_curve(param_.swap_crv()) ? param_.swap_first_fix() : null<number>;
		else
			return param.fixing_crv().fixing_rate(d_, fwd_idx_.tenor(), err_msg);
	}
	else if(type_ == swap_curve_type.SCE_FWD){
		return check_curve(param_.swap_crv()) ? param_.swap_first_fix() : null<number>;
	}
	else
		QL_FAIL("invalid function call");
}
//-------------------------------------------------------
// first_flt_fix_vect
//-------------------------------------------------------
vector(number) swap_curve_fwd_ext.first_flt_fix_vect(vector(instrument) vi)
{
	integer size = v_size(vi);
	vector(number) first_fix[size];
	for (integer i = 0 ; i < size ; i++) {
		if(vi[i].is_swap_fixflt()) 
			first_fix[i] = 	first_flt_fix(vi[i].settle_date());
		else
			first_fix[i] = 	null<number>;
	}
	return first_fix;
}

//-------------------------------------------------------
// swap_curve_fix_adj
//-------------------------------------------------------
curve swap_curve_fwd_ext.swap_curve_fix_adj(curve swap_crv)
{
	vector(instrument) src	= swap_crv.instruments();
	integer size_o = v_size(src);
	integer remove_i = -1;

	for(integer i = 0; i < size_o; i++){					
		if(src[i].is_swap_fixflt() || src[i].is_swap_fltflt2s()|| src[i].is_swap_fltflt()){
			vector(date) pmt_dates_beg, pmt_dates_end,idx_dates_end;
			logical ip_idx_stub = false;
			date start 			= src[i].issue_date();
			CORE_INT.swap_flt_dates(start,null<date>,null<date>, src[i].maturity(),day(start),
									flt_freq(),flt_bd(),flt_cal(), flt_eom(),flt_bd(),flt_cal(),
									flt_eom(),ip_idx_stub,pmt_dates_beg,pmt_dates_end,idx_dates_end);

			if(v_size(pmt_dates_beg) == 1)
				remove_i = i;
			
			break;
		}
	}

	if(remove_i != -1) {
		src = v_exclude (src, remove_i);
		return curve(src, swap_crv.name());
	}
	else
		return swap_crv;
}

//-------------------------------------------------------
// fix_freq
// for the fixed leg in the swap
//-------------------------------------------------------
integer	swap_curve_fwd_ext.fix_freq()
{ 		
	QL_REQUIRE(check_curve(param_.swap_crv()),'swap curve not available');
	return integer(param_.swap_crv().instruments()[0].coupon_freq());	
}
//-------------------------------------------------------
// curve_end  Note: returns the start date of the fwd period
//-------------------------------------------------------
date swap_curve_fwd_ext.curve_end()
{	
	if(!curve_ok())
		return null<date>;

	date m ;
	
	for(integer i = 0; i< cf_.n_instruments(); i++){
		vector(number) fixed_amounts,nominals; 
		vector(date) fixed_dates,disc_dates, fwd_dates;
		logical f = cf_.get_cashflows(i, fixed_dates, fixed_amounts, fwd_dates, disc_dates,nominals);
		if(!f) return null<date>;
		integer size = v_size(fwd_dates);//size==1 for single period instruments
		if (i==0)
			m = fwd_dates[size-1] ;
		else if(fwd_dates[size-1] > m)
			m = fwd_dates[size-1] ;		
	}
	return m;
	
}

//-------------------------------------------------------
// curve_end 
//-------------------------------------------------------
date swap_curve_fwd_ext.curve_end(logical fwd_end)
{
	date s = curve_end();
	if(!fwd_end || null(s))
		return s;
	
	return fwd_idx_.maturity(s);

}
//-------------------------------------------------------
// chop_short
//-------------------------------------------------------
curve swap_curve_fwd_ext.chop_short(curve 			c,
									vector(date) 	f, 
									date 			d)
{
	vector(instrument) v	= c.instruments();
    integer	n				= v_size(v);
    integer	chop			= swap_curve_fwd_ext.find_chop_short(f, d);   
    v = sub_vector(v, chop, n - chop);    
    return curve(v, c.name());
}
//-------------------------------------------------------
// chop_short
//-------------------------------------------------------
integer swap_curve_fwd_ext.find_chop_short(	vector(date) 	f, 
											date 			d)
 {
	integer	n = v_size(f);

	for (integer i = 0 ; i < n ; i++){
	    if (f[i] >= d)
			return i;
	}
	return n;
 }

//-------------------------------------------------------
// chop_short
//-------------------------------------------------------
logical swap_curve_fwd_ext.chop_short(	date 						cut_off,
										curve						orig_curve,
										vector(number) option(nullable) orig_y_adj,//currently not used
										out curve  					cut_curve,
										out vector(number) 			y_adj)//currently not used
{
	if(!curve_ok())
		return false;

	vector(instrument) src	= orig_curve.instruments();
	integer size_o = v_size(src);
	vector(date) f, dates;
	
	for(integer i = 0; i < size_o; i++){					
		if(src[i].is_swap_fixflt() || src[i].is_swap_fltflt2s()|| src[i].is_swap_fltflt()){
			vector(date) float_idx0,float_pmt;
			date isd = src[i].issue_date();
			CORE_INT.swap_pmt_dates(isd,null<date>,null<date>, src[i].maturity(),day(isd), 
									flt_freq(),	flt_bd(),flt_cal(),flt_eom(),dates);
			CORE_INT.dates_1v_to_2v(dates,float_idx0,float_pmt);
			date last_fix = float_idx0[v_size(float_idx0)-1];
			push_back(f, last_fix-1);	
		}
		else if(src[i].is_deposit() || src[i].is_ir_future() || src[i].is_fra())
			push_back(f, src[i].settle_date()-1);			
		else
			QL_FAIL("invalid instrument type in curve (fwd)");

		if(f[v_size(f)-1] > cut_off)
			break;
	}
 
	cut_curve = swap_curve_fwd_ext.chop_short(orig_curve,f,cut_off);
	
	integer size_n = v_size(cut_curve.instruments());
	if(size_o == size_n){
		y_adj = !null(orig_y_adj) ? clone_vector(orig_y_adj) : null<vector(number)>;
		return true;
	}
	else if(size_n == 0)
		return false;
	else {			
		if(null(orig_y_adj) || v_size(orig_y_adj) == 0){
			y_adj = null<vector(number)>;
			return true;
		}
		else {
			resize(y_adj,size_n); 
			for(integer i = 0; i < size_n; i++)
				y_adj[size_n-i-1] = orig_y_adj[size_o-i-1];				
		}
	}
	return true;		
}

//-------------------------------------------------------
// imp_short_curve
//-------------------------------------------------------
curve swap_curve_fwd_ext.imp_short_curve(fwd_func f_flt_idx)
{
	if(!check_curve(param_.short_crv()))
		return null<curve>;

	vector(instrument) vi = param_.short_crv().instruments();
	integer size = v_size(vi);
	
	QL_REQUIRE(size == 1, "invalid short curve");
	
	if(!vi[0].is_depo() && !vi[0].is_deposit_index())
		return null<curve>;

	vi[0] = CORE_INT.create_depo_from_fwd(vi[0], f_flt_idx);
		
	return curve(vi,param_.short_crv().name());
}

//-------------------------------------------------------
// imp_fra_curve
//-------------------------------------------------------
curve swap_curve_fwd_ext.imp_fra_curve(fwd_func f_flt_idx,
									   logical 	incl_synt)
{
	if(!check_curve(param_.fra_crv()))
		return null<curve>;

	vector(instrument) vi = param_.fra_crv().instruments();
	integer size = v_size(vi);

	vector(number) adj = param_.irfut_cvx_adj();
	if(null(adj) || v_size(adj) != size){
		resize(adj,size);
		adj = 0;
	}
	
	if(!incl_synt && check_curve(synt_crv_)) {
		vector(instrument) vsi 	= synt_crv_.instruments();
		vector(integer) vf 		= v_findi(vi.name(),vsi.name());
		vi 						= v_exclude (vi, vf);
		adj 					= v_exclude (adj, vf);
		size 					= v_size(vi);
	}

	vector(number) first_fix = first_flt_fix_vect(vi);		
	
	logical ok = CORE_INT.create_instr_from_fwd(vi, f_flt_idx,df_disc_, first_fix ,adj);
	if(!ok)
		return null<curve>;
	
	return curve(vi,param_.fra_crv().name());
}

//-------------------------------------------------------
// swap_curve
//-------------------------------------------------------
curve swap_curve_fwd_ext.swap_curve(disc_func option(nullable) df,
									fwd_func option(nullable) f_flt_idx,
									number option(nullable) first_fix)//updates only if not null
{
	if(!check_curve(param_.swap_crv()))
		return null<curve>;

	if(null(df) && null(f_flt_idx) && null(first_fix))
		return swap_curve();

	vector(instrument) vi 	= param_.swap_crv().instruments();
	integer size 			= v_size(vi);
	error_info error 		= new error_info(true,false);
	
	for (integer i = 0 ; i < size ; i++) {

		if(vi[i].instr_type(error) == ..instr_type.SWAP_FIXFLT){
			//logical is_fixflt = vi[i].is_swap_fixflt(error);
			if(error.is_error() )
				return null<curve>;					

			swap_fixflt sw = swap_fixflt(vi[i]);
			
			if(!null(df))					
				sw.add_disc_func(df);
			else
				df = sw.disc_func_flt_leg();
			
			if(!null(first_fix))
				sw.add_current_fixing(first_fix);
			
			if(!null(f_flt_idx)){
				QL_REQUIRE(!null(df),"missing disc_func (required when fwd_func is input)");
				tenor_surface ts = sw.create_tenor_surface( df,f_flt_idx);
				sw.add_surface(ts);
			}
				
			vi[i] = sw;
		}
		else if(vi[i].instr_type(error) == ..instr_type.SWAP_FIXOIS){
			
			if(error.is_error() )
				return null<curve>;

			swap_fixois sw = swap_fixois(vi[i]);
			
			if(!null(df))					
				sw.add_disc_func(df);
			else
				df = sw.disc_func_ois_leg();
			
			//if(!null(first_fix))
				//sw.add_current_fixing(first_fix);
			
			if(!null(f_flt_idx)){
				QL_REQUIRE(!null(df),"missing disc_func (required when fwd_func is input)");
				tenor_surface ts = sw.create_tenor_surface( df,f_flt_idx);
				sw.add_surface(ts);
			}
				
			vi[i] = sw;
		}
		else {
			return null<curve>;
		}
		
	}
	
	return curve(vi,param_.swap_crv().name());
}

//-------------------------------------------------------
// imp_swap_curve
//-------------------------------------------------------
curve swap_curve_fwd_ext.imp_swap_curve(fwd_func f_flt_idx)
{
	if(!check_curve(param_.swap_crv()))
		return null<curve>;
	
	vector(instrument) vi = param_.swap_crv().instruments();
	integer size = v_size(vi);
	
	for (integer i = 0 ; i < size ; i++) {

		
		if(vi[i].is_swap_fixflt()) {
			
			number first_fix = 	first_flt_fix(vi[i].settle_date());
		
			vi[i] = CORE_INT.create_fixflt_from_fwd(vi[i],df_disc_,f_flt_idx, first_fix);
		}
		else if(vi[i].is_swap_fixois()) {
			number first_fix = null<number>;
			vi[i] = CORE_INT.create_fixois_from_fwd(vi[i],df_disc_,f_flt_idx,first_fix);
		}
		else {
			return null<curve>	;
		}
	}
	
	return curve(vi,param_.swap_crv().name());
}


//-------------------------------------------------------
// imp_blended_curve
//-------------------------------------------------------
curve swap_curve_fwd_ext.imp_blended_curve(fwd_func f_flt_idx,
										   logical incl_synt)
{
	if(!curve_ok())
		return null<curve>;

	vector(instrument) vi 	= curve_.instruments();
	integer size 			= v_size(vi);

	vector(number) adj = param_.irfut_cvx_adj();
	if(null(adj) || v_size(adj) != size){
		resize(adj,size);
		adj = 0;
	}

	if(!incl_synt && check_curve(synt_crv_)) {
		vector(instrument) vsi 	= synt_crv_.instruments();
		vector(integer) vf 		= v_findi(vi.name(),vsi.name());
		vi 						= v_exclude (vi, vf);
		adj 					= v_exclude (adj, vf);
		size 					= v_size(vi);
	}

	vector(number) first_fix = first_flt_fix_vect(vi);	
	
	logical ok = CORE_INT.create_instr_from_fwd(vi, f_flt_idx,df_disc_, first_fix ,adj);
	if(!ok)
		return null<curve>;
	
	return curve(vi,curve_.name());
}
//-------------------------------------------------------
// cf_fwd
//-------------------------------------------------------
/*zero spread always and throws if not a complete cashflows_fwd (used in risk calc)*/
cashflows_fwd swap_curve_fwd_ext.cf_fwd(vector(instrument) src,
										vector(number) option(nullable)	y_adj,
										logical require_all_cf )
{						
	vector(number) first_fix = first_flt_fix_vect(src);
	return CORE_INT.cf_fwd(src,first_fix,y_adj,require_all_cf);
}
//-------------------------------------------------------
// calc_df ONLY FOR OIS
//-------------------------------------------------------
disc_func swap_curve_fwd_ext.calc_df(fwd_z_model fwd_model)
{
	if(curve_ok() && flt_freq() == 365) {

		/*logical apply_hyman = false;
		interpolator ip;
		try {
			ip = fwd_ip_map(fwd_model.type(), apply_hyman,	0,0);
		}
		catch {
			ip = ip_linear();
		}*/
		date max = CORE_INT.max_curve_date(curve_);
		return bootstrap_on_extrap_flat(curve_, d_,max, flt_cal());
	}
	else {
		return null<disc_func>;
	}
}

//-------------------------------------------------------
// fwd
//-------------------------------------------------------
fwd_func swap_curve_fwd_ext.fwd(fwd_z_model fwd_model)
							   
{
	logical no_disc = no_df_1d_ && null(df_disc_);
	if(no_disc) {
		df_disc_ = calc_df(fwd_model);
	}
	//cashflows_fwd cf = swap_curve_fwd_ext.cf_fwd();
	QL_REQUIRE(!null(df_disc_),'invalid disc_func (fwd)');
	QL_REQUIRE(!null(cf_), curve_msg_ != "" ? curve_msg_ : 'invalid cashflows_fwd (fwd)');
	fwd_func res;
	
	if(!null(ff_pre_)) {
		//interpolator ip = fwd_model.fwd_ip();
		//QL_REQUIRE(!null(ip),'joining of fwd_funcs not available for current fwd model');
		//return bootstrap_fwd_ext(ff_pre_, cut_off_, cf_, df_disc_, ip); 
		res = fwd_model.fwd(ff_pre_, cut_off_, cf_, df_disc_,/* null<curve>*/curve_);	
		//return fwd_comb(fwd_model,ff_pre_,cut_off_);
	}
	else	
		res = fwd_model.fwd(cf_, df_disc_, curve_);

	if(no_disc) {
		vector(instrument) vi 	= curve_.instruments();
		integer size 			= v_size(vi);
		df_disc_ 				= CORE_INT.on_fwd_to_df(d_, vi[size-1].maturity(), res,flt_dc(),flt_cal(),true);
	}
	
	return res;
}
//-------------------------------------------------------
// fwd_spr
//-------------------------------------------------------
fwd_func swap_curve_fwd_ext.fwd_spr(fwd_func 	fwd_base,
									fwd_z_model fwd_model)
							   
{
	logical no_disc = no_df_1d_ && null(df_disc_);
	
	if(no_disc) {
		df_disc_ = calc_df(fwd_model);
	}
	//cashflows_fwd cf = swap_curve_fwd_ext.cf_fwd();
	QL_REQUIRE(!null(df_disc_),'invalid disc_func (fwd)');
	QL_REQUIRE(!null(cf_), curve_msg_ != "" ? curve_msg_ : 'invalid cashflows_fwd (fwd)');

	fwd_func res;
	if(!null(ff_pre_)) {
		//interpolator ip = fwd_model.fwd_ip();
		//QL_REQUIRE(!null(ip),'joining of fwd_funcs not available for current fwd model');	
		res = fwd_model.fwd(ff_pre_, cut_off_, fwd_base, cf_, df_disc_,/* null<curve>*/curve_);			
	}
	else	
		res = fwd_model.fwd(fwd_base, cf_, df_disc_, curve_);

	if(no_disc) {
		vector(instrument) vi 	= curve_.instruments();
		integer size 			= v_size(vi);
		df_disc_ 				= CORE_INT.on_fwd_to_df(d_, vi[size-1].maturity(), res,flt_dc(),flt_cal(),true);
	}
	
	return res;
}
//-------------------------------------------------------
// fwd_spr
//-------------------------------------------------------
fwd_func swap_curve_fwd_ext.fwd_spr(disc_func 	df_base,
									fwd_z_model fwd_model)
							   
{	
	date anchor_date = d_;
	QL_REQUIRE(!null(df_disc_),'invalid disc_func (fwd)');
	fwd_func fwd_base = fwd_func(df_disc_, flt_freq(), anchor_date, flt_dc(), flt_cal(), flt_bd(), flt_eom());
	QL_REQUIRE(!null(fwd_base),'invalid fwd_base (fwd)');
	QL_REQUIRE(!null(cf_), curve_msg_ != "" ? curve_msg_ : 'invalid cashflows_fwd (fwd)');
	if(!null(ff_pre_)) {
		//interpolator ip = fwd_model.fwd_ip();
		//QL_REQUIRE(!null(ip),'joining of fwd_funcs not available for current fwd model');	
		return fwd_model.fwd(ff_pre_, cut_off_, fwd_base, cf_, df_disc_,/* null<curve>*/curve_);			
	}
	else	
		return fwd_model.fwd(fwd_base, cf_, df_disc_, curve_);			
}
//-------------------------------------------------------
// fwd_flt_df_comb
// joining of fwd_funcs
// mostly here for backward compat but allows usage of a different fwd_pre and/or cut_off 
//-------------------------------------------------------
/*fwd_func swap_curve_fwd_ext.fwd_comb(	fwd_func 		fwd_pre,
										date 			cut_off,			
										interpolator  	fwd_ip)
							   
{			
	curve cut_curve;
	vector(number) y_adj;
	logical ok = chop_short(cut_off, curve_orig_, y_adj_orig_, cut_curve, y_adj);		
	QL_REQUIRE(ok && check_curve(cut_curve),'invalid blended fwd curve (chopped curve) (fwd)');
	cashflows_fwd cf = cf_fwd(cut_curve, y_adj);	
	QL_REQUIRE(!null(cf),'invalid cashflows_fwd (chopped curve) (fwd)');		
	return bootstrap_fwd_ext(fwd_pre, cut_off, cf, df_disc_, fwd_ip); 			
}*/
//-------------------------------------------------------
// fwd_flt_df_comb
// joining of fwd_funcs
// mostly here for backward compat but allows usage of a different fwd_pre and/or cut_off 
//-------------------------------------------------------
fwd_func swap_curve_fwd_ext.fwd_comb(	fwd_z_model		fwd_model,
										fwd_func 		fwd_pre,
										date 			cut_off)
							   
{	
	curve cut_curve;
	vector(number) y_adj ;
	logical ok = chop_short(cut_off, curve_orig_, null<vector(number)>, cut_curve, y_adj);		
	QL_REQUIRE(ok && check_curve(cut_curve),'invalid blended fwd curve (chopped curve) (fwd)');
	
	cashflows_fwd cf = cf_fwd(cut_curve.instruments(), y_adj,false);	
	QL_REQUIRE(!null(cf),'invalid cashflows_fwd (chopped curve) (fwd)');
	
	return fwd_model.fwd(fwd_pre, cut_off, cf, df_disc_);				
}
//-------------------------------------------------------
// fwd_comb_blended_curve
// mostly here for backward compat but allows usage of a different cut_off 
//-------------------------------------------------------

curve swap_curve_fwd_ext.fwd_comb_blended_curve(date cut_off)							   
{	
	curve cut_curve;
	vector(number) y_adj;
	logical ok = chop_short(cut_off, curve_orig_, null<vector(number)>, cut_curve, y_adj);	
	QL_REQUIRE(ok && check_curve(cut_curve),'invalid blended fwd curve (chopped curve) (fwd)');
	
	return cut_curve;		
}

//-------------------------------------------------------
// fwd_risk
// if const_basis --> approximation
//-------------------------------------------------------
fwd_func swap_curve_fwd_ext.fwd_risk(	fwd_z_model		fwd_model,
										disc_func option(nullable) df,
										number 			r_shift,
										logical 		const_basis,
										out disc_func 	df_disc)//input and output						   
{
	if(!curve_ok())
		return null<fwd_func>;

	vector(instrument) vi = curve_.instruments();
	vector(number) first_fix = first_flt_fix_vect(vi);	

	if(null(df))
		df = df_disc_;
	
	disc_func df_res;
	fwd_func f;
	if(no_df_1d_)
		f = CORE_INT.fwd_risk_ois(d_,curve_, fwd_model,r_shift, 365, flt_dc(),flt_cal(), /*first_fix,*/df_res);
	else
		f = CORE_INT.fwd_risk(	d_,curve_, fwd_model,df_disc,r_shift,
								const_basis,flt_freq(),flt_dc(), flt_cal(),first_fix,df_res);
	
	df_disc = df_res;
	return f;	
}

//-------------------------------------------------------
// instr_bumper_ff
//-------------------------------------------------------
fwd_func swap_curve_fwd_ext.instr_bumper_ff(fwd_z_model		fwd_model,
											number 			y_shift,															
											out fwd_func	fwd_base)
														
{
	QL_REQUIRE(curve_ok(),"blended curve not created (fwd)");
	if(null(df_disc_)){
		fwd_func tmp;
		tmp = fwd(fwd_model);
	}
	QL_REQUIRE(!null(df_disc_),'invalid disc_func (fwd)');
	vector(number) first_fix = first_flt_fix_vect(curve_.instruments());	
	return CORE_INT.instr_bumper_ff(curve_, fwd_model, y_shift, first_fix, cf_, df_disc_, ff_pre_, cut_off_,fwd_base);	
}

//-------------------------------------------------------
// instr_bumper_ff_vect
//-------------------------------------------------------
vector(fwd_func) swap_curve_fwd_ext.instr_bumper_ff_vect(	fwd_z_model		fwd_model,
															number 			y_shift,															
															vector(string) 	 option(nullable) incl,//names in curve_
															out fwd_func	fwd_base,
															out fwd_func	fwd_all_shift,
															out vector(instrument) all_instr,
															out vector(instrument) bump_instr,//size is equal to number of included instruments 
															out vector(fwd_func) f_all_v)//size is equal to number of instruments in the curve (=curve_)
{
	QL_REQUIRE(curve_ok(),"blended curve not created (fwd)");
	
	vector(integer) incli_v = null(incl) ? null<vector(integer)>: v_findi(curve_.instruments().name(), incl);
	return this.instr_bumper_ff_vect(fwd_model, y_shift, incli_v,  fwd_base, fwd_all_shift, all_instr,bump_instr, f_all_v);
	
}

//-------------------------------------------------------
// instr_bumper_ff_vect
//-------------------------------------------------------
vector(fwd_func) swap_curve_fwd_ext.instr_bumper_ff_vect(	fwd_z_model		fwd_model,
															number 			y_shift,															
															out vector(integer) incl,//input and output, size is equal to number of included instruments
															out fwd_func	fwd_base,
															out fwd_func	fwd_all_shift,
															out vector(instrument) all_instr,
															out vector(instrument) bump_instr,//size is equal to number of included instruments 
															out vector(fwd_func) f_all_v)//size is equal to number of instruments in the curve (=curve_)
{
	QL_REQUIRE(curve_ok(),"blended curve not created (fwd)");
	if(null(df_disc_)){
		fwd_func tmp;
		tmp = fwd(fwd_model);
	}
	QL_REQUIRE(!null(df_disc_),'invalid disc_func (fwd)');
	all_instr				= curve_.instruments();
	vector(number) first_fix = first_flt_fix_vect(all_instr);	
	return CORE_INT.instr_bumper_ff_vect(	curve_, fwd_model, y_shift, first_fix, cf_, df_disc_, ff_pre_, cut_off_,
											incl,fwd_base, fwd_all_shift, bump_instr, f_all_v);
	
}

//-------------------------------------------------------
// wave_bumper_ff_vect
//-------------------------------------------------------
vector(fwd_func) swap_curve_fwd_ext.wave_bumper_ff_vect(vector(instrument) src,
														fwd_z_model		fwd_model,
														number 			y_shift,
														out fwd_func	fwd_base,
														out fwd_func	fwd_all)
{
	cashflows_fwd cf 	= cf_fwd(src,null<vector(number)>,true);
	QL_REQUIRE(v_size(src) == cf.n_instruments(), "invalid cashflow size");
	//return wave_bumper_ff_vect_int(cf, fwd_model, y_shift, fwd_base, fwd_all);

	fwd_base = fwd(fwd_model);
	return CORE_INT.wave_bumper_ff_vect(d_, cf, fwd_model, fwd_base, y_shift, fwd_all);
}

//-------------------------------------------------------
// wave_bumper_ff_vect
//-------------------------------------------------------
vector(fwd_func) swap_curve_fwd_ext.wave_bumper_ff_vect(fwd_z_model		fwd_model,
														number 			y_shift,
														vector(string) option(nullable) incl,//names in curve_
														out fwd_func	fwd_base,
														out fwd_func	fwd_all,
														out vector(instrument) all_instr,
														out vector(instrument) bump_instr)
{
	QL_REQUIRE(curve_ok(),"blended curve not created (fwd)");

	
	vector(integer) incli_v = null(incl) ? null<vector(integer)>: v_findi(curve_.instruments().name(), incl);
	return this.wave_bumper_ff_vect(fwd_model,y_shift,incli_v,fwd_base,fwd_all,all_instr,bump_instr);
}
//-------------------------------------------------------
// wave_bumper_ff_vect
//-------------------------------------------------------
vector(fwd_func) swap_curve_fwd_ext.wave_bumper_ff_vect(fwd_z_model		fwd_model,
														number 			y_shift,
														out vector(integer) incl,
														out fwd_func	fwd_base,
														out fwd_func	fwd_all,
														out vector(instrument) all_instr,
														out vector(instrument) bump_instr)
{
	QL_REQUIRE(curve_ok(),"blended curve not created (fwd)");

	all_instr				= curve_.instruments();
	integer size_all 		= v_size(all_instr);
	
	//check and maybe update the incl vector
	integer size_incl;
	if(null(incl) || v_size(incl) == 0) {
		incl 		= index_vector(size_all);
		size_incl 	= size_all;
	}
	else {
		size_incl 	= v_size(incl);
		QL_REQUIRE(size_incl <= size_all,"invalid size of include vector");
		
		for(integer i=0;i<size_incl;i++){
			QL_REQUIRE(incl[i] >= 0 && incl[i] < size_all,"invalid index in include vector");
			if(i>0){
				QL_REQUIRE(incl[i] > incl[i-1],"invalid index order in include vector");
			}
		}
	}

	cashflows_fwd cf;
	if(size_incl == size_all) {
		cf = cf_;
		bump_instr = all_instr;
	}
	else {
		bump_instr 	= sub_vector(all_instr, incl) ;
		cf = cf_fwd(bump_instr,null<vector(number)>,true);
	}
	
	QL_REQUIRE(v_size(bump_instr) == cf.n_instruments(), "invalid cashflow size");
	
	fwd_base = fwd(fwd_model);
	return CORE_INT.wave_bumper_ff_vect(d_, cf, fwd_model, fwd_base, y_shift, fwd_all);
}

//-------------------------------------------------------
// sens_matrix_bump
//-------------------------------------------------------
matrix(number) swap_curve_fwd_ext.sens_matrix_bump(	vector(instrument) 	src,
													number 			y_shift,
													number			nominal,
													fwd_func		fwd_base,
													fwd_func		fwd_all,
													vector(fwd_func) fwd_incl_v)//must match size of the src-vector 
{
	QL_REQUIRE(v_size(src) > 0,"invalid instrument vector (fwd)");
	//if(null(df_disc_)){
		//fwd_func tmp = fwd(fwd_model);
	//}
	
	QL_REQUIRE(!null(df_disc_),'invalid disc_func (fwd)');
	return CORE_INT.sens_matrix_instr_fwd(nominal,src,df_disc_,fwd_base,fwd_all,fwd_incl_v,y_shift);
}
//-------------------------------------------------------
// sens_matrix_bump
//-------------------------------------------------------
matrix(number) swap_curve_fwd_ext.sens_matrix_bump(	number 			y_shift,
													number			nominal,
													vector(integer) incl,
													fwd_func		fwd_base,
													fwd_func		fwd_all,
													vector(fwd_func) fwd_incl_v)//must match size of the incl-vector 
{
	curve imp_blend			= imp_blended_curve(fwd_base, true);
	QL_REQUIRE(check_curve(imp_blend),"invalid implied blended curve (fwd)");
	vector(instrument) vi 	= sub_vector(imp_blend.instruments(), incl) ;	
	return this.sens_matrix_bump(vi,y_shift,nominal,fwd_base,fwd_all,fwd_incl_v);
}
//-------------------------------------------------------
// sens_matrix_wave
//-------------------------------------------------------
matrix(number) swap_curve_fwd_ext.sens_matrix_wave(	vector(instrument) 	src,
													number 				y_shift,
													number				nominal,
													fwd_func			fwd_base,
													fwd_func			fwd_all,
													vector(fwd_func) 	fwd_v)//must match size of the src-vector 
{		
	QL_REQUIRE(v_size(src) > 0,"invalid instrument vector (fwd)");
	//if(null(df_disc_)){
		//fwd_func tmp = fwd(fwd_model);
	//}
	QL_REQUIRE(!null(df_disc_),'invalid disc_func (fwd)');
	return CORE_INT.sens_matrix_instr_fwd(nominal,src,df_disc_,fwd_base,fwd_all,fwd_v,y_shift);
}
//-------------------------------------------------------
// sens_matrix_wave
//-------------------------------------------------------
matrix(number) swap_curve_fwd_ext.sens_matrix_wave(	number 			y_shift,
													number			nominal,
													vector(integer) incl,
													fwd_func		fwd_base,
													fwd_func		fwd_all,
													vector(fwd_func) fwd_incl_v)//must match size of the incl-vector 
{		
	curve imp_blend	= imp_blended_curve(fwd_base,true);
	QL_REQUIRE(check_curve(imp_blend),"invalid implied blended curve (fwd)");
	vector(instrument) vi 	= sub_vector(imp_blend.instruments(), incl) ;
	return this.sens_matrix_wave(vi, y_shift, nominal, fwd_base, fwd_all, fwd_incl_v);
}

/*----- CREATE FUNCS ------*/

//-------------------------------------------------------
// create func
//-------------------------------------------------------
swap_curve_fwd_ext swap_curve_fwd_ext(	fwd_func_tenor 				fwd_tenor,
										swap_curve_fwd_f_parm		p,	
										date						d,							
										disc_func 					df,
										disc_func option(nullable)	df_synt_base,
										date						synt_end,
										string 						name,
										fwd_func option(nullable) 	ff_pre = null<fwd_func>, 
										date option(nullable)		cut_off = null<date>) 									
{	
	if(null(ff_pre))			
		return new swap_curve_fwd_ext(fwd_tenor,p,d,df,df_synt_base,synt_end,name,null<fwd_func>,null<date>);		
	else {
		QL_REQUIRE(!null(cut_off),"cut-off date is required");
		return new swap_curve_fwd_ext(fwd_tenor,p,d,df,df_synt_base,synt_end,name,ff_pre,cut_off);	
	}
}

//-------------------------------------------------------
// create func
//-------------------------------------------------------
swap_curve_fwd_ext swap_curve_fwd_ext(	fwd_func_tenor 				fwd_tenor,
										swap_curve_fwd_f_parm		p,	
										date						d,							
										disc_func option(nullable)	df,	
										string 						name,
										fwd_func option(nullable) 	ff_pre = null<fwd_func>, 
										date option(nullable)		cut_off = null<date>) 									
{	
	if(null(ff_pre))			
		return new swap_curve_fwd_ext(fwd_tenor,p,d,df,name,null<fwd_func>,null<date>);		
	else {
		QL_REQUIRE(!null(cut_off),"cut-off date is required");
		return new swap_curve_fwd_ext(fwd_tenor,p,d,df,name,ff_pre,cut_off);	
	}
}

//-------------------------------------------------------
// create func [LEGACY version]
//-------------------------------------------------------
swap_curve_fwd_ext swap_curve_fwd_ext(	swap_curve_ext_parm		s,	
										date					d,							
										disc_func option(nullable)	df,	
										string 					name,
										fwd_func option(nullable) ff_pre = null<fwd_func>, 
										date option(nullable)	cut_off = null<date>) 									
{	
	if(null(ff_pre))			
		return new swap_curve_fwd_ext(s,d,df,name);		
	else {
		QL_REQUIRE(!null(cut_off),"cut-off date is required");
		return new swap_curve_fwd_ext(s,d,df,name,ff_pre,cut_off);	
	}
}
//-------------------------------------------------------
// create func  [LEGACY version]
//-------------------------------------------------------
swap_curve_fwd_ext swap_curve_fwd_ext(	swap_curve_ext_parm		s,
										date					d,	
										disc_func option(nullable)	df,								
										integer					flt_freq,
										day_count_method 		flt_dc,
										bd_convention 			flt_bd,
										calendar				flt_cal,
										logical					flt_eom,
										integer					flt_reset_days,
										string					settle_code,
										string 					name,
										fwd_func option(nullable) ff_pre = null<fwd_func>, 
										date option(nullable)	cut_off = null<date>) 
{	
	if(null(ff_pre))
		return new swap_curve_fwd_ext(s,d,df,flt_freq, flt_dc, flt_bd, flt_cal, flt_eom,
									  flt_reset_days, settle_code, name);
	else {
		QL_REQUIRE(!null(cut_off),"cut-off date is required");		
		return new swap_curve_fwd_ext(s,d,df,flt_freq, flt_dc, flt_bd, flt_cal, flt_eom,
									  flt_reset_days, settle_code, name,ff_pre,cut_off);
	}
}



